; Forth92

;; Copyright (c) 1999-2000 David Kuehling  <dvdkhlng TA gmx TOD de>

;; Permission is hereby granted, free of charge, to any person obtaining a copy
;; of this software and associated documentation files (the "Software"), to
;; deal in the Software without restriction, including without limitation the
;; rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
;; sell copies of the Software, and to permit persons to whom the Software is
;; furnished to do so, subject to the following conditions:
;;
;; The above copyright notice and this permission notice (including the next
;; paragraph) shall be included in all copies or substantial portions of the
;; Software.
;;
;; THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
;; IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
;; FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
;; AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
;; LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
;; FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
;; IN THE SOFTWARE.
;
; Unfortunately, this file needs to be preprocessed by my PreFargo
; preprocessor before you can compile it. Send me a mail for it or have a 
; look at http://www.ticalc.org
; If I've some time, I'll try to switch to the m4 preprocessor.
;

	
/********************************************************************************/
		#include <tios.h>
		#include <flib.h>
		#include <forthlib.h>
		#include "FORTH92m.s"		// contains the macros for creating threaded code
/********************************************************************************/
	
/********************************************************************************/
		XDEF	_main
		XDEF	_comment
/********************************************************************************/

/*******************************************************************************
 * INTERRUPT REDIRECTION
 */

/****
 * Redirection Data
 */
	// number of redirections
#define REDIRECT_NUM 0
#define REDIRECT(vect_addr,new_vect) {
	dc.b	vect_addr
	dc.b	new_vect-(*+1)
	#equ	REDIRECT_NUM REDIRECT_NUM+1
}			
#define REDIRECT_END  dc.w 0

Redirections:
/*  0 */	REDIRECT($0C, IntAddrErr)		/* Address Error Vector */
/*  1 */	REDIRECT($10, IntIllegInstr)		/* Illegal Instruction Vector */
/*  2 */	REDIRECT($14, IntZeroDiv)		/* Zero Divide Vector */
/*  3 */	REDIRECT($18, IntChk)			/* CHK Instruction Vector */
/*  4 */	REDIRECT($78, IntAuto6)			/* Auto Int 6 (ON-key int) */
	REDIRECT_END

/****
 * New Interrupt Routines
 */
IntAddrErr:	moveq	#ERR_ALIGN, d0			// set exception-code
		bra	DoError
IntIllegInstr:	move.w	#ERR_ILLEG_INSTR, d0		// set exception-code
		bra	DoError
IntZeroDiv:	moveq	#ERR_DIVISION, d0		// set exception-code
		bra	DoError
	/** ON-Key interrupt: check for ON+C combo which quits FORTH */
IntAuto6:	move.w	#%1111110111, $600018		// set keyboard-row mask
		move.l	(a7), (a7)			// some dealy...
		move.l	(a7), (a7)
		move.l	(a7), (a7)
		btst.b	#1, $60001B			// check keyboard-column mask
		bne	\NoOnPlusC			// if ON+C is hold down:
		move	#$0000, SR			//    switch to user mode
		bra	ExitFORTH			//    and exit FORTH

\NoOnPlusC:	move.l	OldInts+4*4, -(a7)		// if ON+C isn't hold down:
		rts					//    trigger original AutoInt6
		
/** 
 * Chk is triggered, when float or parameter-stack over- or underflow or when an 
 * illegal address is accessed. So now we have to check the float and parameter stack
 * to get out which of those events happened.
 */
IntChk:		// d0 = number of bytes on stack
		move.l	USP, a0		// check for stack over- and underflow
		move.l	PSB, d0
		sub.l	a0, d0
		bmi	\PSUnderflow			
		cmpi.w	#PS_SIZE, d0
		bgt	\PSOverflow
		move.l	FSB, d0		// check for float-stack over- and underflow
		sub.l	FSP, d0
		bmi	\FSUnderflow
		cmpi.w	#FS_SIZE*6, d0
		bgt	\FSOverflow
						// if it wasn't a stack- over or underflow:
		moveq	#ERR_INV_ADDRESS, d0	//   it was certainly an access to an illegal address
		bra	DoError
\FSOverflow:	moveq	#ERR_FS_OVERFLOW, d0	// floating-point-stack-overflow
		bra	DoError
\FSUnderflow:	moveq	#ERR_FS_UNDERFLOW, d0	// floating-point-stack-underflow
		bra	DoError
\PSOverflow:	moveq	#ERR_S_OVERFLOW, d0	// parameter-stack-overflow
		bra	DoError
\PSUnderflow:	moveq	#ERR_S_UNDERFLOW, d0	// parameter-stack-underflow

DoError:	addq.l	#6, a7			// remove return-address and stored SR
		move	#$0000, SR		// switch back to user mode
             	bra	DoThrow			// and trigger FORTH exception


/****
 * Begin redirection -- load address of `Redirection' and `OldInts', switch off ints and memory protection
 * Output:
 *	d0.l = old interrupt mask
 *    	a0.l = address of `Redirections'
 *      a1.l = address of `OldInts'
 */
StartRedirect:	move.w	#$0700, d0			// switch off ints
		trap	#1
		bclr.b	#2, $600001			// switch off memory protection
		lea	Redirections(PC), a0		// a0 = address of redirection-data
		lea	OldInts, a1			// a1 = address of table that holds the old vectors
		rts

/****
 * End redirection -- switch ints and memory protection on
 * Input:
 *	d0.w = old interrupt mask
 */
EndRedirect:	trap	#1				// restore interrupt mask
		bset.b	#2, $600001			// switch memory protection on
		rts

/**** 
 * Get redirection address pair
 * Input:
 * 	a0.l = pointer to current entry in `Redirections'
 * Output:
 *      a0.l = pointer to next entry
 *      a2.l = address of vector
 *      a3.l = address of new routine
 *      d1.w = 0 if no more addresses in `Redirections', else undefined
 *               (Z-flag is set corresponding to d1)
 *      d2.w destroyed
 */
GetRedirect:	clr.w	d1
		move.b	(a0)+, d1			// d1.w = address of vector
		beq	\End				//   if it is zero: end of list
                clr.w	d2
		move.b	(a0)+, d2			// d2.w = relative address of new routine
		movea.w	d1, a2				// a2.l = address of vector
		move.l	a0, a3				// a3.l = absolute address of new routine
		adda.w	d2, a3
\End:		rts

/****
 * Redirect all the ints, that are stored in the table `Redirections'
 */
RedirectInts:	bsr	StartRedirect			// a0.l = *redirect-table  a1.l = *old-vectors

\Loop:		bsr	GetRedirect			// a2.l = *vector  a3.l = *new-routine
		beq	\End
		move.l	(a2), (a1)+			// save the old vector
		move.l	a3, (a2)			// set vector to new routine
		bra	\Loop

\End:		bsr	EndRedirect
		rts

/****
 * Restore the vectors, that were redirected by `RedirectInts'
 */
UnRedirectInts:	bsr	StartRedirect			// a0.l = *redirect-table  a1.l = *old-vectors

\Loop:		bsr	GetRedirect			// a2.l = *vector
		beq	\End
		move.l	(a1)+, (a2)			// restore the old vector
		bra	\Loop

\End:		bsr	EndRedirect
		rts
		

/******************************************************************************
 * MAIN - initialize registers, redirect ints and start execution of hello
 ******************************************************************************/
_main:		/******* Create the FORTH Text Screen */
		jsr	forthlib::CreateScreen
		tst.l	d0
		beq	\ScreenOk
		rts

\ScreenOk:	bsr	RedirectInts			// redirect interrupts
		move.l	a7, SavedSP			// save a7 (is used by InitRegs)
			
BootFORTH:	lea	hello+2(PC), IP			// IP: instruction-pointer
		lea	RelativePointerBase(PC), BP	// BP: base-pointer
		movea.l	SavedSP(PC), PSP		// load PSP = a7 = parameter-stack-ptr
		suba.w	#MAX_UNDERFLOW, PSP		// add underflow safety-area to Parameter Stack
		move.l	PSP, PSB			// PSB: parameter-stack-base (bottom)
		move.l	#ReturnStackTop, RST		// RST: return-stack-top
		lea	ReturnStack, RSP		// RSP: return-stack-pointer
		lea	FloatStack, FSP			// FSP: float-stack-pointer
		move.l	FSP, FSB			// FSB: float-stack-base (bottom)
		lea	Next(PC), NP			// NP: next-pointer
		jmp	(NP)				// start interpretation

/*******************************************************************************
 * Exit FORTH
 */
ExitFORTH:	movea.l SavedSP(PC), a7			// restore the SP that returns to Fargo
		bsr	UnRedirectInts			// restore vector table
		jsr	forthlib::DestroyScreen		// destroy the FORTH screen
		rts					// return to Fargo

/*---------------------------------------
 * Helper routine, that has the purpose of tios::DEREF d0,a0 where d0 is NOT modified!!! 
 * (function calls are much smaller than the DEREF macro)
 */
DEREF_d0_a0:	move.w	d0, -(a7)
		tios::DEREF	d0, a0
		move.w	(a7)+, d0
		rts

/** Central Next. Primary words jump here by  `jmp (NP)'  (unless they contain next) */
Next:		NEXT

/******************************************************************************
 ******************************************************************************
 *** The FORTH dictionary						    ***
 ******************************************************************************
 ******************************************************************************/

/******************************************************************************
 * 1. Stack- and Memory-Access, Arithmetic and Logical Operations
 ******************************************************************************/
/**
 * 1.1 Access the Data- and Returnstack
 */
PRIM_WORD (DUP, "DUP")			// ( x -- x x )					CORE
		move.w	(PSP), -(PSP)
		DO_NEXT
PRIM_WORD (TWO_DUP, "2DUP")		// ( x1 x2 -- x1 x2 x1 x2 )			CORE
		move.l	(PSP), -(PSP)
		DO_NEXT
PRIM_WORD (QUESTION_DUP, "?DUP")	// ( x -- 0 | x x )				CORE
		move.w	(PSP), d0
		beq	\Zero
		move.w	d0, -(PSP)
\Zero:		DO_NEXT
PRIM_WORD (DROP, "DROP")		// ( x -- )					CORE
		addq.l	#2, PSP
		DO_NEXT
PRIM_WORD (TWO_DROP, "2DROP")		// ( x1 x2 -- )					CORE
		addq.l	#4, PSP
		DO_NEXT
PRIM_WORD (OVER, "OVER")		// ( x1 x2 -- x1 x2 x1 )			CORE
		move.w	2(PSP), -(PSP)
		DO_NEXT
PRIM_WORD (TWO_OVER, "2OVER")		// ( x1 x2 x3 x4 -- x1 x2 x3 x4 x1 x2 )		CORE
		move.l	4(PSP), -(PSP)
		DO_NEXT
PRIM_WORD (ROT, "ROT")			// ( x1 x2 x3 -- x2 x3 x1 )			CORE
		move.l	(PSP)+, d0
		move.w	(PSP)+, d1
		move.l	d0, -(PSP)
		move.w	d1, -(PSP)
		DO_NEXT
PRIM_WORD (TWO_ROT, "2ROT")		// ( x1 x2 x3 x4 x5 x6 -- x3 x4 x5 x6 x1 x2 )	DOUBLE EXT
		movem.l	(PSP)+, d0-d2
		exg.l	d0, d2
		exg.l	d1, d2
		movem.l	d0-d2, -(PSP)	
		DO_NEXT
PRIM_WORD (_SWAP, "SWAP")		// ( x1 x2 -- x2 x1 )				CORE
		move.l	(a7)+, d0
		swap	d0
		move.l	d0, -(a7)
		DO_NEXT
PRIM_WORD (TWO_SWAP, "2SWAP")		// ( x1 x2 x3 x4 -- x3 x4 x1 x2 )		CORE
		move.l	(PSP)+, d0
		move.l	(PSP)+, d1
		move.l	d0, -(PSP)
		move.l	d1, -(PSP)
		DO_NEXT
PRIM_WORD (DEPTH, "DEPTH")		// ( -- n )					CORE
		move.l	PSB, d0
		sub.l	PSP, d0
		asr.w	#1, d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (NIP, "NIP")			// ( x1 x2 -- x2 )				CORE EXT
		move.w	(PSP)+, (PSP)
		DO_NEXT
PRIM_WORD (PICK, "PICK")		// ( xu ... x1 x0 u -- xu ... x1 x0 xu )	CORE EXT
		move.w	(PSP), d0
		lsl.w	#1, d0
		move.w	2(PSP,d0.w), (PSP)
		DO_NEXT
PRIM_WORD (ROLL, "ROLL")		// ( xu xu-1 ... x0 u -- xu-1 ... x0 xu )	CORE EXT
		move.w	(PSP)+, d0			// d0 = u
		move.w	d0, d1		
		subq.w	#1, d1				// d1 = counter
		bmi	\NoOp
		lsl.w	#1, d0				// d0 = 2u
		lea	0(PSP,d0.w), a0			// a0 = *xu
		move.w	(a0), d0			// d0 = xu
\Loop:		move.w	-(a0), 2(a0)
		dbra	d1, \Loop
		move.w	d0, (PSP)
\NoOp:		DO_NEXT
PRIM_WORD (TUCK, "TUCK")		// ( x1 x2 -- x2 x1 x2 )			CORE EXT
		move.l	(PSP), d0
		swap	d0
		move.l	d0, (PSP)
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (R_FROM, "R>")		// ( -- x ) ( R: x -- ) 			CORE
		move.w	(RSP)+, -(PSP)
		DO_NEXT
PRIM_WORD (TWO_R_FROM, "2R>")		// ( -- x1 x2 ) ( R: x1 x2 -- )			CORE EXT
		move.l	(RSP)+, -(PSP)
		DO_NEXT
PRIM_WORD (TO_R, ">R")			// ( x -- ) ( R: -- x ) 			CORE
		move.w	(PSP)+, -(RSP)
		DO_NEXT
PRIM_WORD (TWO_TO_R, "2>R")		// ( x1 x2 -- ) ( R: x2 x1 -- )			CORE EXT
		move.l	(PSP)+, -(RSP)
		DO_NEXT
PRIM_WORD (R_FETCH, "R@")		// ( -- x ) ( R: x -- x ) 			CORE
		move.w	(RSP), -(PSP)
		DO_NEXT
PRIM_WORD (TWO_R_FETCH, "2R@")		// ( -- x1 x2 ) ( R: x1 x2 -- x1 x2 )		CORE EXT
		move.l	(RSP), -(PSP)
		DO_NEXT
INTERNAL_PRIM_WORD (sp_store, "sp!")	// ( i*x -- )					internal
		move.l	PSB, PSP			// initialize Parameterstack Pointer (PSP)
		jmp	(NP)
INTERNAL_PRIM_WORD (rp_store, "rp!")	// ( R: i*x -- )				internal
		lea	ReturnStack, RSP		// initialize Returnstack Pointer (RSP)
		jmp	(NP)
		

/**
 * 1.2 Read, Write and Transfer Memory Contents
 */
PRIM_WORD (FETCH, "@")			// ( a-addr -- x ) 				CORE
		move.w	(PSP), d0
		move.w	0(BP, d0.w), (PSP)
		DO_NEXT
PRIM_WORD (TWO_FETCH, "2@")		// ( a-addr -- x1 x2 )				CORE
		move.w	(PSP)+, d0
		move.l	0(BP, d0.w), -(PSP)
		DO_NEXT
PRIM_WORD (C_FETCH, "C@")		// ( c-addr -- char )				CORE
		move.w	(PSP), d0
		clr.w	d1
		move.b	0(BP, d0.w), d1
		move.w	d1, (PSP)
		DO_NEXT
PRIM_WORD (STORE, "!")			// ( x a-addr -- ) 				CORE
		move.w	(PSP)+, d0
		chk	#RELP(AccessBoundary), d0		// check address
		move.w	(PSP)+, 0(BP, d0.w)
		DO_NEXT
PRIM_WORD (TWO_STORE, "2!")		// ( x1 x2 a-addr -- )				CORE
		move.w	(PSP)+, d0
		chk	#RELP(AccessBoundary), d0		// check address
		move.l	(PSP)+, 0(BP, d0.w)
		DO_NEXT
PRIM_WORD (C_STORE, "C!")		// ( char c-addr -- )				CORE
		move.w	(PSP)+, d0
		chk	#RELP(AccessBoundary), d0		// check address
		move.w	(PSP)+, d1
		move.b	d1, 0(BP, d0.w)
		DO_NEXT
PRIM_WORD (PLUS_STORE, "+!")		// ( n|u a-addr -- )				CORE
		move.w	(PSP)+, d0
		chk	#RELP(AccessBoundary), d0		// check address
		move.w	(PSP)+, d1
		add.w	d1, 0(BP, d0.w)
		DO_NEXT
PRIM_WORD (ALIGNED, "ALIGNED")		// ( addr -- a-addr )				CORE
		addq.w	#1, (PSP)
		and.w	#$FFFE, (PSP)
		DO_NEXT
PRIM_WORD (FILL, "FILL")		// ( c-addr u char -- )				CORE
		move.w	(PSP)+, d0			// d0 = char
__DO_FILL__:	move.w	(PSP)+, d1			// d1 = u (#chars)
		beq	\Zero
		subq	#1, d1				// d1 = counter
		move.w	(PSP)+, d2			// d2 = c-addr
\Loop:		chk	#RELP(AccessBoundary), d2	   // check, whether address is in range
		move.b	d0, 0(BP, d2.w)		           // write char
		addq.w	#1, d2				   // go to next character
		dbra	d1, \Loop
\Zero:		DO_NEXT
PRIM_WORD (ERASE, "ERASE")		// ( addr u -- )				CORE
		clr.w	d0				// fill with zeroes
		bra	__DO_FILL__
PRIM_WORD (_MOVE, "MOVE")		// ( src-addr dst-addr u -- )			CORE
		move.w	2(PSP), d0			// d0 = dst-addr
		cmp.w	4(PSP), d0
		bgt	CMOVE_UP+2			// dst-addr > src-addr: CMOVE>
		bra	CMOVE+2				// dst-addr <= src-addr: CMOVE
		
					// ( a-addr1 -- a-addr2 )			CORE
SAME_PRIM_WORD (CELL_PLUS, "CELL+", TWO_PLUS)
 					// ( n1 -- n2 )					CORE
SAME_PRIM_WORD (CELLS, "CELLS",	TWO_STAR) 	
				 	// ( n1 -- n2 )					CORE
SAME_PRIM_WORD (CHAR_PLUS, "CHAR+", ONE_PLUS)
NOOP_WORD (CHARS, "CHARS")		// ( n1 -- n2 )					CORE


PRIM_WORD (A_TO_R, "A>R")		// ( far-addr -- addr )				????
		move.l	(PSP)+, a0
		suba.l	BP, a0
		move.w	a0, -(PSP)
		DO_NEXT
PRIM_WORD (R_TO_A, "R>A")		// ( addr - far-addr )				????
		move.w	(PSP)+, a0
		adda.l	BP, a0
		move.l	a0, -(PSP)
		DO_NEXT
PRIM_WORD (A_FETCH, "A@")		// ( far-a-addr -- x )				????
		move.l	(PSP)+, a0
		move.w	(a0), -(PSP)
		DO_NEXT
PRIM_WORD (A_STORE, "A!")		// ( x far-a-addr -- )				????
		move.l	(PSP)+, a0
		move.w	(PSP)+, (a0)
		DO_NEXT
PRIM_WORD (TWO_A_FETCH, "2A@")		// ( far-a-addr -- x1 x2 )			????
		move.l	(PSP)+, a0
		move.l	(a0), -(PSP)
		DO_NEXT
PRIM_WORD (TWO_A_STORE, "2A!")		// ( x1 x2 far-a-addr -- )			????
		move.l	(PSP)+, a0
		move.l	(PSP)+, (a0)
		DO_NEXT
PRIM_WORD (CA_FETCH, "CA@")		// ( far-addr -- char )				????
		move.l	(PSP)+, a0
		clr.w	d0
		move.b	(a0), d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (CA_STORE, "CA!")		// ( char far-addr -- )				????
		move.l	(PSP)+, a0
		move.w	(PSP)+, d0
		move.b	d0, (a0)
		DO_NEXT
PRIM_WORD (DEREF, "DEREF")		// ( hdl -- far-addr )				????
		move.w	(PSP)+, d0
		bsr	DEREF_d0_a0
		move.l	a0, -(PSP)
		jmp	(NP)

/**
 * 1.3 Arithmetic and Logical/Binary Operations
 */
PRIM_WORD (PLUS, "+")			// ( n1|u1 n2|u2 -- n3|u3 )			CORE
		move.w	(PSP)+, d0
		add.w	d0, (PSP)
		DO_NEXT
PRIM_WORD (D_PLUS, "D+")		// ( d1|ud1 d2|ud2 -- d3|ud3 )			DOUBLE
		move.l	(PSP)+, d0
		add.l	d0, (PSP)
		DO_NEXT
PRIM_WORD (M_PLUS, "M+")		// ( d1|ud1 n -- d2|ud2 )			DOUBLE
		move.w	(PSP)+, d0
		ext.l	d0
		add.l	d0, (PSP)
		DO_NEXT
PRIM_WORD (MINUS, "-")			// ( n1|u1 n2|u2 -- n3|u3 )			CORE
		move.w	(PSP)+, d0
		sub.w	d0, (PSP)
		DO_NEXT
PRIM_WORD (D_MINUS, "D-")		// ( d1|ud1 d2|ud2 -- d3|ud3 )			DOUBLE
		move.l	(PSP)+, d0
		sub.l	d0, (PSP)
		DO_NEXT
PRIM_WORD (STAR, "*")			// ( n1 n2 -- n3 )				CORE
		move.w	(PSP)+, d0
		muls	(PSP), d0
		move.w	d0, (PSP)
		DO_NEXT
PRIM_WORD (SLASH, "/")			// ( n1 n2 -- n3 )				CORE
		move.w	(PSP)+, d2			// d2 = n2
		move.w	(PSP), d1			// d1 = n1
		ext.l	d1
		divs	d2, d1				// d1 = n1 / n2
		move.w	d1, (PSP)
		DO_NEXT
PRIM_WORD (MOD, "MOD")			// ( n1 n2 -- n3 )				CORE
		move.w	(PSP)+, d2			// d2 = n2
		move.w	(PSP), d1			// d1 = n1
		ext.l	d1
		divs	d2, d1				// d1(high) = n1 MOD n2
		swap	d1
		move.w	d1, (PSP)
		DO_NEXT
PRIM_WORD (SLASH_MOD, "/MOD")		// ( n1 n2 -- n3 n4 )				CORE
		move.w	(PSP)+, d2			// d2 = n2
		move.w	(PSP)+, d1			// d1 = n1
		ext.l	d1
		divs	d2, d1				// d1(low) = n1 / n2; d1(high) = n1 MOD n2
		swap	d1				// d1(high) = n1 / n2; d1(low) = n1 MOD n2
		move.l	d1, -(PSP)			// -- d1(low) d1(high)
		DO_NEXT

/********************************************************************************
 * M*			n1 n2 -- d				 	CORE	
 */
PRIM_WORD (M_STAR, "M*")
		move.w	(PSP)+, d0
		muls	(PSP)+, d0
		move.l	d0, -(PSP)
		DO_NEXT

//********************************************************************************
// */			n1 n2 n3 -- n4				 	CORE	
//
PRIM_WORD (STAR_SLASH, "*/")
		move.w	(PSP)+, d2			// d2 = n3
		move.w	(PSP)+, d1			// d1 = n2
		muls	(PSP), d1			// d1 = n1 * n2
		divs	d2, d1				// d1 = n1 * n2 / n3
		move.w	d1, (PSP)
		DO_NEXT

//*******************************************************************************
// M*/			d1 n1 +n2 -- d2				 	DOUBLE
//
PRIM_WORD (M_STAR_SLASH, "M*/")
		clr.w	d3			// d3 = sign-flag
		move.w	(PSP)+, d2		// d2 = n2
		move.w	(PSP)+, d1		// d1 = n1
		bpl	\N1Pos
		not.w	d3			// transfer sign to d3
		neg.w	d1
\N1Pos:		move.l	(PSP)+, d0		// d0 = d1
		bpl	\D1Pos
		not.w	d3			// transfer sign to d3
		neg.l	d0
\D1Pos:
		move.w	d3, -(a7)
		move.w	d0, d3			// d3 = low(d1)
		swap	d0
		mulu	d1, d0			// d0 = high(d1)*n1
		mulu	d1, d3			// d3 = low(d1)*n1
		move.w	d3, d1			// d1.w = low(low(d1)*n1)
		clr.w	d3
		swap	d3			// d3.l = high(low(d1)*n1)
		add.l	d3, d0			// d0.l = high(d1)*n1 + high(low(d1)*n1)
	
		divu	d2, d0			// divide (d0|d1) by d2 ...
		move.w	d0, d3
		move.w	d1, d0
		divu	d2, d0			
		swap	d3
		move.w	d0, d3
		
		tst.w	(a7)+			// adjust sign 
		beq	\D2Pos
		neg.l	d3
\D2Pos:		move.l	d3, -(PSP)		// push d2
		jmp	(NP)

//*******************************************************************************
// */MOD			n1 n2 n3 -- n4 n5			 	CORE	
//
PRIM_WORD (STAR_SLASH_MOD, "*/MOD")
		move.w	(PSP)+, d2			// d2 = n3
		move.w	(PSP)+, d1			// d1 = n2
		mulu	(PSP)+, d1			// d1 = n1 * n2
		divu	d2, d1				// d1(low) = n1 * n2 / n3; d1(high) = remainder
		swap	d1				// d1(high) = n1 * n2 / n3; d1(low) = remainder
		move.l	d1, -(PSP)			// -- d1(low) d1(high)
		DO_NEXT

/********************************************************************************
 * UM*		u1 u2 -- ud					 	CORE	
 */
PRIM_WORD (UM_STAR, "UM*")	
		move.w	(PSP)+, d0
		mulu	(PSP)+, d0
		move.l	d0, -(PSP)
		DO_NEXT

/********************************************************************************
 * UM/MOD	 ud1 u1 -- u2 u3 				 	CORE	
 */
PRIM_WORD (UM_SLASH_MOD, "UM/MOD")
		move.w	(PSP)+, d2			// d2 = n2
		move.l	(PSP)+, d1			// d1 = n1
		divu	d2, d1				// d1(low) = n1 / n2; d1(high) = n1 MOD n2
		swap	d1				// d1(high) = n1 / n2; d1(low) = n1 MOD n2
		move.l	d1, -(PSP)			// -- d1(low) d1(high)
		DO_NEXT

/********************************************************************************
 * SM/REM		d n1 -- n2 n3				 	CORE	
 */
PRIM_WORD (SM_SLASH_REM, "SM/REM")
		move.w	(PSP)+, d2		// d2 = n2
		move.l	(PSP)+, d1		// d1 = n1
		divs	d2, d1			// d1(low) = n1 / n2; d1(high) = n1 MOD n2
		swap	d1			// d1(high) = n1 / n2; d1(low) = n1 MOD n2
		move.l	d1, -(PSP)		// -- d1(low) d1(high)
		DO_NEXT

/********************************************************************************
 * FM/MOD		d n1 -- n2 n3				 	CORE	
 */
PRIM_WORD (FM_SLASH_MOD, "FM/MOD")
		move.w	(PSP)+, d2		// d2 = n2
		move.l	(PSP)+, d1		// d1 = n1
		divs	d2, d1			// d1(low) = n1 / n2; d1(high) = n1 MOD n2
		bpl	\End1			// if quotient is negative
		swap	d1
		tst.w	d1			// and not zero: we have to adjust the result
		beq	\End2
		add.w	d2, d1			//   increase remainder 
		swap	d1
		subq.w	#1, d1			//   and decrease quotient
\End1:		swap	d1
\End2:		move.l	d1, -(PSP)	
		DO_NEXT

/********************************************************************************
 * 1+			n1|u1 -- n2|u2				 	CORE	
 */
PRIM_WORD (ONE_PLUS, "1+")
		addq.w	#1, (PSP)
		DO_NEXT

/********************************************************************************
 * 1-			n1|u1 -- n2|u2				 	CORE	
 */
PRIM_WORD (ONE_MINUS, "1-")
		subq.w	#1, (PSP)
		DO_NEXT

/********************************************************************************
 * 2+			n1|u1 -- n2|u2				 	????
 */
PRIM_WORD (TWO_PLUS, "2+")	
		addq.w	#2, (PSP)
		DO_NEXT

/********************************************************************************
 * 2-			n1|u1 -- n2|u2				 	????
 */
PRIM_WORD (TWO_MINUS, "2-")		
		subq.w	#2, (PSP)
		DO_NEXT

/********************************************************************************
 * 1+!			a-addr --				 	????
 * 
 * possible Definition:   : 1+!   1 SWAP +! ;
 */
PRIM_WORD (ONE_PLUS_STORE, "1+!")	
		move.w	(PSP)+, d0
		addq.w	#1, 0(BP, d0.w)
		DO_NEXT

/********************************************************************************
 * 1-!			a-addr --				 	????
 * 
 * possible Definition:   : 1-!   -1 SWAP +! ;
 */
PRIM_WORD (ONE_MINUS_STORE, "1-!")
		move.w	(PSP)+, d0
		subq.w	#1, 0(BP, d0.w)
		DO_NEXT

/********************************************************************************
 * 2*			x1 -- x2				 	CORE	
 */
PRIM_WORD (TWO_STAR, "2*")		
		move.w	(PSP), d0
		asl.w	#1, d0
		move.w	d0, (PSP)
		DO_NEXT

/********************************************************************************
 * D2*			xd1 -- xd2				 	DOUBLE
 */
PRIM_WORD (D_TWO_STAR, "D2*")
		move.l	(PSP), d0
		asl.l	#1, d0
		move.l	d0, (PSP)
		DO_NEXT

/********************************************************************************
 * 2/			x1 -- x2				 	CORE	
 */
PRIM_WORD (TWO_SLASH, "2/")	
		move.w	(PSP), d0
		asr.w	#1, d0
		move.w	d0, (PSP)
		DO_NEXT

/********************************************************************************
 * D2/			xd1 -- xd2				 	DOUBLE	
 */
PRIM_WORD (D_TWO_SLASH, "D2/")		// ( xd1 -- xd2 )				DOUBLE
		move.l	(PSP), d0
		asr.l	#1, d0
		move.l	d0, (PSP)
		DO_NEXT

/********************************************************************************
 * NEGATE		n1 -- n2				 	CORE	
 */
PRIM_WORD (NEGATE, "NEGATE")		// ( n1 -- n2 )					CORE
		neg.w	(PSP)
		DO_NEXT

/********************************************************************************
 * DNEGATE		d1 -- d2				 	DOUBLE
 */
PRIM_WORD (DNEGATE, "DNEGATE")		// ( d1 -- d2 )					DOUBLE
		neg.l	(PSP)
		DO_NEXT

PRIM_WORD (ABS, "ABS")			// ( n -- u )					CORE
		tst.w	(PSP)
		bpl	\NotNeg
		neg.w	(PSP)
\NotNeg:	DO_NEXT
PRIM_WORD (DABS, "DABS")		// ( d -- ud )					DOUBLE
		tst.l	(PSP)
		bpl	\NotNeg
		neg.l	(PSP)
\NotNeg:	DO_NEXT
PRIM_WORD (MAX, "MAX")			// ( n1 n2 -- n3 )				CORE
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP), d0			// n2 > n1 ?
		ble	\N1Greater
		move.w	d0, (PSP)			//    -> replace n1 by n2
\N1Greater:	DO_NEXT
PRIM_WORD (D_MAX, "DMAX")		// ( d1 d2 -- d3 )				DOUBLE
		move.l	(PSP)+, d0			// D0 = d2
		cmp.l	(PSP), d0			// d2 > d1 ?
		ble	\N1Greater
		move.l	d0, (PSP)			//    -> replace d1 by d2
\N1Greater:	DO_NEXT
PRIM_WORD (MIN, "MIN")			// ( n1 n2 -- n3 )				CORE
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP), d0			// n2 < n1 ?
		bge	\N1Smaller
		move.w	d0, (PSP)			//    -> replace n1 by n2
\N1Smaller:	DO_NEXT
PRIM_WORD (D_MIN, "DMIN")		// ( d1 d2 -- d3 )				DOUBLE
		move.l	(PSP)+, d0			// D0 = d2
		cmp.l	(PSP), d0			// d2 < d1 ?
		bge	\N1Smaller
		move.l	d0, (PSP)			//    -> replace d1 by d2
\N1Smaller:	DO_NEXT
PRIM_WORD (S_TO_D, "S>D")		// ( n -- d )					CORE
		move.w	(PSP)+, d0
		ext.l	d0
		move.l	d0, -(PSP)
		DO_NEXT
SAME_PRIM_WORD (D_TO_S, "D>S", DROP)	// ( d -- n )					DOUBLE
PRIM_WORD (INVERT, "INVERT")		// ( x1 -- x2 )					CORE
		not.w	(PSP)
		DO_NEXT
PRIM_WORD (_AND, "AND")			// ( x1 -- x2 )					CORE
		move.w	(PSP)+, d0
		and.w	d0, (PSP)
		DO_NEXT
PRIM_WORD (_OR, "OR")			// ( x1 -- x2 )					CORE
		move.w	(PSP)+, d0
		or.w	d0, (PSP)
		DO_NEXT
PRIM_WORD (XOR, "XOR")			// ( x1 -- x2 )					CORE
		move.w	(PSP)+, d0
		eor.w	d0, (PSP)
		DO_NEXT
PRIM_WORD (LSHIFT, "LSHIFT")		// ( x1 u -- x2 )				CORE
		move.w	(PSP)+, d0
		move.w	(PSP), d1
		lsl.w	d0, d1
		move.w	d1, (PSP)
		DO_NEXT
PRIM_WORD (RSHIFT, "RSHIFT")		// ( x1 u -- x2 )				CORE
		move.w	(PSP)+, d0
		move.w	(PSP), d1
		lsr.w	d0, d1
		move.w	d1, (PSP)
		DO_NEXT
PRIM_WORD (TOGGLE, "TOGGLE")		// ( c-addr char -- )				????
		move.w	(PSP)+, d0
		move.w	(PSP)+, d1
		eor.b	d0, 0(BP, d1.w)
		DO_NEXT
		
/**
 * 1.4 Comparison Operators
 */
PRIM_WORD (EQUALS, "=")			// ( n1 n2 -- flag )				CORE
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP)+, d0			// n2 == n1
__DO_EQ__:	seq	d0				
		ext.w	d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (D_EQUALS, "D=")			// ( d1 d2 -- flag )			DOUBLE
		move.l	(PSP)+, d0		
		cmp.l	(PSP)+, d0			
		bra	__DO_EQ__
PRIM_WORD (NOT_EQUALS, "<>")			// ( n1 n2 -- flag )			CORE EXT
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP)+, d0			// n2 == n1
__DO_NE__:	sne	d0				
		ext.w	d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (LESS_THAN, "<")		// ( n1 n2 -- flag )				CORE
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP)+, d0			// n2 > n1
__DO_GT__:	sgt	d0				
		ext.w	d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (D_LESS_THAN, "D<")		// ( d1 d2 -- flag )				DOUBLE
		move.l	(PSP)+, d0	
		cmp.l	(PSP)+, d0		
		bra	__DO_GT__
PRIM_WORD (GREATER_THAN, ">")		// ( n1 n2 -- flag )				CORE
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP)+, d0			// n2 < n1
__DO_LT__:	slt	d0
		ext.w	d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (D_GREATER_THAN, "D>")	// ( d1 d2 -- flag )				DOUBLE
		move.l	(PSP)+, d0
		cmp.l	(PSP)+, d0
		bra	__DO_LT__
PRIM_WORD (U_LESS_THAN, "U<")		// ( u1 u2 -- flag )				CORE
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP)+, d0			// n2 > n1
__DO_HI__:	shi	d0				
		ext.w	d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (DU_LESS_THAN, "DU<")		// ( du1 du2 -- flag )				DOUBLE
		move.l	(PSP)+, d0		
		cmp.l	(PSP)+, d0			
		bra	__DO_HI__
PRIM_WORD (U_GREATER_THAN, "U>")	// ( u1 u2 -- flag )				CORE EXT
		move.w	(PSP)+, d0			// d0 = n2
		cmp.w	(PSP)+, d0			// n2 < n1
		scs	d0
		ext.w	d0
		move.w	d0, -(PSP)
		DO_NEXT
PRIM_WORD (ZERO_EQUALS, "0=")		// ( x -- flag )				CORE
		tst.w	(PSP)+
		bra	__DO_EQ__
PRIM_WORD (ZERO_NOT_EQUALS, "0<>")	// ( x -- flag )				CORE EXT
		tst.w	(PSP)+
		bra	__DO_NE__
PRIM_WORD (D_ZERO_EQUALS, "D0=")	// ( xd -- flag )				CORE
		tst.l	(PSP)+
		bra	__DO_EQ__
PRIM_WORD (ZERO_LESS, "0<")		// ( x -- flag )				CORE
		tst.w	(PSP)+
		bra	__DO_LT__
PRIM_WORD (D_ZERO_LESS, "D0<")		// ( d -- flag )				DOUBLE
		tst.l	(PSP)+
		bra	__DO_LT__
PRIM_WORD (ZERO_GREATER, "0>")		// ( x -- flag )				CORE EXT
		tst.w	(PSP)+
		bra	__DO_GT__
PRIM_WORD (WITHIN, "WITHIN")		// ( n1|u1 n2|u2 n3|u3 -- flag )		CORE EXT
// possible definition:   :  WITHIN  ( test low high -- flag )   OVER - >R - R>  U<  ;
		movem.w	(PSP)+, d1-d3			// d1..d3 = n1|u1..n3|u3
		sub.w	d2, d3				// OVER -
		sub.w	d2, d1				// >R - R>
		cmp.w	d3, d1				// U<
		bra	__DO_HI__
		
/*==================================================================================
       Routines of the Textinterpreter, System- and Dictionaryroutines              **
  ==================================================================================**
   ***********************************************************************************/

/*     Separate the Tokens of the Input Source, Interpretation                    
 *=================================================================================*/

/********************************************************************************
 * SOURCE-ID		-- n					 	CORE	
 */
VALUE (SOURCE_ID, "SOURCE-ID", 0)

/********************************************************************************
 * SOURCE		-- c-addr u				 	CORE	
 *
 * Definition:  
 *    : SOURCE
 *       BLK @ ?DUP  IF BLOCK 1024 EXIT THEN   ( if block: return block's address )
 *       input-buffer 2@ ;                     ( else return input-buffer addr + len )
 */
SEC_WORD (SOURCE, "SOURCE")
	XT(BLK) XT(FETCH) XT(QUESTION_DUP)    IF({ XT(BLOCK) LIT(1024) XT(EXIT) })/*THEN*/
	XT(input_buffer) XT(TWO_FETCH) XT(EXIT)

/********************************************************************************
 * SAVE-INPUT		-- xn ... x1 n			 		CORE EXT
 *
 * Definition: 
 *    : SAVE-INPUT   >IN @  >line 2@  BLK @  4 ;
 */
SEC_WORD (SAVE_INPUT, "SAVE-INPUT")
	XT(TO_IN) XT(FETCH)  XT(to_line) XT(TWO_FETCH)  XT(BLK) XT(FETCH)  LIT(5) XT(EXIT)

/********************************************************************************
 * RESTORE-INPUT			xn ... x1 n --		 	CORE EXT
 *
 * Definition:
 *    : RESTORE-INPUT  
 *       4 <> -257 ?throw                     ( if n is not 4: error )
 *       BLK !  >line 2!  >IN !               ( restore the variables )
 *       ?reload ;                            ( if input comes from a file: reload current line )
 */
SEC_WORD (RESTORE_INPUT, "RESTORE-INPUT")
	XT(FOUR) XT(NOT_EQUALS) LIT(-257) XT(question_throw)
	XT(BLK) XT(STORE)  XT(to_line) XT(TWO_STORE)  XT(TO_IN) XT(STORE)
	XT(question_reload) XT(EXIT)

/*-------------------------------------------------------------------------------
 * ?reload		-- flag			 			internal
 *
 * If input comes from a file, reload the current input line.
 *
 * Definition:   
 *    : ?reload
 *       SOURCE-ID 0> IF                        ( if input comes from a file )
 *          >line 2@ SOURCE-ID REPOSITION-FILE DROP    ( seek to begin of current line )
 *          >IN @ REFILL DROP >IN !                    ( and reload it )
 *       THEN ;
 */
INTERNAL_SEC_WORD (question_reload, "?reload")
	XT(SOURCE_ID) XT(ZERO_GREATER) IF({
	   XT(to_line) XT(TWO_FETCH) XT(SOURCE_ID) XT(REPOSITION_FILE) XT(DROP)
	   XT(TO_IN) XT(FETCH) XT(REFILL) XT(DROP) XT(TO_IN) XT(STORE) 
	})/*THEN*/ XT(EXIT)

/*-------------------------------------------------------------------------------
 * input>r		    -- 				 		internal
 *			R:  -- input-source-specification
 * Save the complete input source specification onto the return stack.
 */
	// size of input source specification on return stack
#define SIZEOF_RS_ISS	14
INTERNAL_PRIM_WORD (input_to_r, "input>r")
		pea	(NP)				// return address is Next
InputToR:	move.l	input_buffer+2(PC), -(RSP)	// .l input-buffer
		move.l	to_line+2(PC), -(RSP)		// .l >line
		move.w	SOURCE_ID+2(PC), -(RSP)		// .w SOURCE-ID
		move.w	TO_IN+2(PC), -(RSP)		// .w >IN
		move.w	BLK+2(PC), -(RSP)		// .w BLK
		rts					// return to next or 'SaveInput' caller

/*-------------------------------------------------------------------------------
 * r>input		    -- 				 		internal
 *			R:  input-source-specification --
 * Restore the complete input source specification from the returnstack, that
 * was put there by input>r.
 */
INTERNAL_PRIM_WORD (r_to_input, "r>input")
		pea	(NP)				// return address is Next
RToInput:	move.w	(RSP)+, RELP(BLK+2)(BP)
		move.w	(RSP)+, RELP(TO_IN+2)(BP)
		move.w	(RSP)+, RELP(SOURCE_ID+2)(BP)
		move.l	(RSP)+, RELP(to_line+2)(BP)
		move.l	(RSP)+, RELP(input_buffer+2)(BP)
		rts

/********************************************************************************
 * PARSE		char "ccc<char>" -- c-addr u			CORE
 *
 * Definition:
 *    : PARSE
 *       SOURCE interpreted 2!	         \ interpreted holds most recently interpreted string
 *       SOURCE >IN @ /STRING            ( S: char input-addr #remaining-chars )
 *       ROT OVER 0 ?DO                  ( S: input-addr #remaining-chars char )
 *          >IN 1+!                                 
 *          2 PICK I + C@ OVER = IF      ( current character equals char ? )
 *             NIP I SWAP LEAVE             ( S: input-addr I char )
 *          THEN
 *       LOOP                            ( S: input-addr #chars char )
 *       DROP  2DUP parsed 2! ;          ( S: input-addr #chars )
 */
SEC_WORD (PARSE, "PARSE")
        XT(SOURCE) XT(interpreted) XT(TWO_STORE)
	XT(SOURCE) XT(TO_IN) XT(FETCH) XT(SLASH_STRING)
	XT(ROT) XT(OVER) XT(ZERO) QDO_LOOP({
	   XT(TO_IN) XT(ONE_PLUS_STORE)
	   XT(TWO) XT(PICK) XT(I) XT(PLUS) XT(C_FETCH) XT(OVER) XT(EQUALS) IF({
	      XT(NIP) XT(I) XT(_SWAP) XT(LEAVE)
	   /*THEN*/})
	})/*LOOP*/
	XT(DROP)   XT(TWO_DUP) XT(parsed) XT(TWO_STORE) XT(EXIT)
	
/********************************************************************************
 * WORD			char "<chars>ccc<char>" -- c-addr 		CORE
 *
 * Definition:
 *    : WORD  (char "<chars>ccc<char>" -- c-addr )
 *       BEGIN                             ( skip leading chars )
 *          DUP  SOURCE >IN @ /STRING      ( S:  char char input-addr not-end-of-input? )
 *          SWAP C@ ROT = AND              ( current character not equals char ? )
 *       WHILE >IN 1+! REPEAT              ( while not-end-of-input and current-char = char )
 *       PARSE 255 MIN                     ( parse at most 255 characters )
 *       DUP HERE C!                       ( store string count at here )
 *       HERE 1+ SWAP MOVE   HERE ;        ( copy string to here+1 and return here )
 */
SEC_WORD (WORD, "WORD")
	BEGIN_WHILE({
	   XT(DUP)  XT(SOURCE) XT(TO_IN) XT(FETCH) XT(SLASH_STRING)
	   XT(_SWAP) XT(C_FETCH) XT(ROT) XT(EQUALS) XT(_AND)
	},{/*WHILE*/  XT(TO_IN) XT(ONE_PLUS_STORE) })/*REPEAT*/
	XT(PARSE) LIT(255) XT(MIN)
	XT(DUP) XT(HERE) XT(C_STORE)
	XT(HERE) XT(ONE_PLUS) XT(_SWAP) XT(_MOVE)   XT(HERE) XT(EXIT)

/*-------------------------------------------------------------------------------
 * word'		char "<chars>ccc<char>" -- c-addr		internal
 *
 * Read a word from the input source, and trigger an error #-16 if there are
 * no more words in it. (if WORD returns empty string)
 *
 * Definition:
 *    : word'   BL WORD  DUP C@ 0=  -16 ?throw ;
 */
INTERNAL_SEC_WORD (word_, "word'")
	XT(BL) XT(WORD)  XT(DUP) XT(C_FETCH) XT(ZERO_EQUALS)  LIT(-16) XT(question_throw) XT(EXIT)

/********************************************************************************
 * '			"<spaces>name" -- xt		 		CORE
 *
 * Definition:
 *    : '   word' FIND 0= -13 ?throw ;
 */
SEC_WORD (TICK, "'")
	XT(word_) XT(FIND)  XT(ZERO_EQUALS) LIT(-13) XT(question_throw) XT(EXIT)

/********************************************************************************
 * [']			Compilation: "<spaces>name" -- 	 		CORE
 *			Run-time:                   -- xt
 * Definition:
 *    : [']   ' [COMPILE] LITERAL ; IMMEDIATE 
 */
SEC_WORD (BRACKET_TICK, "[']", IMMEDIATE)   
	XT(TICK) XT(LITERAL) XT(EXIT)

/********************************************************************************
 * CHAR			"<spaces>name" -- char		 		CORE
 *
 * Definition:
 *    : CHAR   word' 1+ C@ ; 
 */
SEC_WORD (CHAR, "CHAR")
	XT(word_)  XT(ONE_PLUS) XT(C_FETCH) XT(EXIT)

/********************************************************************************
 * [CHAR]		Compilation: "<spaces>name" -- 			CORE 
 *			Run-time:                   -- char
 *
 * Definition:
 *    : [CHAR]   CHAR [COMPILE] LITERAL ; IMMEDIATE
 */
SEC_WORD (BRACKET_CHAR, "[CHAR]", IMMEDIATE)
	XT(CHAR) XT(LITERAL) XT(EXIT)

/********************************************************************************
 * [ELSE]		"<spaces>name" ... -- 				TOOLS EXT
 *
 * Definition:
 *    : [ELSE]   
 *       1 BEGIN                                          ( S: level )
 *          BEGIN   BL WORD COUNT   DUP WHILE             ( R: c-addr u )
 *             2>R
 *             2R@ S" [IF]" COMPARE 0=   IF 1+ THEN       \ [IF] -> increase level
 *             2R@ S" [ELSE]" COMPARE 0=   OVER 1 = AND   \ [ELSE] on base level or
 *             2R> S" [THEN]" COMPARE 0= OR   IF 1- THEN  \ [THEN] -> decrese level
 *	       ?DUP 0= IF EXIT THEN                       \ if level=0: exit
 *          REPEAT   2DROP
 *       REFILL 0= UNTIL                                  \ refill input source
 *       DROP ; IMMEDIATE
 *       
 */
SEC_WORD (BRACKET_ELSE, "[ELSE]", IMMEDIATE)
	XT(ONE) BEGIN_UNTIL({
	   BEGIN_WHILE({   XT(BL) XT(WORD) XT(COUNT)   XT(DUP) /*WHILE*/},{
	      XT(TWO_TO_R)
	      XT(TWO_R_FETCH) SLIT("[IF]") XT(COMPARE) XT(ZERO_EQUALS)   IF({ XT(ONE_PLUS) /*THEN*/})
	      XT(TWO_R_FETCH) SLIT("[ELSE]") XT(COMPARE) XT(ZERO_EQUALS)   XT(OVER) XT(ONE) XT(EQUALS) XT(_AND)
	      XT(TWO_R_FROM) SLIT("[THEN]") XT(COMPARE) XT(ZERO_EQUALS) XT(_OR)   IF({ XT(ONE_MINUS) /*THEN*/})
	      XT(QUESTION_DUP) XT(ZERO_EQUALS) IF({ XT(EXIT) /*THEN*/})
	   /*REPEAT*/})   XT(TWO_DROP)
	 XT(REFILL) XT(ZERO_EQUALS) /*UNTIL*/})
	 XT(DROP) XT(EXIT)
	
/********************************************************************************
 * [IF]			flag | flag "<spaces>name" ... -- 		TOOLS EXT
 *
 * Definition:
 *    : [IF]    0= IF POSTPONE [ELSE] THEN ; IMMEDIATE
 */
SEC_WORD (BRACKET_IF, "[IF]", IMMEDIATE)
	XT(ZERO_EQUALS) IF({ XT(BRACKET_ELSE) /*THEN*/ }) XT(EXIT)

/********************************************************************************
 * [THEN]		-- 						TOOLS EXT
 *
 * Definition:
 *    : [THEN]   ; IMMEDIATE
 */
NOOP_WORD (BRACKET_THEN, "[THEN]", IMMEDIATE)

/********************************************************************************
 * REFILL		-- flag			 			CORE EXT
 *
 * Definition:
 *    : REFILL  
 *        BLK @ IF                        ( if input comes from a block )
 *           BLK 1+!  0 >IN !  TRUE EXIT        ( increase BLK and set >IN to zero )
 *        THEN
 *        SOURCE-ID 0= IF                 ( if input comes from terminal )
 *           tib DUP 84 ACCEPT  SPACE           ( get one input line )
 *           input-buffer 2!   0 >IN !          ( make it the input buffer and set >IN to zero )
 *           TRUE EXIT                          ( return TRUE )
 *        THEN   
 *        SOURCE-ID 0> IF                 ( if input comes from a file )
 *           SOURCE-ID FILE-POSITION DROP >line 2!    ( store file position of beginning of line )
 *           fib DUP 128 SOURCE-ID                    ( read a line into the file input buffer )
 *           READ-LINE DROP                           ( and discard ior )
 *           >R input-buffer 2! R>                    ( set input buffer to line buffer )
 *           0 >IN ! EXIT                             ( set >IN to zero and return EOF flag )
 *        THEN     
 *        FALSE ;                         ( if input comes from EVALUATEd string: return false )
 */
SEC_WORD (REFILL, "REFILL")
	XT(BLK) XT(FETCH) IF({
           XT(BLK) XT(ONE_PLUS_STORE)  XT(ZERO) XT(TO_IN) XT(STORE)  XT(TRUE) XT(EXIT)
        })
	XT(SOURCE_ID) XT(ZERO_EQUALS) IF({  
	   XT(tib) XT(DUP) LIT(TIB_SIZE) XT(ACCEPT)  XT(SPACE)
           XT(input_buffer) XT(TWO_STORE)  XT(ZERO) XT(TO_IN) XT(STORE)  
	   XT(TRUE) XT(EXIT)
        })/*THEN*/ 
	XT(SOURCE_ID) XT(ZERO_GREATER) IF({
	   XT(SOURCE_ID) XT(FILE_POSITION) XT(DROP) XT(to_line) XT(TWO_STORE) 
	   XT(fib) XT(DUP) LIT(FIB_SIZE) XT(SOURCE_ID)
	   XT(READ_LINE) XT(DROP)
	   XT(TO_R) XT(input_buffer) XT(TWO_STORE) XT(R_FROM)
	   XT(ZERO) XT(TO_IN) XT(STORE) XT(EXIT)
	})/*THEN*/
	XT(FALSE) XT(EXIT)


/*-------------------------------------------------------------------------------
 * interpret		-- i*x						internal
 *
 * FORTH text interpreter: 
 *    Read and process words from the input-buffer, until WORD returns an empty string
 * 
 * Definition:
 *    : interpret
 *       BEGIN  BL WORD DUP C@ WHILE        ( process until no more words in buffer )
 *          FIND ?DUP IF                            ( word is a FORTH word? )
 *             STATE @ +                            ( IMMEDIATE or interpretation STATE ? )
 *             IF  EXECUTE  ELSE  COMPILE, THEN     ( execute or compile word )
 *          ELSE                                  ( try to convert word to number )
 *             COUNT number 0= -13 ?throw           ( error -13 if conversion impossible )
 *          THEN
 *       REPEAT DROP ;
 *
 */
INTERNAL_SEC_WORD (interpret, "interpret")	
	BEGIN_WHILE({  XT(BL) XT(WORD) XT(DUP) XT(C_FETCH) },{/*WHILE*/
	   XT(FIND) XT(QUESTION_DUP) IF({
	      XT(STATE) XT(FETCH) XT(PLUS) 
	      IF({   XT(EXECUTE)  },{/*ELSE*/  XT(COMPILE_COMMA) })/*THEN*/
	   },{/*ELSE*/
	      XT(COUNT) XT(number) XT(ZERO_EQUALS) LIT(-13) XT(question_throw)
           })/*THEN*/
	})/*REPEAT*/  XT(DROP) XT(EXIT)

/********************************************************************************
 * QUIT			        --					CORE
 *			R:  i*x --
 * Definition:
 *    : QUIT  
 *       rp!  0 TO SOURCE-ID  0 BLK !        \ clear returnstack, store 0 to SOURCE-ID and BLK
 *       POSTPONE [                          \ switch to interpretation state
 *       BEGIN
 *          REFILL DROP           		     \ refill terminal input-buffer
 *          ['] interpret CATCH                 \ interpret, TOS is exception number (or zero)
 *          ?DUP IF                               \ if there was an exception:
 *             >R sp!                              \ clear data stack (perform function of ABORT)
 *             R> CASE 		                \ process throw code
 *               -1 OF   ENDOF                         \ ABORT   
 *               -2 OF abort"-str COUNT TYPE   ENDOF   \ ABORT"
 *               disp-exc   CR warm                    \ any error: display info and reboot system
 *             ENDCASE
 *          ELSE                                \ on success:
 *             STATE @ 0= IF ." <OK" THEN          \ display OK
 *          THEN   CR
 *       AGAIN ;
 */
SEC_WORD (QUIT, "QUIT")
	XT(rp_store)  XT(ZERO) TO(SOURCE_ID)  XT(ZERO) XT(BLK) XT(STORE)  
	XT(LEFT_BRACKET)
	BEGIN_AGAIN({
	   XT(REFILL) XT(DROP)
	   LIT(RELP(interpret)) XT(CATCH)
	   XT(QUESTION_DUP) IF({
   	      XT(TO_R) XT(sp_store) 
	      XT(R_FROM) CASE({
	         LIT(-1) /*OF*/},{   /*ENDOF*/},{
	         LIT(-2) /*OF*/},{   XT(abort_quote_str) XT(COUNT) XT(TYPE)   /*ENDOF*/},{
	         XT(disp_exc)  XT(CR) XT(warm)
	      /*ENDCASE*/})
	   /*ELSE*/},{
	      XT(STATE) XT(FETCH) XT(ZERO_EQUALS) IF({ STR(25,"OK") })
	   /*THEN*/}) XT(CR) 
	/*AGAIN*/})

/********************************************************************************
 * EVALUATE		i*x c-addr u -- j*x				CORE
 *
 * Definition:
 *    : EVALUATE  
 *       input>r                               ( store input source specification on ret stack ) 
 *       -1 TO SOURCE-ID   0 >IN !   0 BLK !   ( set variables ...)
 *       input-buffer 2! interpret             ( interpret that string )
 *       r>input ?reload ;                     ( restore prior input source specification )
 */
SEC_WORD (EVALUATE, "EVALUATE")
	XT(input_to_r) XT(MINUS_ONE) TO(SOURCE_ID)  
	XT(ZERO) XT(TO_IN) XT(STORE)   XT(ZERO) XT(BLK) XT(STORE)
	XT(input_buffer) XT(TWO_STORE) XT(interpret)
	XT(r_to_input) XT(question_reload) XT(EXIT)

/********************************************************************************
 * LOAD		   	i*x u -- j*x					BLOCK
 *
 * Definition:
 *    : LOAD
 *       input>r                               ( store input source specification on ret stack )
 *       BLK !   0 >IN !   interpret           ( store zero in >IN and interpret that block )
 *       r>input ?reload ;                     ( restore prior input source specification )
 */
SEC_WORD (LOAD, "LOAD")
	XT(input_to_r)
	XT(BLK) XT(STORE)   XT(ZERO) XT(TO_IN) XT(STORE)  XT(interpret)
	XT(r_to_input) XT(question_reload) XT(EXIT)

/********************************************************************************
 * THRU		   	i*x u1 u2 -- j*x				BLOCK EXT
 *
 * Definition:
 *    : THRU   1+ SWAP ?DO I LOAD LOOP ;
 */
SEC_WORD (THRU, "THRU")
 	XT(ONE_PLUS) XT(_SWAP) QDO_LOOP({ XT(I) XT(LOAD) })/*LOOP*/ XT(EXIT)

/********************************************************************************
 * INCLUDE-FILE	   	i*x fileid -- j*x				FILE
 *
 * Definition:
 *    : INCLUDE-FILE
 *       input>r                               ( save input source specification onto ret stack )
 *       TO SOURCE-ID  0 BLK !                 ( store fileid to SOURCE-ID, set BLK to 0 )
 *       BEGIN REFILL WHILE interpret REPEAT   ( read a line, interpret ... )
 *       r>input ?reload ;                     ( restore prior input source specification )
 */
SEC_WORD (INCLUDE_FILE, "INCLUDE-FILE")
	XT(input_to_r)
	TO(SOURCE_ID)  XT(ZERO) XT(BLK) XT(STORE)
	BEGIN_WHILE({ XT(REFILL) },{/*WHILE*/ XT(interpret) })/*REPEAT*/
	XT(r_to_input) XT(question_reload) XT(EXIT)

/********************************************************************************
 * INCLUDED	   	i*x c-addr u -- j*x				FILE
 *
 * Definition:
 *    : INCLUDED
 *       R/O OPEN-FILE -37 ?throw  
 *       DUP >R INCLUDE-FILE  R> CLOSE-FILE DROP ;
 */
SEC_WORD (INCLUDED, "INCLUDED")
	XT(R_O) XT(OPEN_FILE) LIT(-37) XT(question_throw) 
	XT(DUP) XT(TO_R) XT(INCLUDE_FILE)  XT(R_FROM) XT(CLOSE_FILE) XT(DROP) XT(EXIT) 

/*     System-Directives
 *=================================================================================*/

/*-------------------------------------------------------------------------------
 * warm		 	      --					internal
 *
 * "warm" system reboot. Initialize some vars to default
 *
 * Definition:   : warm    ONLY DEFINITIONS   DECIMAL   QUIT ;
 */
INTERNAL_SEC_WORD (warm, "warm")
	XT(ONLY) XT(DEFINITIONS)   XT(DECIMAL)   XT(QUIT)

/*-------------------------------------------------------------------------------
 * hello	 	      --					internal
 *
 * Start FORTH: display an info and make warm system reboot
 *
 * Definition:
 *    : hello   
 *       CR 12 SPACES UL ." FORTH92 v0.1.00{CR}{TAB} by David Khling 1998/99{CR}{CR}" UL
 *       warm
 */
INTERNAL_SEC_WORD (hello, "hello")
	XT(CR) LIT(12) XT(SPACES) XT(UL)
	STR("FORTH92  v0.1.00",CHR_LF,CHR_TAB,"by David Khling 1998/99",CHR_LF,CHR_LF)  XT(UL)
	XT(warm)

/********************************************************************************
 * BYE		   							TOOLS EXT
 *
 * Flush block buffers and exit FORTH
 */
SEC_WORD (BYE, "BYE")
		XT(FLUSH)
		XT(\bye)
\bye:		dc.w	RELP(ExitFORTH)

/********************************************************************************
 * COLD			   i*x --					????
 *                      R: j*x --
 * Cold system reboot: set the dictionary to standard values, thus removing all
 * user-defined words. Execute hello.
 */
PRIM_WORD (COLD, "COLD")
			// initialize wordlist-table to 1 wordlist: FORTH
		move.l	#$10000+RELP(__HERE), RELP(WordListTbl)(BP) 
			// initialize dictionary pointer (HERE) to first byte of user memory
		move.w	#RELP(UserDict), RELP(HERE+2)(BP)
			// reinitialize all registers and "reboot" FORTH
		bra	BootFORTH
		
/********************************************************************************
 * ABORT		   i*x --					????
 *                      R: j*x --
 * Definition:
 *    : ABORT   -1 THROW ;  
 */
SEC_WORD (ABORT, "ABORT")
	XT(MINUS_ONE) XT(THROW)

/*-------------------------------------------------------------------------------
 * abort"		   i*x x1 -- j*x |				internal
 *			R:    k*x -- k*x |
 * Compiled by ABORT". Set abort"-str to the literal string following abort"
 * and trigger THROW -2.
 *
 * Definition:
 *    : abort" 
 *       2R> A>R TUCK TO abort"-str       \ addr of string is on stack: store it to abort"-str
 *       -2 ?throw                        \ trigger THROW -2 if x1 is true
 *       COUNT bra __Skip_SLITERAL__ ;    \ else continue execution after string, same as .")
 */
INTERNAL_SEC_WORD (abort_quote, 'abort"')
	XT(TWO_R_FROM) XT(A_TO_R) XT(TUCK) TO(abort_quote_str) 
	XT(MINUS_TWO) XT(question_throw)
	XT(COUNT) XT(_bra) XT(__Skip_SLITERAL)

/********************************************************************************
 * ABORT"		Compilation: "ccc<quote>"			CORE
 * 			Execution:     i*x x1 -- j*x |		
 *		     	            R:    k*x -- k*x |
 * Definition:
 *    : ABORT" 
 *       compile abort"  bra __Quote_SLITERAL  IMMEDIATE   ( compile abort", ... )
 */
SEC_WORD (ABORT_QUOTE, 'ABORT"', IMMEDIATE)
	XT(compile) XT(abort_quote)  XT(_bra) XT(__Quote_SLITERAL)

/********************************************************************************
 * EXECUTE	   	i*x xt -- j*x					CORE
 */
PRIM_WORD (EXECUTE,"EXECUTE")
		move.w	(PSP)+, WP
		move.w	0(BP, WP.w), d0
		jmp	0(BP, d0.w)

/********************************************************************************
 * UNUSED   		-- u						CORE EXT
 *
 * Definition:   : UNUSED   xxx HERE - ;
 */
SEC_WORD (UNUSED, "UNUSED")
	LIT(RELP(DictEnd)) XT(HERE) XT(MINUS) XT(EXIT)

/********************************************************************************
 * (	   		"ccc<paren>" -- 				CORE
 *
 * Definition:
 *    : (   
 *       BEGIN  
 *          [CHAR] )  DUP PARSE             ( skip till r-bracket or end of line )
 *          + C@ <>  SOURCE-ID 0> AND       ( if r-bracket isn't found in file-input-mode )
 *       WHILE 
 *          REFILL 0= IF EXIT THEN          ( read all following lines until it is found )
 *       REPEAT ;
 *       
 */
SEC_WORD (PAREN, "(", IMMEDIATE)
	BEGIN_WHILE({
	   LIT(')') XT(DUP) XT(PARSE)
	   XT(PLUS) XT(C_FETCH) XT(NOT_EQUALS)  XT(SOURCE_ID) XT(ZERO_GREATER) XT(_AND) 
	},{/*WHILE*/
	   XT(REFILL) XT(ZERO_EQUALS) IF({ XT(EXIT) })/*THEN*/
	})/*REPEAT*/ XT(EXIT)

/********************************************************************************
 * \	   		"ccc<eol>" --					CORE EXT
 *
 * Definition:
 *    : \   
 *       BLK @ IF                                  ( if input-source is a block )
 *          C/L >IN @ C/L MOD -   >IN +!   EXIT        ( go to begin of next line )
 *       THEN 
 *       0 PARSE 2DROP ;                           ( else: discard remainder of parse area )
 */
SEC_WORD (BACKSLASH, "\", IMMEDIATE)
	XT(BLK) XT(FETCH) IF({
	   XT(C_PER_L)   XT(TO_IN) XT(FETCH) XT(C_PER_L) XT(MOD) XT(MINUS)   XT(TO_IN) XT(PLUS_STORE) XT(EXIT)
	}) /*THEN*/
	XT(ZERO) XT(PARSE) XT(TWO_DROP) XT(EXIT)

/*     Operations within the dictionary
 *=================================================================================*/

/********************************************************************************
 * GET-CURRENT 		-- wid						SEARCH
 */
PRIM_WORD (GET_CURRENT, "GET-CURRENT")
		move.w	RELP(Current)(BP), -(PSP)
		jmp	(NP)

/********************************************************************************
 * SET-CURRENT 		wid --						SEARCH
 */
PRIM_WORD (SET_CURRENT, "SET-CURRENT")
		move.w	(PSP)+, RELP(Current)(BP)
		jmp	(NP)

/********************************************************************************
 * DEFINITIONS		--						SEARCH
 */
PRIM_WORD (DEFINITIONS, "DEFINITIONS")
		move.w	RELP(SearchOrder+2)(BP), RELP(Current)(BP)
		jmp 	(NP)

/********************************************************************************
 * GET-ORDER		-- widn ... wid1 n				SEARCH
 */
PRIM_WORD (GET_ORDER, "GET-ORDER")
		lea	RELP(SearchOrder)(BP), a0	// a0 points to SearchOrder table
		move.w	(a0), d0			// d0 = number of word lists
		move.w	d0, d1
		lsl.w	#1, d1
		lea	2(a0, d1.w), a0			// a0 points to end of SearchOrder table
\Loop:		move.w	-(a0), -(PSP)			// push the wids, and finally n
		dbra	d0, \Loop
		jmp	(NP)				

/********************************************************************************
 * SET-ORDER		widn ... wid1 n --				SEARCH
 */
PRIM_WORD (SET_ORDER, "SET-ORDER")
		lea	RELP(SearchOrder)(BP), a0	// a0 points to SearchOrder table
		move.w	(PSP), d0			// d0 = number of word lists in search order
		bmi	\Minimum			//   if it is negative: -> minimum search order
		cmpi.w	#MAX_ORDER, d0                  //   if it is greater than MAX_ORDER: -> overflow
		bgt	SOrderOverflow
\Loop:		move.w	(PSP)+, (a0)+			// copy n, and wids to that table
		dbra	d0, \Loop
		jmp	(NP)
\Minimum:	move.l	#$10000+RELP(FORTH_Wordlist), (a0)
		addq.l	#2, PSP
		jmp	(NP)

SOrderOverflow:	move.w	#ERR_SO_OVERFLOW, d0
		bra	DoThrow

/********************************************************************************
 * WORDLIST		-- wid						SEARCH
 */
PRIM_WORD (WORDLIST, "WORDLIST")
		lea	WordListTbl(PC), a0
		move.w	(a0), d0
		cmpi.w	#MAX_WORDLISTS, d0		//   if allready the maximum amount of wordlists:
		bge	SOrderOverflow			//      -> overflow
		addq.w	#1, (a0)			// increase number of word lists
		lsl.w	#1, d0				// d0 = relative address of first free cell in table
		add.w	#RELP(WordListTbl)+2, d0
		clr.w	0(BP, d0.w)			// clear that new cell
		move.w	d0, -(PSP)			// push this address (it is the wid)
		jmp	(NP)

/********************************************************************************
 * FORTH-WORDLIST	-- wid						SEARCH
 */
CONSTANT (FORTH_WORDLIST, "FORTH-WORDLIST", RELP(FORTH_Wordlist))

/********************************************************************************
 * FORTH		-- 						SEARCH EXT
 */
PRIM_WORD (FORTH, "FORTH")
		move.w	#RELP(FORTH_Wordlist), RELP(SearchOrder+2)(BP)
		jmp	(NP)

/********************************************************************************
 * ONLY			-- 						SEARCH EXT
 *
 * Definition:   : ONLY   -1 SET-ORDER ;
 */
SEC_WORD (ONLY, "ONLY")
	XT(MINUS_ONE) XT(SET_ORDER) XT(EXIT)

/********************************************************************************
 * ALSO			--						SEARCH EXT
 *
 * Definition:   : ALSO   GET-ORDER   OVER SWAP 1+   SET-ORDER ;
 */
SEC_WORD (ALSO, "ALSO")
	XT(GET_ORDER)  XT(OVER) XT(_SWAP) XT(ONE_PLUS) XT(SET_ORDER) XT(EXIT)

/********************************************************************************
 * PREVIOUS		--						SEARCH EXT
 *
 * Definition:   : PREVIOUS   GET-ORDER   DUP 0= -50 ?throw   NIP 1- SET-ORDER ;
 */
SEC_WORD (PREVIOUS, "PREVIOUS")	
	XT(GET_ORDER)  XT(DUP) XT(ZERO_EQUALS) LIT(-50) XT(question_throw)  XT(NIP) XT(ONE_MINUS) XT(SET_ORDER) XT(EXIT)

/********************************************************************************
 * VOCABULARY		"name<space>" --				????
 *
 * Definition:   
 *   : VOCABULARY 
 *      CREATE WORDLIST ,                ( create a wordlist, store its wid into current definition )
 *      DOES>  @ >R                          ( Execution: load wid to top of return stack )
 *       GET-ORDER NIP R> SWAP SET-ORDER ;   ( replace first wordlist of the searchorder with it )
 */
SEC_WORD (VOCABULARY, "VOCABULARY")
	XT(CREATE) XT(WORDLIST) XT(COMMA)
	XT(semi_code) bsr DoDoes/*DOES>*/  XT(FETCH) XT(TO_R) 
	XT(GET_ORDER) XT(NIP) XT(R_FROM) XT(_SWAP) XT(SET_ORDER) XT(EXIT)

/********************************************************************************
 * MARKER		"name<space>" --				CORE EXT
 *
 * Definition:
 *    : MARKER
 *          HERE                              ( S: HERE )
 *          dict-data HERE #dictdata MOVE     ( copy dictionary-data to HERE )
 *          #dict-data ALLOT                  ( and allocate suficient memory )
 *          CREATE ,                          ( save HERE, the address where the data lie )
 *       DOES>                             ( when the marker is executed : )
 *          @ TO HERE                         ( restore HERE )
 *          HERE dict-data #dict-data MOVE ;  ( restore the dictionary-data )
 */
SEC_WORD (MARKER, "MARKER")
	   XT(HERE)
	   XT(dict_data) XT(HERE) XT(num_dict_data) XT(_MOVE)
	   XT(num_dict_data) XT(ALLOT)
	   XT(CREATE) XT(COMMA)
	/*DOES>*/XT(semi_code) bsr DoDoes
	   XT(FETCH)  TO(HERE)
	   XT(HERE) XT(dict_data) XT(num_dict_data) XT(_MOVE) XT(EXIT)

/********************************************************************************
 * >BODY		xt -- a-addr					CORE
 */
SAME_PRIM_WORD (TO_BODY, ">BODY", TWO_PLUS)

/*-------------------------------------------------------------------------------
 * word		c-addr1 -- c-addr2					internal
 *
 * Print a word, the pointer to its bit field given in TOS (c-addr1).
 * If the word is  hidden, it will be displayed underlined. 
 * Return pointer to the next words bit field.
 * Note that a space will precede the word.
 *
 * Definition:
 *    : word 
 *       SPACE
 *       DUP COUNT DUP 31 AND     ( read word's bitfield and extract length )
 *       SWAP 64 AND IF           ( if smudge bit is set )
 *          UL TYPE UL                ( type the word unerlined )
 *       ELSE   TYPE THEN         ( else type the word normal )
 *       2- @ ;                   ( read link-field -- get address of next word )
 */
INTERNAL_SEC_WORD (word, "word")
	XT(SPACE)
	XT(DUP) XT(COUNT) XT(DUP) LIT(31) XT(_AND)
	XT(_SWAP) LIT(SMUDGE) XT(_AND) IF({
	   XT(UL) XT(TYPE) XT(UL)
	},{/*ELSE*/ XT(TYPE) })/*THEN*/
	XT(TWO_MINUS) XT(FETCH) XT(EXIT) 
	
/********************************************************************************
 * ORDER		--						SEARCH EXT
 *
 * Definition:
 *    : ORDER
 *       GET-ORDER 1+ GET-CURRENT SWAP 0 DO    ( for all wordlists and the compilation wordlist )
 *          CR   @ word word word DROP ." ..."    ( print the 3 last words )
 *       LOOP ;
 */
SEC_WORD (ORDER, "ORDER")
	XT(GET_ORDER) XT(ONE_PLUS) XT(GET_CURRENT) XT(_SWAP) XT(ZERO) QDO_LOOP({
	   XT(CR)   XT(FETCH) XT(word) XT(word) XT(word) XT(DROP) STR("...")
	})/*LOOP*/ XT(EXIT)

/********************************************************************************
 * WORDS		--						TOOLS
 *
 * Definition:
 *    : WORDS  ( -- )
 *       0 GET-ORDER DROP @       ( S: 0 widn ... wid2  pointer-to-last-word )
 *       1 TO out                 ( set out, the number of printed characters, to 1 )
 *       BEGIN 
 *          out 586 > IF                   ( if screen is full )
 *             EKEY DROP   0 TO out        ( wait for keypress, set out to zero )
 *          THEN   word                    ( display a word )
 *       ?DUP 0= UNTIL ;                   ( loop until pointer-to-word is NULL )
 *       BEGIN 0= UNTIL ;         ( remove 0 widn ... wid2 )
 */
SEC_WORD (WORDS, "WORDS")			// ( -- )				TOOLS
	XT(ZERO) XT(GET_ORDER) XT(DROP) XT(FETCH)
	XT(ONE) TO(out)
	BEGIN_UNTIL({
	   XT(out) LIT(586) XT(GREATER_THAN) IF({
	      XT(EKEY) XT(DROP)   XT(ZERO) TO(out)
	   })/*THEN*/   XT(word)
	XT(QUESTION_DUP) XT(ZERO_EQUALS) })/*UNTIL*/
	BEGIN_UNTIL({ XT(ZERO_EQUALS) })/*UNTIL*/ XT(EXIT)

/********************************************************************************
 * SEARCH-WORDLIST	c-addr u wid -- xt -1 | xt 1 | 0		SEARCH
 *
 * (possible) definition: (ASM is 10 times faster and a little smaller)
 *    : SEARCH-WORDLIST
 *       @                    ( wid points to the pointer that points to the latest entry's BitField )
 *       BEGIN ?DUP WHILE     ( search as long as the pointer is not a NULL pointer )
 *          >R R@ COUNT  31 AND       ( load entry's BitField, get length (look forth92m.s)
 *          2OVER COMPARE             ( if the two names are equal )
 *          R@ C@ 64 AND  OR 0= IF    ( and the smudge bit in the BitField is not set )
 *             2DROP                            ( remove c-addr u )
 *             R@ COUNT  31 AND +  ALIGNED      ( get BitField, calculate address of code field )
 *             -1  R> C@ 32 AND  IF NEGATE THEN ( push 1 if immediat bit in BitField is set  -1, else)
 *             EXIT
 *          THEN
 *       R> 2-  @ REPEAT      ( go to next word (read next address from link field)
 *       2DROP 0 ;            ( when loop exits: no success, return 0 )
 */
PRIM_WORD (SEARCH_WORDLIST, "SEARCH-WORDLIST")
		move.w	(PSP)+, d0			// d0 = wid = ''bf
		move.w	(PSP)+, d3			// d3 = string-len

		addq.w	#2, d0				// d0 = ''bf+2  (link pointer also points to ''bf+2)
\Loop:		move.w	-2(BP,d0.w), d0			   // d0 = 'bf   (read contents of link-field)
		beq	\End				   // if it is zero: end of linked list -> return 0
		lea	0(BP,d0.w), a1			   // a1 = 'bf   (calculate absolute address)
		move.b	(a1)+, d1			   // d1 =  contents of 'bf (bitfield)
		btst.b	#SMUDGE_BIT, d1			   // if smudge bit is set: 
		bne	\__Skip				      // skip this word
		moveq	#1, d2				   // d2 = return-flag (1 if immediate, -1 else)
		btst.b	#IMMEDIATE_BIT, d1		
		bne	\__Immediate
		moveq	#-1, d2
\__Immediate:	
		/******* compare string with word's name */
		andi.w	#$001F, d1			   // d1 = length of word's name
		cmp.w	d1, d3				   // if string-len <> word-name-len
		bne	\__Skip				      // skip this word
		subq.w	#1, d1				   // d1 = counter   (for name comparison)
		movea.w	(PSP), a0			   // a0 = 'string
		adda.l	BP, a0
\__CmpLoop:	cmpm.b	(a1)+, (a0)+			     // compare two characters of string and name
		bne	\__Skip				     // if they aren't equal: skip this word
		dbra	d1, \__CmpLoop			   // loop until end of string is reached
		/****** word has been found */		
		add.w	d3, d0				   // calculate address of code field ( = xt)
		addq.w	#2, d0				
		bclr.b	#0, d0				   // has to be an aligned address
		move.w	d0, (PSP)			   // push xt 
		move.w	d2, -(PSP)			   // push flag
		jmp	(NP)				   // exit

\__Skip:	bra	\Loop				// loop back -- d0 = 'bf = 'link-field+2

\End:		clr.w	(PSP)				// push 0 --> word not found!
		jmp	(NP)				// exit

/********************************************************************************
 * FIND			c-addr -- xt -1 | xt 1 | c-addr 0		CORE
 *
 * Definition:
 *    : FIND   
 *       0 GET-ORDER 1+ ROLL                    ( S: widn ... wid0 c-addr )
 *       BEGIN  TUCK COUNT ROT ?DUP  WHILE         ( while the current wid is not zero )
 *          SEARCH-WORDLIST ?DUP IF                    ( search the word in the current wordlist )
 *             BEGIN ROT 0= UNTIL  EXIT                ( if it was found: clear stack, return  xt n )
 *          THEN
 *       REPEAT  2DROP 0 ;                         ( word not found: clear stack, return c-addr 0 )
 */
SEC_WORD (FIND, "FIND")				// ( c-addr -- xt n | c-addr 0 )
	XT(ZERO) XT(GET_ORDER) XT(ONE_PLUS) XT(ROLL) 
	BEGIN_WHILE({ XT(TUCK) XT(COUNT) XT(ROT) XT(QUESTION_DUP) /*WHILE*/},{
           XT(SEARCH_WORDLIST) XT(QUESTION_DUP) IF({
	      BEGIN_UNTIL({ XT(ROT) XT(ZERO_EQUALS) })/*UNTIL*/ XT(EXIT)
	   })*THEN*/
	})/*REPEAT*/ XT(TWO_DROP) XT(ZERO) XT(EXIT)


/*-------------------------------------------------------------------------------
 * find'		    "<spaces>name" i*x --  | i*x xt -1|1 	CORE
 *			 R:                j*x --  | j*x
 * Search the next word of the input source in the dictionary. Trigger error
 * -13 if it is not found. Uses word' to read a word.
 *
 * Definition:   : find'   word' FIND DUP 0=  -13 ?throw ;
 */
INTERNAL_SEC_WORD (find_, "find'")
	XT(word_) XT(FIND) XT(DUP) XT(ZERO_EQUALS) LIT(-13) XT(question_throw) XT(EXIT)

/*==================================================================================
       Codegeneration routines
  ==================================================================================**
   ***********************************************************************************/

/*     Basic routines
 *=================================================================================*/

/********************************************************************************
 * ,			x --						CORE
 *
 * Definition:   : ,   HERE 2 ALLOT ! ;
 */
SEC_WORD (COMMA, ",")
	XT(HERE) XT(TWO) XT(ALLOT) XT(STORE) XT(EXIT)

/********************************************************************************
 * COMPILE,		xt --						CORE EXT
 *
 * Definition:   : COMPILE,   , ;
 */
SEC_WORD (COMPILE_COMMA, "COMPILE,")
	XT(COMMA) XT(EXIT)

/********************************************************************************
 * C,			char --						CORE
 *
 * Definition:   : C,   HERE 1 ALLOT C! ;
 */
SEC_WORD (C_COMMA, "C,")		
	XT(HERE) XT(ONE) XT(ALLOT) XT(C_STORE) XT(EXIT)

/********************************************************************************
 * ALIGN		--						CORE
 *
 * Definition:   : ALIGN   HERE ALIGN TO HERE ;
 */
SEC_WORD (ALIGN, "ALIGN")		
	XT(HERE) XT(ALIGNED) TO(HERE) XT(EXIT)

/********************************************************************************
 * ALLOT		n --						CORE
 *
 * Definition:   : ALLOT   HERE +   DUP xxx U> -8 ?throw  TO HERE ;
 */
SEC_WORD (ALLOT, "ALLOT")		
	XT(HERE) XT(PLUS)  XT(DUP) LIT(RELP(DictEnd-1)) XT(U_GREATER_THAN) LIT(-8) XT(question_throw)
	TO(HERE) XT(EXIT)

/*-------------------------------------------------------------------------------
 * lit			-- x						internal
 *
 * Is compiled by LITERAL. Push the cell pointed to by the current IP onto the
 * data stack. Continue execution after that cell.
 */
INTERNAL_PRIM_WORD (lit, "lit")	
		move.w	(IP)+, -(PSP)
		DO_NEXT

/*-------------------------------------------------------------------------------
 * 2lit			-- dx						internal
 *
 * Is compiled by 2LITERAL. Push the cell-pair pointed to by the current IP onto 
 * the data stack. Continue execution after these cells.
 */
INTERNAL_PRIM_WORD (two_lit, "2lit")
		move.l	(IP)+, -(PSP)
		DO_NEXT

/********************************************************************************
 * POSTPONE		"<spaces>name" --				CORE
 *
 * Definition:
 *    : POSTPONE   'find 0< IF comp comp THEN   , ; IMMEDIATE
 */
SEC_WORD (POSTPONE, "POSTPONE", IMMEDIATE)
	XT(find_) XT(ZERO_LESS) IF({  XT(compile) XT(compile) })/*THEN*/ XT(COMMA) XT(EXIT)

/*-------------------------------------------------------------------------------
 * cf			-- a-addr					internal
 *
 * Returns the pointer to the code field of the most recently defined word.
 *
 * Definition:   0 VALUE cf
 */
INTERNAL_VALUE (cf, "cf", 0)	

/********************************************************************************
 * CREATE		"<spaces>name" --				CORE
 *
 * Definition:
 *    : CREATE
 *       GET-CURRENT @ ,                    ( create linkfield )
 *       word' C@   DUP 31 > -19 ?throw     ( read word, error -19 if name too long S: name-len )
 *       HERE GET-CURRENT !                 ( link word to list )
 *       HERE 128 TOGGLE                    ( set bit 7 in bitfield of word )
 *       1+ ALLOT ALIGN  HERE 0 , TO cf     ( allocate code field and set cf to point to it)
 *       ;CODE ...                          ( append assembly runtime semantics to current definition...)
 */
SEC_WORD (CREATE, "CREATE")
	XT(GET_CURRENT) XT(FETCH) XT(COMMA)  
	XT(word_)  XT(C_FETCH) XT(DUP) LIT(31) XT(GREATER_THAN) LIT(-19) XT(question_throw)
	XT(HERE) XT(GET_CURRENT) XT(STORE)
	XT(HERE) LIT(128) XT(TOGGLE)
	XT(ONE_PLUS) XT(ALLOT) XT(ALIGN)  XT(HERE) XT(ZERO) XT(COMMA) TO(cf)
	XT(semi_code)
		// ;CODE ...
DoVar:		addq.w	#2, WP				// push address of body
		move.w	WP, -(PSP)
		NEXT

/*-------------------------------------------------------------------------------
 * ;code		            --					internal
 *              	R: nest-sys --
 * Is compiled by ;CODE and DOES>. Append assembly run time semantics that
 * follow ;code to the current definition. Return to the calling procedure of
 * the procedure that called ;code.
 *
 * Definition:   ;code   2R> A>R cf ! ;
 */
INTERNAL_SEC_WORD (semi_code, ";code")
	XT(TWO_R_FROM) XT(A_TO_R)  XT(cf) XT(STORE) XT(EXIT)
		

/*     Definition Words
 *=================================================================================*/

/*
 * Summary of syntax checking stack values:
 *==========================================
 *  orig:	a-addr -4
 *  dest:	a-addr -2
 *  do-sys: 	orig dest -1 
 *  case-sys:	0 orig1 ... origN 0
 *  of-sys:	orig 1
 *  colon-sys:	2
 *  code-sys:   4
 */

/********************************************************************************
 * :			"<spaces>name" -- colon-sys			CORE
 *
 * Definition:
 *    : :  
 *       CREATE smudge         ( create new word, hide it )
 *       POSTPONE ]  2         ( enter compilation state, push colon-sys
 *       ;CODE ...             ( add DoCall execution semantics ... )
 */
SEC_WORD (COLON, ":")
		XT(CREATE) XT(smudge)
__DoDefinition:	XT(RIGHT_BRACKET)  XT(TWO)  
		XT(semi_code)
		// ;CODE ...
DoCall:		cmpa.l	RST, RSP		// check for return-stack overflow
		ble	\RSOverflow
		move.l	IP, -(RSP)		// push return-address
		lea	2(BP,WP.w), IP
		NEXT
\RSOverflow:	moveq	#ERR_RS_OVERFLOW, d0    
		bra	DoThrow

/********************************************************************************
 * :NONAME		"<spaces>name" -- xt colon-sys			CORE EXT
 *
 * Definition:
 *    : :NONAME  
 *       HERE 0 , DUP TO cf   ( create a codefield, and push it's address (this is xt)
 *       bra __DoDefinition   ( everyting else is similar to ":" )
 */
SEC_WORD (COLON_NONAME, ":NONAME")
	XT(HERE) XT(ZERO) XT(COMMA) XT(DUP) TO(cf)
	XT(_bra) LBL(__DoDefinition)

/********************************************************************************
 * ;			colon-sys --					CORE
 *
 * Definition:
 *    : ;  
 *       2 ?pairs  comp EXIT             ( check colon-sys, compile EXIT to end of current def. )
 *       smudge POSTPONE [ ; IMMEDIATE   ( unhide word, enter interpretation state )
 */
SEC_WORD (SEMICOLON, ";", IMMEDIATE)
	XT(TWO) XT(question_pairs)  XT(compile) XT(EXIT)  
	XT(smudge) XT(LEFT_BRACKET) XT(EXIT)

/********************************************************************************
 * CODE			"<spaces>name" -- code-sys			CORE
 *
 * Definition:
 *    : CODE   CREATE smudge   HERE cf !   4 ;
 */
SEC_WORD (CODE, "CODE")
	XT(CREATE) XT(smudge) XT(HERE) XT(cf) XT(STORE)  XT(FOUR) XT(EXIT)

/********************************************************************************
 * END-CODE		code-sys --					????
 *
 * Definition:
 *    : END-CODE   4 ?pairs   smudge ; IMMEDIATE
 */
SEC_WORD (END_CODE, "END-CODE", IMMEDIATE)
	XT(FOUR) XT(question_pairs)  XT(smudge) XT(EXIT)

/********************************************************************************
 * VARIABLE		"<spaces>name" --				CORE
 *
 * Definition:   : VARIABLE   CREATE 0 , ;
 */
SEC_WORD (_VARIABLE, "VARIABLE")	
	XT(CREATE)  XT(ZERO) XT(COMMA) XT(EXIT)	

/********************************************************************************
 * CONSTANT		x "<spaces>name" --				CORE
 *
 * Definition:   : CONSTANT   CREATE , ;CODE ...
 */
SEC_WORD (_CONSTANT, "CONSTANT")
	XT(CREATE)  XT(COMMA)  XT(semi_code)
		// ;CODE ...
DoCon:		move.w	2(BP,WP.w), -(PSP)	// push contents of constant
		NEXT

/********************************************************************************
 * 2VARIABLE		"<spaces>name" --				DOUBLE
 *
 * Definition:   : 2VARIABLE   VARIABLE 0 , ;
 */
SEC_WORD (_2VARIABLE, "2VARIABLE")
	XT(_VARIABLE) XT(ZERO) XT(COMMA) XT(EXIT)

/********************************************************************************
 * 2CONSTANT		dx "<spaces>name" --				DOUBLE
 *
 * Definition:   : 2CONSTANT   CREATE , , ;CODE ...
 */
SEC_WORD (_2CONSTANT, "2CONSTANT")
	XT(CREATE)  XT(COMMA) XT(COMMA) XT(semi_code)
Do2Con:		move.l	2(BP,WP.w), -(PSP)		// push contents of constant
		DO_NEXT

/********************************************************************************
 * VALUE		x "<spaces>name" --				CORE EXT
 *
 * Definition:   : VALUE   CONSTANT compile DoTo ;
 */
SEC_WORD (_VALUE, "VALUE")
	XT(_CONSTANT) XT(compile) XT(DoTo) XT(EXIT)
DoTo:		move.w	(PSP)+, -2(BP,WP.w)		// set value to TOS
		NEXT

/********************************************************************************
 * TO			Compiling:      "<spaces>name" --		CORE EXT
 *                      Interpreting: x "<spaces>name" --
 *			Execution:                   x --
 * Definition:
 *    : TO
 *       ' 4 +   DUP @ xxx <> -32 ?throw     ( get address of 2nd codefield for TO, check whether realy a VALUE )
 *       STATE @ IF  , EXIT THEN  EXECUTE ;  ( compile or execute, depending on mode )
 *       IMMEDIATE
 */
SEC_WORD (_TO, "TO", IMMEDIATE)	
	XT(TICK) XT(FOUR) XT(PLUS)  XT(DUP) XT(FETCH) LIT(RELP(DoTo)) XT(NOT_EQUALS) LIT(-32) XT(question_throw)
	XT(STATE) XT(FETCH) IF({  XT(COMMA) XT(EXIT) })/*THEN*/  XT(EXECUTE) XT(EXIT) 

/*     Meta Definition Words
 *=================================================================================*/

/*
 * DOES> - handler, FORTH83 - like
 */
DoDoes:		move.l	IP, -(RSP)	// push IP (do call)
		move.l	(a7)+, IP	// address of DOES>-commands is on stack (due to BSR)
		addq.w	#2, WP		// push address of body
		move.w	WP, -(PSP)
		DO_NEXT

/********************************************************************************
 * ;CODE		colon-sys -- code-sys				TOOLS EXT
 *
 * Definition:
 *    : ;CODE
 *       2 ?pairs  compile ;code   [COMPILE] [   4 ;
 */
SEC_WORD (SEMI_CODE, ";CODE", IMMEDIATE)
	XT(TWO) XT(question_pairs)  XT(compile) XT(semi_code) XT(LEFT_BRACKET)  XT(FOUR) XT(EXIT)

/********************************************************************************
 * DOES>		colon-sys -- colon-sys				CORE
 * 
 * Definition:
 *    : DOES>  
 *        DUP 2 ?pairs                                         ( check, whether colon-sys )
 *        compile ;code   compile ASM-BSR ['] DoDoes HERE - ,  ( Execution: compile a call to DoDoes )
 *        ; IMMEDIATE
 */
SEC_WORD (DOES, "DOES>", IMMEDIATE)
	XT(DUP) XT(TWO) XT(question_pairs) 
        XT(compile) XT(semi_code) XT(compile) dc.w %0110000100000000  /* BSR */
	LIT(RELP(DoDoes))  XT(HERE) XT(MINUS) XT(COMMA)
	XT(EXIT)


/*     Structure Words and Branches
 *=================================================================================*/

/********************************************************************************
 * EXIT				    --					CORE
 *                      R: nest-sys --
 */
PRIM_WORD (EXIT, "EXIT")
DoRet:		move.l	(RSP)+, IP		// Load address from Returnstack into IP
		NEXT				// Directly execute NEXT (macro, not jmp)

/*-------------------------------------------------------------------------------
 * bra			--						internal
 */
INTERNAL_PRIM_WORD (_bra, "bra")
		movea.w	(IP)+, IP			// branch to address, pointed to by (IP)
		adda.l	BP, IP
		NEXT

/*-------------------------------------------------------------------------------
 * ?bra  		flag --						internal
 */
INTERNAL_PRIM_WORD (question_bra, "?bra")
		tst.w	(PSP)+				// branch if flag = zero
		beq	_bra+2
		addq.l	#2, IP
		NEXT

/*-------------------------------------------------------------------------------
 * bra'			--						internal
 *
 * Branch with stack and Hand-ON checking (for loops).
 */
INTERNAL_PRIM_WORD (bra_, "bra'")
		move.l	PSB, d0				// check for stack-over- and underflow
		sub.l	PSP, d0
		chk	#PS_SIZE, d0
		cmp.w	#$848, tios::kb_vars+2		// hand-ON pressed?
		bne	_bra+2
\WaitNoKey:	tst.w	tios::kb_vars+2			// wait, until keys are released
		bne	\WaitNoKey
		moveq	#ERR_USER_INTERRUPT, d0		// User Interrupt Exception
		bra	DoThrow

/*-------------------------------------------------------------------------------
 * ?bra'		flag --						internal
 *
 * Conditional branch with stack and hand-ON checking (for loops).
 */
INTERNAL_PRIM_WORD (question_bra_, "?bra'")
		tst.w	(PSP)+				// branch if flag = zero
		beq	bra_+2
		addq.l	#2, IP
		DO_NEXT

/*-------------------------------------------------------------------------------
 * of			x1 x2 -- x1					internal
 *
 * Is compiled by OF. Check whether x1 and x2 are equal and discard x2. 
 * If they are equal also discard x1 and continue execution behind of. Else
 * branch to the address stored in the cell that follows of in the threaded-code.
 */
INTERNAL_PRIM_WORD (of, "of")
		move.w	(PSP)+, d0		// discard x2
		cmp.w	(PSP), d0
		bne	_bra+2
		addq.l	#2, PSP			// discard x1
		addq.l	#2, IP			// continue execution behind of
		DO_NEXT

/*-------------------------------------------------------------------------------
 * do			n1|u1 n2|u2 --					internal
 *
 * Is compiled by DO
 * Create loop-sys:	R: leave-addr limit index
 */
INTERNAL_PRIM_WORD (do, "do")	
		move.w	(IP)+, -(RSP)			// push leave-addr
		move.l	(PSP)+, -(RSP)			// push limit; push index=start
		DO_NEXT

/*-------------------------------------------------------------------------------
 * ?do			n1|u1 n2|u2 --					internal
 *
 * Is compiled by ?DO
 */
INTERNAL_PRIM_WORD (question_do, "?do")
		move.w	(PSP)+, d0			// d0 = start
		move.w	(PSP)+, d1			// d1 = limit
		cmp.w	d0, d1				// if start=limit: branch to loop-end
		beq	bra_+2
		move.w	(IP)+, -(RSP)                   // else push loop-sys: leave-addr
		move.w	d1, -(RSP)                      //                     limit
		move.w	d0, -(RSP)			//                     index=start
		DO_NEXT

/*-------------------------------------------------------------------------------
 * loop			--						internal
 *			R: loop-sys1 -- loop-sys2 |
 * Is compiled by LOOP
 */
INTERNAL_PRIM_WORD (loop, "loop")
		addq.w	#1, (RSP)			// increase loop index
		move.w	(RSP), d0
		cmp.w	2(RSP), d0			// if index <> limit: branch to loop begin
		bne	bra_+2
		addq.l	#6, RSP				// else: remove loop-sys
		addq.l	#2, IP				// and continue execution after loop
		DO_NEXT

/*-------------------------------------------------------------------------------
 * +loop		n --						internal
 *			R: loop-sys1 -- loop-sys2 |
 * Is compiled by +LOOP
 */
INTERNAL_PRIM_WORD (plus_loop, "+loop")	
		move.w	2(RSP), d1			// d1 = limit
		move.w	(RSP), d0			// d0 = index
		move.w	d0, d2				// d2 = new index
		add.w	(PSP)+, d2			// increase/decrease index
		move.w	d2, (RSP)
		cmp.w	d1, d0				// check, in which direction boundary has to be crossed
		blt	\FromBelow
		/** limit has to cross the boundary (limit-1 limit) from above limit */
		cmp.w	d1, d2				// if index >= limit: branch to loop begin
		bge	bra_+2
		bra	\NoJmp
		/** limit has to cross the boundary (limit-1 limit) from below limit-1 */
\FromBelow:	cmp.w	d1, d2				// if index < limit: branch to loop begin
		blt	bra_+2
\NoJmp:		addq.l	#6, RSP				// else: remove loop-sys
		addq.l	#2, IP				// and continue execution after loop
		jmp	(NP)

/********************************************************************************
 * LEAVE			    --					CORE
 *                      R: loop-sys --
 */
PRIM_WORD (LEAVE, "LEAVE")	
		addq.l	#4, RSP				// remove limit and index
		movea.w	(RSP)+, IP		 	// branch to end-addr
		add.l	BP, IP
		DO_NEXT

/********************************************************************************
 * UNLOOP			    --					CORE
 *                      R: loop-sys --
 */
PRIM_WORD (UNLOOP, "UNLOOP")
		addq.l	#6, RSP
		DO_NEXT

/********************************************************************************
 * I		   	-- n|u						CORE
 */
SAME_PRIM_WORD (I, "I", R_FETCH)

/********************************************************************************
 * J			-- n|u						CORE
 */
PRIM_WORD (J, "J")	
		move.w	6(RSP), -(PSP)
		DO_NEXT

/*-------------------------------------------------------------------------------
 * mark>		-- orig 					internal
 *
 * Put the address of a new unresolved forward reference onto the stack, allocate
 * one cell that holds the branch destination.
 *
 * Definition:
 *    : mark>   HERE 0 , -4 ;
 */
INTERNAL_SEC_WORD (mark_greater, "mark>")
	XT(HERE) XT(ZERO) XT(COMMA) XT(MINUS_FOUR) XT(EXIT)

/*-------------------------------------------------------------------------------
 * resolve>		orig -- 					internal
 *
 * Resolve forward reference, created by mark>.
 *
 * Definition:
 *    : resolve>   -4 ?pairs   HERE SWAP ! ;
 */
#define resolve_greater THEN

/*-------------------------------------------------------------------------------
 * mark<		-- dest 					internal
 *
 * Put the next location for a transfer of controll onto the stack.
 *
 * Definition:
 *    : mark<   HERE -2 ;
 */
#define	mark_less BEGIN

/*-------------------------------------------------------------------------------
 * resolve<		dest -- 					internal
 *
 * Add the address of a transfer of controll to 'dest' to the dictionary.
 *
 * Definition:
 *    : mark<   -2 ?pairs , ;
 */
INTERNAL_SEC_WORD (resolve_less, "resolve<")
	XT(MINUS_TWO) XT(question_pairs) XT(COMMA) XT(EXIT)

/********************************************************************************
 * CS-PICK	destu ... orig0|dest0 u -- destu ... orig0|dest0 destu 	TOOLS EXT
 * 
 * Definition:    : CS-PICK    2* 1+ >R   R@ PICK   R> PICK ;
 */
SEC_WORD (CS_PICK, "CS-PICK")
	XT(TWO_STAR) XT(ONE_PLUS) XT(TO_R)   XT(R_FETCH) XT(PICK)   XT(R_FROM) XT(PICK) XT(EXIT)

/********************************************************************************
 * CS-ROLL	origu|destu origu-1|destu-1 ... orig0|dest0 u --	TOOLS EXT
 *                         origu-1|destu-1 ... orig0|dest0 origu|destu 
 * 
 * Definition:    : CS-ROLL    2* 1+ >R   R@ ROLL   R> ROLL ;
 */
SEC_WORD (CS_ROLL, "CS-ROLL")
	XT(TWO_STAR) XT(ONE_PLUS) XT(TO_R)   XT(R_FETCH) XT(ROLL)   XT(R_FROM) XT(ROLL) XT(EXIT)

/********************************************************************************
 * AHEAD		-- orig 					TOOLS EXT
 * 
 * Definition:    : AHEAD    compile bra mark> ; IMMEDIATE
 */
SEC_WORD (AHEAD, "AHEAD", IMMEDIATE)
	XT(compile) XT(_bra) XT(mark_greater) XT(EXIT)

/********************************************************************************
 * IF			-- orig						CORE
 *
 * Definition:    : IF   compile ?bra  mark> ; IMMEDIATE  
 */
SEC_WORD (_IF, "IF", IMMEDIATE)
	XT(compile) XT(question_bra) XT(mark_greater) XT(EXIT)

/********************************************************************************
 * THEN			orig --						CORE
 *
 * Definition:    : THEN   resolve> ; IMMEDIATE 
 */
SEC_WORD (THEN, "THEN", IMMEDIATE)
	XT(MINUS_FOUR) XT(question_pairs)  XT(HERE) XT(_SWAP) XT(STORE) XT(EXIT)

/********************************************************************************
 * ELSE			orig1 -- orig2					CORE
 *
 * Definition:    : ELSE   compile bra mark>  2SWAP resolve> ; IMMEDIATE
 */
SEC_WORD (ELSE, "ELSE", IMMEDIATE)
	XT(compile) XT(_bra) XT(mark_greater)  XT(TWO_SWAP) XT(resolve_greater) XT(EXIT)

/********************************************************************************
 * BEGIN		-- dest						CORE
 *
 * Definition:    : BEGIN  mark< ; IMMEDIATE   
 */
SEC_WORD (BEGIN, "BEGIN", IMMEDIATE)
	XT(HERE) XT(MINUS_TWO) XT(EXIT)

/********************************************************************************
 * UNTIL		dest --						CORE
 *
 * Definition:    : UNTIL   compile ?bra' resolve< ; IMMEDIATE
 */
SEC_WORD (UNTIL, "UNTIL", IMMEDIATE)
	XT(compile) XT(question_bra_) XT(resolve_less) XT(EXIT)

/********************************************************************************
 * AGAIN		dest --						CORE EXT
 *
 * Definition:    : AGAIN   compile bra' resolve< ; IMMEDIATE 
 */
SEC_WORD (AGAIN, "AGAIN", IMMEDIATE)
	XT(compile) XT(bra_) XT(resolve_less) XT(EXIT)

/********************************************************************************
 * WHILE		dest --	orig dest				CORE
 *
 * Definition:    : WHILE   DUP -2 ?pairs   POSTPONE IF   2SWAP ; 
 */
SEC_WORD (WHILE, "WHILE", IMMEDIATE)
	XT(DUP) XT(MINUS_TWO) XT(question_pairs)  XT(_IF) XT(TWO_SWAP) XT(EXIT)

/********************************************************************************
 * REPEAT		orig dest --					CORE
 *
 * Definition:    : REPEAT   POSTPONE AGAIN   resolve> ; IMMEDIATE
 */
SEC_WORD (REPEAT, "REPEAT", IMMEDIATE)
	XT(AGAIN) XT(resolve_greater) XT(EXIT)

/********************************************************************************
 * DO			-- do-sys					CORE
 *
 * Definition:    : DO   compile do   mark> mark< -1 ; IMMEDIATE 
 */
SEC_WORD (DO, "DO", IMMEDIATE)	
	XT(compile) XT(do)  __DO__: XT(mark_greater) XT(mark_less) XT(MINUS_ONE) XT(EXIT)

/********************************************************************************
 * LOOP			do-sys --					CORE
 *
 * Definition:    : LOOP   compile loop   -1 ?pairs    resolve< resolve> ;
 */
SEC_WORD (LOOP, "LOOP", IMMEDIATE)
	XT(compile) XT(loop)  __LOOP__: XT(MINUS_ONE) XT(question_pairs) 
	XT(resolve_less) XT(resolve_greater) XT(EXIT)

/********************************************************************************
 * ?DO			-- do-sys					CORE EXT
 *
 * Definition:
 *    : ?DO   compile ?do bra __DO__ ( ; )  IMMEDIATE  ( compile a ?do, everything else: same as DO )
 */
SEC_WORD (QUESTION_DO, "?DO", IMMEDIATE)
	XT(compile) XT(question_do)  XT(_bra) XT(__DO__)

/********************************************************************************
 * +LOOP			do-sys --				CORE
 *
 * Definition:
 *    : +LOOP   compile +loop bra __LOOP__ ( ; ) IMMEDIATE   ( compile a +loop, same as in LOOP )
 */
SEC_WORD (PLUS_LOOP, "+LOOP", IMMEDIATE)
	XT(compile) XT(plus_loop)  XT(_bra) XT(__LOOP__)

/********************************************************************************
 * CASE			-- case-sys					CORE EXT
 *
 * Possible Definition:   : CASE   0 0 ; IMMEDIATE
 */
PRIM_WORD (_CASE, "CASE", IMMEDIATE)
		clr.l	-(PSP)
		jmp	(NP)

/********************************************************************************
 * OF		 	Compilation:       -- of-sys 			CORE EXT
 *                    	Run-time:    x1 x2 -- | x1
 * Definition:  
 *    : OF   
 *       DUP 0 ?pairs                     ( check whether case-sys )
 *       compile of mark> 1 ; IMMEDIATE   ( create of and of-sys )
 */
SEC_WORD (OF, "OF", IMMEDIATE)
	XT(DUP) XT(ZERO) XT(question_pairs)   
	XT(compile) XT(of) XT(mark_greater) XT(ONE) XT(EXIT)

/********************************************************************************
 * ENDOF	 	case-sys1 of-sys  -- case-sys2 			CORE EXT
 *
 * Definition:  
 *    : ENDOF   
 *       1 ?pairs                          ( check whether of-sys )
 *       compile bra mark>                 ( create branch to endcase )
 *       2SWAP resolve>  ROT ; IMMEDIATE   ( resolve OF, put 0 to top of case-sys )
 */
SEC_WORD (ENDOF, "ENDOF", IMMEDIATE)
	XT(ONE) XT(question_pairs) 
        XT(compile) XT(_bra) XT(mark_greater)
	XT(TWO_SWAP) XT(resolve_greater) XT(ROT) XT(EXIT)
	
/********************************************************************************
 * ENDCASE	 	Compilation: case-sys -- 			CORE EXT
 *                      Run-time:           x -- 
 * Definition:
 *    : ENDCASE
 *       0 ?pairs   POSTPONE DROP    ( check whether case-sys, append run-time semantics of ENDCASE to current definition )
 *       BEGIN ?DUP WHILE            ( until end of case-sys (ends on 0)
 *          resolve>                 ( resolve one ENDOF )
 *       REPEAT ; IMMEDIATE      
 */
SEC_WORD (ENDCASE, "ENDCASE", IMMEDIATE)
	XT(ZERO) XT(question_pairs)   XT(compile) XT(DROP)
	BEGIN_WHILE({ XT(QUESTION_DUP) },{/*WHILE*/
	   XT(resolve_greater)
	})/*REPEAT*/ XT(EXIT)

/********************************************************************************
 * RECURSE			--					CORE
 *
 * Definition:   : RECURSE   cf , ; IMMEDIATE
 */
SEC_WORD (RECURSE, "RECURSE", IMMEDIATE)
	XT(cf) XT(COMMA) XT(EXIT)

/** 
 * 3.5 Compiler-Directives
 */		
PRIM_WORD (LEFT_BRACKET, "[", IMMEDIATE)	// ( -- )				CORE
		/*( possibly ) : [  0 STATE ! ;  IMMEDIATE*/
		clr.w	RELP(STATE)+2(BP)
		jmp	(NP)
PRIM_WORD (RIGHT_BRACKET, "]", IMMEDIATE)	// ( -- )				CORE
		/*( possibly ) : ]  1 STATE ! ;  IMMEDIATE*/
		move.w	#1, RELP(STATE)+2(BP)
		jmp	(NP)

/*-------------------------------------------------------------------------------
 * compile	 	-- 						internal
 *
 * Add the cell that follows compile to the dictionary.
 */
INTERNAL_PRIM_WORD (compile, "compile")	// ( -- )					internal
		move.w	HERE+2(PC), d0
		move.w	(IP)+, 0(BP, d0.w)
		addq.w	#2, RELP(HERE+2)(BP)
		jmp	(NP)

SEC_WORD (LITERAL, "LITERAL", IMMEDIATE)	// ( x -- )					CORE
	XT(compile) XT(lit) XT(COMMA) XT(EXIT)		// : LITERAL comp lit , ;
SEC_WORD (TWO_LITERAL, "2LITERAL", IMMEDIATE)	// ( dx -- )					CORE
	XT(compile) XT(two_lit) XT(COMMA) XT(COMMA) XT(EXIT)	// : 2LITERAL comp 2lit , , ;
// toggles the smudge bit of the last definition, if it is 1 the word can't be found by FIND
INTERNAL_SEC_WORD (smudge, "smudge")	// ( -- )					internal
		//: smudge  GET-CURRENT @  64 TOGGLE ;
		XT(GET_CURRENT) XT(FETCH) LIT(SMUDGE) XT(TOGGLE) XT(EXIT)
SEC_WORD (IMMEDIATE, "IMMEDIATE", IMMEDIATE)// ( -- )					CORE
		//: smudge  GET-CURRENT @  32 TOGGLE ;
		XT(GET_CURRENT) XT(FETCH) LIT(IMMEDIATE) XT(TOGGLE) XT(EXIT)

/*==================================================================================
       I/O Routines
  ==================================================================================**
   ***********************************************************************************/

/*     Keyboard and Terminal Input
 *=================================================================================*/

/*-------------------------------------------------------------------------------
 * 'key			-- abs-address					internal
 *
 * Return the address of the current character in the keyboard-structure
 */
INTERNAL_PRIM_WORD (_key, "'key")
		pea	tios::kb_globals+$1E
		jmp	(NP)

/*-------------------------------------------------------------------------------
 * 'key?		-- abs-address					internal
 *
 * Return the address of the word in the keyboard-structure, that holds the
 * the number of keys in the keyboard buffer.
 *
 * Definition:
 *    : 'key?   'key -2 M+ ;
 */
INTERNAL_SEC_WORD (_key_question, "'key?")
	XT(_key) XT(MINUS_TWO) XT(M_PLUS) XT(EXIT)

/********************************************************************************
 * EKEY?		-- flag					FACILITY EXT
 *
 * Definition:
 *    : EKEY?   'key? A@ 0<> ;
 */
SEC_WORD (EKEY_QUESTION, "EKEY?")	
	XT(_key_question) XT(A_FETCH) XT(ZERO_NOT_EQUALS) XT(EXIT)

/********************************************************************************
 * EKEY			-- u					FACILITY EXT
 *
 * Possible definition (in case flib becomes incompatible):
 *    : EKEY   
 *       BEGIN ?EKEY UNTIL                            ( wait for a keypress )
 *       'key A@   0 'key? A!                         ( read key, clear keyboard buffer )
 *       DUP [ HEX ] $210B = IF SHUTDOWN EKEY THEN ;  ( if it was diamond+on: switch off, get another key )
 */
PRIM_WORD (EKEY, "EKEY")			
		jsr	flib::idle_loop
		move.w	d0, -(PSP)
		jmp	(NP)

/********************************************************************************
 * EKEY>CHAR		u -- u false | char true		FACILITY EXT 
 *
 * Definition:
 *    : EKEY>CHAR   DUP 14 256 WITHIN ;
 */
SEC_WORD (EKEY_TO_CHAR, "EKEY>CHAR")
		XT(DUP) LIT(14) LIT(256) XT(WITHIN) XT(EXIT) 

/********************************************************************************
 * KEY?			-- flag						FACILITY
 *
 * Definition:
 *    : KEY?   
 *       EKEY? IF       ( is there a key? )
 *          'key A@ EKEY>CHAR NIP IF    ( if it is a character-key )
 *             TRUE EXIT                 ( return true )
 *          THEN
 *          EKEY DROP                    ( if it isn't a character-key: clear keyboard-buffer )
 *       THEN   FALSE ;                  ( and return false )
 */
SEC_WORD (KEY_QUESTION, "KEY?")
	XT(EKEY_QUESTION) IF({
	   XT(_key) XT(A_FETCH) XT(EKEY_TO_CHAR) XT(NIP) IF({
	      XT(TRUE) XT(EXIT)
	   })
	   XT(EKEY) XT(DROP)
	})/*THEN*/   XT(FALSE) XT(EXIT)

/********************************************************************************
 * KEY			-- char						FACILITY
 *
 * Definition:
 *    : KEY  
 *       0   BEGIN   DROP   EKEY EKEY>CHAR UNTIL ; 
 */
SEC_WORD (KEY, "KEY")
	XT(ZERO) BEGIN_UNTIL({ XT(DROP) XT(EKEY) XT(EKEY_TO_CHAR) })/*UNTIL*/ XT(EXIT)

/********************************************************************************
 * ACCEPT		c-addr +n1 -- +n2				CORE
 *
 * Definition:
 *    : ACCEPT
 *       OVER + SWAP DUP BEGIN    ( S: end-addr start-addr current-addr )
 *          EKEY                          ( read key )
 *          case
 *             257 of                                  ( backspace )
 *                2DUP =  IF ." #" 
 *                ELSE 1- ." ###" THEN  
 *             endof 
 *             13  of                                  ( enter )
 *                SWAP - NIP EXIT  
 *             endof 
 *             OVER 4 PICK <                           ( other keys )
 *             SWAP EKEY>CHAR ROT AND 
 *             IF  DUP EMIT OVER C! 1+
 *             ELSE  DROP ." #" THEN
 *          endcase
 *	 AGAIN ( ; )  ( never returns from there )
 */
SEC_WORD (ACCEPT, "ACCEPT")
	XT(OVER) XT(PLUS) XT(_SWAP) XT(DUP) BEGIN_AGAIN({
	   XT(EKEY)
	   /*case*/
	      LIT(257) XT(of) LBL(\endof1)  
	         XT(TWO_DUP) XT(EQUALS) IF({ STR(CHR_BEL) 
		 },{/*ELSE*/  XT(ONE_MINUS) STR(CHR_BS,' ', CHR_BS) })/*THEN*/
	      /*endof*/XT(_bra) LBL(\endcase)
\endof1:      LIT(13)  XT(of) LBL(\endof2)  
                 XT(_SWAP) XT(MINUS) XT(NIP) XT(EXIT)
	      /* no code for endof, since EXIT makes it unreachable */
\endof2:      XT(OVER) XT(FOUR) XT(PICK) XT(LESS_THAN)  
              XT(_SWAP) XT(EKEY_TO_CHAR) XT(ROT) XT(_AND) 
	      IF({  XT(DUP) XT(EMIT) XT(OVER) XT(C_STORE) XT(ONE_PLUS)
	      },{/*ELSE*/  XT(DROP) STR(CHR_BEL) })/*THEN*/
\endcase:   /*endcase (does nothing since DROP is already done) */
	})/*AGAIN*/
	  
		
/*     Terminal Output
 *=================================================================================*/

/*-------------------------------------------------------------------------------
 * out			-- u						internal
 *
 * out is a value that is incremented every time EMIT is executed. It is used
 * by words like WORDS to output to the screen until it is filled, wait for 
 * keypress and continue.
 *
 * Definition:   0 VALUE out
 */
INTERNAL_VALUE(out, "out", 0)

/********************************************************************************
 * EMIT			x -- 						CORE
 */
PRIM_WORD (EMIT, "EMIT")
		move.w	(PSP)+, d0
		addq.w	#1, RELP(out+2)(BP)		// increase out
		jsr	forthlib::Emit
		jmp	(NP)
		
/********************************************************************************
 * CR			-- 						CORE
 *
 * possible Definition:   : CR   10 EMIT ;
 */
SEC_WORD (CR, "CR")			
		STR(CHR_LF) XT(EXIT)

/********************************************************************************
 * PAGE	   	    	-- 						CORE
 *
 * possible Definition:    : PAGE   12 EMIT ;
 */
SEC_WORD (PAGE, "PAGE")
		STR(CHR_FF) XT(EXIT)	

/********************************************************************************
 * INV	   	    	-- 						CORE
 *
 * Invert text color (switch between black on white and white on black).
 *
 * possible Definition:    : INV   1 EMIT ;
 */
SEC_WORD (INV, "INV")
		STR(CHR_INV) XT(EXIT)

/********************************************************************************
 * UL	   	    	-- 						CORE
 *
 * Toggle whether outputed characters should be underlined.
 *
 * possible Definition:    : UL   2 EMIT ;
 */
SEC_WORD (UL, "UL")	
		STR(CHR_UL) XT(EXIT)

/********************************************************************************
 * BOLD	   	    	-- 						CORE
 *
 * Toggles whether outputed characters should be bold.
 *
 * possible Definition:    : BOLD   3 EMIT ;
 */
SEC_WORD (BOLD, "BOLD")
		STR(CHR_BOLD) XT(EXIT)

/********************************************************************************
 * SPACE		-- 						CORE
 *
 * possible Definition:   : SPACE   BL EMIT ;
 */
SEC_WORD (SPACE, "SPACE")	
		STR(" ") XT(EXIT)	

/********************************************************************************
 * SPACES	       	n -- 						CORE
 *
 * Definition:    : SPACE   0 MAX 0 ?DO SPACE LOOP ;
 */
SEC_WORD (SPACES, "SPACES")		// ( n -- )					CORE
	XT(ZERO) XT(MAX) XT(ZERO) QDO_LOOP({ XT(SPACE) /*LOOP*/}) XT(EXIT)

/********************************************************************************
 * AT-XY	       	u1 u2 -- 					CORE
 */
PRIM_WORD (AT_XY, "AT-XY")
		move.w	(PSP)+, d1
		move.w	(PSP)+, d0
		jsr	forthlib::SetCursor
		jmp	(NP)

/********************************************************************************
 * WHERE-XY	       	-- u1 u2 					????
 *
 * Return current cursor position. 'u1' is the cursor's column, 'u2' the row.
 */
PRIM_WORD (WHERE_XY, "WHERE-XY")
		jsr	forthlib::GetCursor
		move.w	d0, -(PSP)
		move.w	d1, -(PSP)
		jmp	(NP)

/********************************************************************************
 * CURSOR-HEIGHT	u --	 					????
 *
 * Set height of text cursor to 'u' pixels. 'u' is automatically bounded to fall
 * into the range 1...8.
 */
PRIM_WORD (CURSOR_HEIGHT, "CURSOR-HEIGHT")
		move.w	(PSP)+, d0
		jsr	forthlib::CursorHeight
		jmp	(NP)

PRIM_WORD (SCROLL_DOWN, "SCROLL\")	// ( -- )					????
		jsr	forthlib::Scroll1Down
		jmp	(NP)
PRIM_WORD (SCROLL_UP, "SCROLL/")	// ( -- )					????
		jsr	forthlib::Scroll1Up
		jmp	(NP)
PRIM_WORD (SCROLL_LEFT, "SCROLL>")	// ( -- )					????
		jsr	forthlib::Scroll1Left
		jmp	(NP)
PRIM_WORD (SCROLL_RIGHT, "SCROLL<")	// ( -- )					????
		jsr	forthlib::Scroll1Right
		jmp	(NP)
PRIM_WORD (PLUS_CURSOR, "+CURSOR")	// ( -- ) 					????
		jsr	forthlib::ShowCursor
		jmp	(NP)
PRIM_WORD (MINUS_CURSOR, "-CURSOR")	// ( -- )					????
		jsr	forthlib::HideCursor
		jmp	(NP)
SEC_WORD (TYPE, "TYPE")			// ( c-addr u -- )				CORE
/* : TYPE ( c-addr u )    OVER + SWAP ?DO I C@ EMIT LOOP ; */
		XT(OVER) XT(PLUS) XT(_SWAP) QDO_LOOP({ XT(I) XT(C_FETCH) XT(EMIT) })/*LOOP*/ XT(EXIT)
SEC_WORD (COUNT, "COUNT")		// ( c-addr1 -- c-addr2 u )			CORE
		XT(DUP) XT(ONE_PLUS) XT(_SWAP) XT(C_FETCH)	// : COUNT  DUP 1+ SWAP C@ ;
		XT(EXIT)

/*-------------------------------------------------------------------------------
 * (.")			"ccc<quote" -- 					internal
 *
 * Is compiled by .". Display a string literal.
 *
 * Definition:
 *    : ."   2R> A>R COUNT   2DUP TYPE   bra __Skip_SLITERAL
 */
INTERNAL_SEC_WORD (dot_quote, '(.")')	
	XT(TWO_R_FROM) XT(A_TO_R) XT(COUNT) 
	XT(TWO_DUP) XT(TYPE) XT(_bra) XT(__Skip_SLITERAL)

/********************************************************************************
 * ."			 "ccc<quote>" -- 				CORE
 *
 * Definition:
 *    : ."   compile (.")   [CHAR] " PARSE   bra __Do_SLITERAL ;
 */
SEC_WORD (DOT_QUOTE, '."', IMMEDIATE)
			XT(compile) XT(dot_quote) 
__Quote_SLITERAL: 	LIT('"') XT(PARSE) XT(_bra) XT(__Do_SLITERAL)

/********************************************************************************
 * .(			"ccc<paren>" --	 			CORE
 *
 * Definition:
 *    : .(   [CHAR] ) PARSE TYPE ;
 */
SEC_WORD (DOT_PAREN, ".(", IMMEDIATE)
	LIT(')') XT(PARSE) XT(TYPE) XT(EXIT)


/*   BLOCK word set
 *=================================================================================*/

/*
 * Block Descriptors
 */
		EVEN
Descr1:	
Descr1.Blk:	dc.w	0			// block number (unassigned)
Descr1.Addr:	dc.w	RELP(BlockBuffer1)	// address of buffer
Descr1.UpdFlag:	dc.b	0			// update-flag

		EVEN
Descr2:	
Descr2.Blk:	dc.w	0			// block number (unassigned)
Descr2.Addr:	dc.w	RELP(BlockBuffer2)	// address of buffer
Descr2.UpdFlag:	dc.b	0			// update-flag
BlockLoaded:	dc.b	0			// flag that is set if a block is allready in memory

		EVEN
		
/*
 * Tool routines
 */		

/** 
 * Swap the two block descriptors (block buffer 2 becomes active block) 
 * Input: none
 * Output: a0/a1/d0/d1 destroyed
 */
SwapDescriptors:
		lea	Descr1(PC), a0
		lea	Descr2(PC), a1
		moveq	#5-1, d1		// d1 = sizeof descriptor -1 (counter)
\SwapLoop:	move.b	(a0), d0
		move.b	(a1), (a0)+
		move.b	d0, (a1)+
		dbra	d1, \SwapLoop
		rts
/**
 * Save the block buffer #1 back to mass storage device, if it is assigned and its UPDATE flag 
 * is set, and clear the UPDATE flag.
 * Output: a0-a1/d0 is destroyed
 */
SaveBuffer1:	lea	Descr1(PC), a1			// a1 points to the block descriptor
		move.w	(a1)+, d0			// d0 = number of block
		beq	\DontUpdate			// if block is unassigned: don't update!
		movea.w	(a1)+, a0			// a0 = address of buffer
		adda.l	BP, a0
		tst.b	(a1)				// if UPDATE-flag is set
		beq	\DontUpdate
		jsr	forthlib::WriteBlock               // store it back to mass storage device
		bne	\ErrBlkWrite			   // Z-flag = 0: Block write error
		clr.b	(a1)				   // clear UPDATE flag
\DontUpdate:	rts

\ErrBlkWrite:	moveq	#ERR_BLK_WRITE, d0
		bra	DoThrow

/********************************************************************************
 * UPDATE		--						BLOCK
 */
PRIM_WORD (UPDATE, "UPDATE")
		st.b	RELP(Descr1.UpdFlag)(BP)	// set update-flag of first descriptor
		jmp	(NP)

/********************************************************************************
 * SAVE-BUFFERS		--						BLOCK
 */

PRIM_WORD (SAVE_BUFFERS, "SAVE-BUFFERS")
		bsr	SaveBuffer1			// save Buffer #1, if UPDATEd
		bsr	SwapDescriptors			// Buffer #2 is now Buffer #1
		bsr	SaveBuffer1			// save it, too 
		bsr	SwapDescriptors			// swap buffers back
		jmp	(NP)

/********************************************************************************
 * EMPTY-BUFFERS	--						BLOCK EXT
 */
PRIM_WORD (EMPTY_BUFFERS, "EMPTY-BUFFERS")
		clr.w	RELP(Descr1.Blk)(BP)		// assign buffers to block #0 (=unassigned)
		clr.w	RELP(Descr2.Blk)(BP)
		clr.b	RELP(Descr1.UpdFlag)(BP)	// clear update flags
		clr.b	RELP(Descr2.UpdFlag)(BP)
		jmp	(NP)

/********************************************************************************
 * FLUSH		--						BLOCK
 *
 * Definition:   : FLUSH   SAVE-BUFFERS EMPTY-BUFFERS ;
 */
SEC_WORD (FLUSH, "FLUSH")
		XT(SAVE_BUFFERS) XT(EMPTY_BUFFERS) XT(EXIT)
				
/********************************************************************************
 * BUFFER		u -- a-addr					BLOCK
 */
PRIM_WORD (BUFFER, "BUFFER")	
		move.w	(PSP)+, d2			// d2 = block number
		cmpi.w	#9999, d2			// check block number
		bhi	\InvalidBlk

		/******* Block is already loaded ? *****/
		st.b	RELP(BlockLoaded)(BP)		

		cmp.w	Descr1.Blk(PC), d2		// Buffer #1 contains this block ?
		beq	\End				
		bsr	SwapDescriptors			// Buffer #2 is now Buffer #1
		cmp.w	Descr1.Blk(PC), d2		// does it contain this block ?
		beq	\End			     

		/******* Block isn't already loaded ! **/
		sf.b	RELP(BlockLoaded)(BP)	
		bsr	SaveBuffer1			// now we use Block Buffer #1
		move.w	d2, RELP(Descr1.Blk)(BP)	// assigne Block Buffer #1 to that block
		
\End:		move.w	Descr1.Addr, -(PSP)		// return address of Block Buffer #1
		jmp	(NP)

\InvalidBlk:	moveq	#ERR_INV_BLK, d0
		bra	DoThrow

/********************************************************************************
 * BLOCK		u -- a-addr					BLOCK
 */
SEC_WORD (BLOCK, "BLOCK")
		XT(DUP) XT(BUFFER) XT(_SWAP) XT(\Block)
\Block:		XT(\Block+2)
		move.w	(PSP)+, d0			// d0 = number of block
		movea.w	(PSP), a0			// a0 = address of buffer
		adda.l	BP, a0
		tst.b	RELP(BlockLoaded)(BP) 		// if the current block had allready been loaded
		bne	\DontLoad		 	   // don't load it again
		jsr	forthlib::ReadBlock		// else transfer data from that block to the buffer
\DontLoad:	bra	DoRet

/********************************************************************************
 * LIST			u -- 						BLOCK EXT
 *
 * Definition:
 *    : LIST
 *       DUP SCR !   BLOCK 0   BEGIN           ( store u in SCR, load block; S:  a-addr column-scroll )
 *          2DUP +                             ( S: offset of first visible character in line #0 )
 *          PAGE  16 0 DO                      ( clear screen, print block )
 *             BASE @ HEX   I 1 INV .R INV  BASE !   ( print line number in one-digit-hex, inverted )
 *             I C/L * OVER +   39 I 15 = +   TYPE   ( print 39 chars of line, but only 38 chars in line #15)
 *          LOOP   DROP
 *          EKEY CASE                          ( process a key-press )
 *             340 OF  8 + ENDOF               ( key #340: scroll right )
 *             337 OF  8 - ENDOF               ( key #337: scroll left )
 *             264 OF  2DROP EXIT ENDOF        ( key #267: EScape -> exit)
 *             ." #"
 *          ENDCASE
 *          25 MIN 0 MAX                       ( bound column-scroll to 0...25 )
 *      AGAIN  ( ; ) ( never returns from here )
 */
SEC_WORD (LIST, "LIST")
	XT(DUP) XT(SCR) XT(STORE)  XT(BLOCK) XT(ZERO)   BEGIN_AGAIN({
 	   XT(TWO_DUP) XT(PLUS)
	   XT(PAGE) LIT(16) XT(ZERO) DO_LOOP({
	      XT(BASE) XT(FETCH) XT(HEX) XT(I) XT(ONE)  XT(INV) XT(DOT_R) XT(INV) XT(BASE) XT(STORE) 
	      XT(I) XT(C_PER_L) XT(STAR) XT(OVER) XT(PLUS)  LIT(39) XT(I) LIT(15) XT(EQUALS) XT(PLUS)  XT(TYPE)
	   })/*LOOP*/   XT(DROP)
	   XT(EKEY) /*CASE*/
	   LIT(340) XT(of) LBL(\endof1)  LIT(8) XT(PLUS) /*ENDOF*/XT(_bra) LBL(\endcase)
\endof1:   LIT(337) XT(of) LBL(\endof2)  LIT(8) XT(MINUS) /*ENDOF*/XT(_bra) LBL(\endcase)
\endof2:   LIT(264) XT(of) LBL(\endof3)  XT(TWO_DROP) XT(EXIT) /*no ENDOF due to EXIT */
\endof3:   STR(CHR_BEL)
	   XT(DROP)
\endcase:  /*ENDCASE (DROP is compiled by ENDCASE)*/
  	   LIT(25) XT(MIN) XT(ZERO) XT(MAX)
	})/*AGAIN*/


/*     File Access Wordset
 *=================================================================================*/

	// maximum size of a file (virtual size as returned by FILE-SIZE)
#define FILE_MAX_SIZE  (65535-6)
 
/********************************************************************************
 * R/O			-- fam 						FILE
 */
SAME_PRIM_WORD (R_O, "R/O", ZERO)
 
/********************************************************************************
 * W/O			-- fam 						FILE
 */
SAME_PRIM_WORD (W_O, "W/O", ONE)

/********************************************************************************
 * R/W			-- fam 						FILE
 */
SAME_PRIM_WORD (R_W, "R/W", TWO)

/********************************************************************************
 * BIN			fam1 -- fam2					FILE
 */
PRIM_WORD (BIN, "BIN")
		st	(PSP)
		jmp	(NP)

/*---------------------------------------
 * Helper routine, that converts a counted string into a zero terminated string. To do so
 * the string is copied to the temporal buffer `TempBuffer1'
 * Input:  c-addr u (stack)
 * Output: a0.l = pointer to string 
 *         d0/a1 destroyed
 */
ToCString:	movea.l	(a7)+, a0		// a0 = return-address
		move.w	(PSP)+, d0		// d0 = u
		move.w	(PSP)+, a1		// a1 = c-addr
		move.l	a0, -(a7)		// restore return-address
		adda.l	BP, a1
		
		subq.w	#1, d0			// d0 = counter
		cmpi.w	#17, d0			// bound string length to 18 characters
		bls	\StrLenOk
		moveq	#17, d0
\StrLenOk:	lea	TempBuffer1, a0		// a0 = destination

		move.l	a0, -(a7)		// save a0 
\CopyLoop:	move.b	(a1)+, (a0)+
		dbra	d0, \CopyLoop
		clr.b	(a0)
		movea.l	(a7)+, a0		// a0 = pointer to buffer begin
		rts

/********************************************************************************
 * DELETE-FILE		c-addr u -- ior					FILE
 */
PRIM_WORD (DELETE_FILE, "DELETE-FILE")
		bsr	ToCString		// a0 = zero terminated file name
		jsr	forthlib::DeleteFile	// delete the file
ChkIOError:     tst.w	d0			// if d0 is nonzero: file IO exception
                bne	IOError
		jmp	(NP)

/********************************************************************************
 * RENAME-FILE		c-addr1 u1 c-addr2 u2 -- ior			FILE
 *
 * It may also be used to move files from one folder to another. If the 
 * destination file exists, it will be removed. 
 */
PRIM_WORD (RENAME_FILE, "RENAME-FILE")
		bsr	ToCString		// a0 = zero terminated destination file name
		lea	TempBuffer2, a1
\CopyLoop:	move.b	(a0)+, (a1)+		// copy destination name to another buffer (a1)
		bne	\CopyLoop
		bsr	ToCString		// a0 = zero terminated source file name
		lea	TempBuffer2, a1		// a1 = zero terminated destination file name
		jsr	forthlib::MoveFile	// rename the file
		bra	ChkIOError

/********************************************************************************
 * OPEN-FILE		c-addr u fam -- fileid ior			FILE
 */
PRIM_WORD (OPEN_FILE, "OPEN-FILE")
		move.w	(PSP)+, d1			// d1 = fam
		bsr	ToCString			// a0 = file name, zero terminated
		jsr	forthlib::OpenFile		// d0 = handle of file (fileid) (or zero on error)
		move.w	d0, -(PSP)			// push  fileid
		beq	IOError				// handle=0: file I/O exception
		
		move.l	a0, a1				// a0 = pointer to SYM_ENTRY of file
				
		/** check whether file is a text variable */
		bsr	DEREF_d0_a0			// a0 = pointer to variable
		moveq	#0, d2		
		move.w	(a0)+, d2			// d2.l = length of var
		cmpi.b	#$E0, -1(a0, d2.l)		// file is not a text variable ?
		bne	IOError				// --> file I/O exception
		move.w	#1, (a0)			// set the file-position field to one

		/** lock file if it was opened in BIN mode */
		tst.w	d1				// if fam is negative (BIN mode)
		bpl	\NotBIN
		bset.b	#3, tios::SYM_ENTRY.flags+1(a1)	// --> lock the file
\NotBIN:
		clr.w	-(PSP)				// ior=zero (success)
		jmp	(NP)
		
IOError:	move.w	#ERR_FILE_IO, -(PSP)		// push ior = number of file I/O exception
		jmp	(NP)

/********************************************************************************
 * CREATE-FILE		c-addr u fam -- fileid ior			FILE
 */
PRIM_WORD (CREATE_FILE, "CREATE-FILE")
					// create the handle of the file (an empty TEXT)
		pea	(7).w				// shorter than move.l !!!
		jsr	tios::HeapAlloc			// d0 = handle
		addq.l	#4, a7
		tst.w	d0				// if handle = zero:
		beq	\IOError			// --> error (not enough memory)
		bsr	DEREF_d0_a0			// a0 = pointer to newly created mem blk
		move.l	#$00050001, (a0)+		// set the size-specifier and cursor pos.
		move.w	#$2000, (a0)+			// output one space + EOF character
		move.b	#$E0, (a0)+			// output TEXT identifier
		move.w	d0, d2				// d2 = file-handle

		move.w	(PSP)+, d1			// d1 = fam
		tst.b	d1				// low byte of fam is R/O W/O or R/W
		beq	\FreeAndError			// if fam = R/O --> error
		
		bsr	ToCString			// a0 = pointer to file name
		move.w	d2, d0				// d0 = handle to associate with file
		jsr	forthlib::CreateFile		// create the file; d0 = zero on success
		bne	\FreeAndError

		tst.w	d1				// if fam is negative (BIN mode)
		bpl	\NotBIN
		bset.b	#3, tios::SYM_ENTRY.flags+1(a0)	// --> lock the file
\NotBIN:
		move.w	d2, -(PSP)			// push the file's handle (ior)
		clr.w	-(PSP)				// push ior = zero
		jmp	(NP)
		
\FreeAndError:	move.w	d2, -(a7)			// file creation error: free the mem blk
		jsr	tios::HeapFree
		addq.l	#2, a7
\IOError:	clr.w	-(PSP)				// push fileid=zero
		bra	IOError				// and push ior= number of file I/O exception


/*-------------------------------------------------------------------------------
 * PrepRead			Helper Routine				internal
 *
 * Prepare for reading from a file
 *
 * Input:  d0.w = fileid
 *         d1.w = #bytes to read
 * Output: d0.w = remaining bytes in file
 *	   d1.w = #bytes to read, decreased to don't overflow the file end
 *	   a0.l = pointer to file-position word
 *         a1.l = pointer to byte at current file position
 *         d2.l modified
 */
PrepRead:	bsr	DEREF_d0_a0
		move.w	(a0)+, d0		// d0.w = file length
		moveq	#0, d2
		move.w	(a0), d2		// d2.l = file position
		lea	2(a0, d2.l), a1		// a1 = pointer to byte at file position
		sub.w	d2, d0			// d0 = number of remaining bytes
		subq.l	#4, d0
		cmp.w	d0, d1			// adjust d1
		bls	\D1Ok
		move.w	d0, d1
\D1Ok:		rts

/********************************************************************************
 * READ-FILE		c-addr u1 fileid -- u2 ior			FILE
 */
PRIM_WORD (READ_FILE, "READ-FILE")
		move.w	(PSP)+, d0		// d0.w = fileid
		move.w	(PSP)+, d1		// d2.w = u1
		
		bsr	PrepRead		// prepare for reading (see above)
					
		add.w	d1, (a0)		// adjust file-position word
		
		move.w	(PSP)+, a0		// a0.l = c-addr (destination)
		adda.l	BP, a0
		move.w	d1, -(PSP)		// push u2=d1
		
		subq.w	#1, d1			// d1 is our counter for copying
		bmi	\End
\CopyLoop:	move.b	(a1)+, (a0)+		// copy from file (a1) to c-addr (a0)
		dbra	d1, \CopyLoop
		
\End:		clr.w	-(PSP)			// push ior=zero
		jmp	(NP)

/********************************************************************************
 * READ-LINE		c-addr u1 fileid -- u2 flag ior			FILE
 */
PRIM_WORD (READ_LINE, "READ-LINE")
		move.w	(PSP)+, d0		// d0.w = fileid
		move.w	(PSP)+, d1		// d1.w = u1
		move.l	a2, -(RSP)		// save a2
		move.w	(PSP)+, a2		// a2.l = c-addr (destination)
		adda.l	BP, a2
		clr.w	-(PSP)			// push u2=zero (is incremented during copy)
		
		bsr	PrepRead		// prepare for reading (see above)
		tst.w	d0			// check whether2 end of file (return flag=zero)
		beq	\FileEnd

		subq.w	#1, d1			// d1 is our counter for copying
		bmi	\End
						// copy from file (a1) to c-addr (a0)
\CopyLoop:	cmpi.b	#13, (a1)		  // line break ?
		beq	\LineBreak
		move.b	(a1)+, (a2)+		  // copy one byte
		addq.w	#1, (PSP)		  // increment u2
		addq.w	#1, (a0)		  // increment file-position word
		dbra	d1, \CopyLoop

\End:		move.w	#-1, -(PSP)		// push flag = true
		clr.w	-(PSP)			// push ior = zero
\Ret:		move.l	(RSP)+, a2		// restore a2
		jmp	(NP)

\LineBreak:	addq.w	#1, (a0)		// set file-position behind line break
		cmpi.b	#32, 1(a1)			
		bne	\End
		addq.w	#1, (a0)		// line break has normaly two chars (13, 32)
		bra	\End

\FileEnd:	clr.l	-(PSP)			// push flag=ior=zero
		bra	\Ret
		
/********************************************************************************
 * FILE-POSITION		fileid -- ud ior			FILE
 */
PRIM_WORD (FILE_POSITION, "FILE-POSITION")
		move.w	(PSP)+, d0
		bsr	DEREF_d0_a0
		move.w	2(a0), -(PSP)			// push the file position
		subq.w	#1, (PSP)			// make it the virtual file position
		clr.l	-(PSP)				// extend it to double cell, push ior=0
		jmp	(NP)

/********************************************************************************
 * FILE-SIZE			fileid -- ud ior			FILE
 */
PRIM_WORD (FILE_SIZE, "FILE-SIZE")
		move.w	(PSP)+, d0
		bsr	DEREF_d0_a0
		move.w	(a0), -(PSP)			// push the file's size
		subq.w	#5, (PSP)			// convert it to virtual size
		clr.l	-(PSP)				// extend it to double cell, push ior=0
		jmp	(NP)

/********************************************************************************
 * REPOSITION-FILE		ud fileid -- ior			FILE
 */
PRIM_WORD (REPOSITION_FILE, "REPOSITION-FILE")
		move.w	(PSP)+, d0		// d0 = handle of file
		bsr	DEREF_d0_a0		// a0 = pointer to begin of variable
		move.l	(PSP)+, d0		// d0 = virtual fileposition
		cmpi.l	#FILE_MAX_SIZE, d0	// bound file position (so we don't get overflows)
		bhi	\InvPos
		
		move.w	(a0)+, d2		// d2 = file size
		subq.w	#5, d2			// d2 = virtual file size (=FILE-SIZE)
		
		cmp.w	d0, d2			// filesize <= fileposition 
		bls	\InvPos			//   -> file position overflow
		
		addq.w	#1, d0			// set the file position
		move.w	d0, (a0)		//  (fileposition+1 is stored to the file)
		clr.w	-(PSP)			// return ior=zero (success)
		jmp	(NP)
		
\InvPos:	move.w	#ERR_INV_FILE_POS, -(PSP)  // return ior = exception number
		jmp	(NP)


/********************************************************************************
 * WRITE-FILE		c-addr u fileid -- ior				FILE
 */
PRIM_WORD (WRITE_FILE, "WRITE-FILE")
		move.w	(PSP)+, d0		// d0 = fileid
		
		bsr	DEREF_d0_a0		// a0 = pointer to variable
		move.w	(a0)+, d1		// d1 = actual_filesize
		moveq	#0, d3			// d3 must be accessable as long word
		move.w	(a0), d3		// d3.l = actual_filepos = virtual_filepos + 1
		add.w	(PSP), d3		// d3.l = new actual_filepos (at end of operation)
		bcs	\IOError
		cmpi.w	#FILE_MAX_SIZE+1, d3	// check file size
		bhi	\IOError
		addq.w	#4, d3			// d3.l = actual_filepos+4
		cmp.w	d1, d3			// if actual_filepos+4 > actual_filesize
		bls	\DontResize		//  --> we have to resize the file
		
		/******* Resize Variable *******/
					// WORD HeapRealloc(WORD handle, LONG newsize)
		move.l	d3, -(a7)		// new size of handle is actual_filepos+6
		addq.l	#2, (a7)
		move.w	d0, -(a7)		// push handle (fileid)
		jsr	tios::HeapRealloc
		addq.l	#6, a7
		tst.w	d0			// d0 = zero: error (not enough memory)
		beq	\IOError
		bsr	DEREF_d0_a0		// a0 = pointer to reallocated memory region
		move.w	d3, (a0)+		// write the file-size word at the var's begin
		clr.b	-2(a0, d3.l)		// create a new end-of-text-file mark
		move.b	#$E0, -1(a0, d3.l)	// create a new TEXT identifier

		/******* Write To Variable *****/
\DontResize:	move.w	(PSP)+, d1		// d1 = u
		move.w	(PSP)+, a1		// a1 = c-addr
		adda.l	BP, a1
		
		moveq	#0, d2			
		move.w	(a0), d2		// d2.l = file position (relative)
		add.w	d1, (a0)+		// adjust file position to end of data
		adda.l	d2, a0			// a0.l = file position (absolute)

		subq.w	#1, d1			// d1 = counter
		bmi	\End
\Loop:		move.b	(a1)+, (a0)+		// copy data
		dbra	d1, \Loop

\End:		clr.w	-(PSP)			// return ior = 0 (success)
		jmp	(NP)

\IOError:	addq.l	#4, PSP			// remove c-addr and u
		bra	IOError			// push ior = file I/O exception number

/********************************************************************************
 * RESIZE-FILE		ud fileid -- ior				FILE
 */
PRIM_WORD (RESIZE_FILE, "RESIZE-FILE")
		move.w	(PSP)+, d0		// d0 = fileid
		move.l	(PSP)+, d3		// d3 = new size
		cmpi.l	#FILE_MAX_SIZE, d3	// check file size
		bhi	IOError
		addq.w	#5, d3			// size+5 is stored at the var's begin
		
					// WORD HeapRealloc(WORD handle, LONG newsize)
		move.l	d3, -(a7)
		addq.l	#2, (a7)		// actual file size is size+7
		move.w	d0, -(a7)		// push handle
		jsr	tios::HeapRealloc
		addq.l	#6, a7
		tst.w	d0
		beq	IOError
		
		bsr	DEREF_d0_a0		// a0 = pointer to variable begin
		move.w	d3, (a0)+		// change the variable-size-word
		clr.b	-2(a0, d3.l)		// create a new end-of-text-file mark
		move.b	#$E0, -1(a0, d3.l)	// create a new TEXT identifier
		move.w	#1, (a0)+		// seek to var's begin
		
		clr.w	-(PSP)			// return ior=zero (success)
		jmp	(NP)
		
/********************************************************************************
 * WRITE-LINE		c-addr u fileid -- ior				FILE
 *
 * Definition:   
 *    : WRITE-LINE   
 *       DUP >R WRITE-FILE            ( write line )
 *       S" ##" R> WRITE-FILE OR ;    ( write line terminator and return ior = ior1 or ior2 )
 */
SEC_WORD (WRITE_LINE, "WRITE-LINE")
	XT(DUP) XT(TO_R) XT(WRITE_FILE) 
	SLIT(13,' ') XT(R_FROM) XT(WRITE_FILE) XT(_OR) XT(EXIT)

/********************************************************************************
 * CLOSE-FILE		fileid -- ior					FILE
 *
 * Definition:   : CLOSE-FILE   0 0 ROT REPOSITION-FILE ;
 */
SEC_WORD (CLOSE_FILE, "CLOSE-FILE")
	XT(ZERO) XT(ZERO) XT(ROT) XT(REPOSITION_FILE) XT(EXIT)


/*==================================================================================
       Number Conversion
  ==================================================================================**
   ***********************************************************************************/

/*     Number Input Conversion
 *=================================================================================*/

/*-------------------------------------------------------------------------------
 * 1stchar=?		c-addr u char -- c-addr u flag			internal
 *
 * Return whether the first character in string c-addr u is equal to char.
 * If the string is empty return false.
 *
 * Definition:   : 1stchar=?   2 PICK C@ =  OVER 0<> AND ;
 */
INTERNAL_SEC_WORD (firstchar_equals_question, "1stchar=?")
	XT(TWO) XT(PICK) XT(C_FETCH) XT(EQUALS)  XT(OVER) XT(ZERO_NOT_EQUALS) XT(_AND) XT(EXIT)
 
/*-------------------------------------------------------------------------------
 * >sign		c-addr1 u1 -- c-addr2 u2 1 | -1			internal
 *
 * Return the sign of the number string c-addr1 u1. Return the string c-addr2 u2
 * with the sign removed.
 *
 * Definition:   
 *    : >sign   
 *        [CHAR] - 1stchar=? IF 1 /STRING  -1 EXIT THEN
 *        [CHAR] + 1stchar=? IF 1 /STRING THEN
 *        1 ;
 */
INTERNAL_SEC_WORD (to_sign, ">sign")
	LIT('-') XT(firstchar_equals_question) IF({ XT(ONE) XT(SLASH_STRING) XT(MINUS_ONE) XT(EXIT) })/*THEN*/
	LIT('+') XT(firstchar_equals_question) IF({ XT(ONE) XT(SLASH_STRING) })/*THEN*/
	XT(ONE) XT(EXIT) 

/********************************************************************************
 * >NUMBER		ud1 c-addr1 u1 -- ud2 c-addr2 u2		CORE
 */
PRIM_WORD (TO_NUMBER, ">NUMBER")		
		move.w	(PSP)+, a0		// a0 = u1
		move.w	(PSP)+, a1		// a1 = c-addr1		
		move.l	(PSP)+, d0		// d0 = ud1
		
\Loop:		move.w	a0, d2			// u1 == zero ?
		beq	\End			//   -> end of conversion
		
		moveq	#0, d1			// d1.l = current character
		move.b	0(BP, a1.w), d1
		subi.w	#'0', d1		// convert d1.w to digit
		cmpi.w	#9, d1			//   digit 0...9
		ble	\__Digit0to9
		sub.w	#'A'-'0', d1		//   digit A...Z
		bmi	\End
		add.w	#10, d1
\__Digit0to9:	cmp.w	BASE+2(PC), d1		//   digit in range 0...BASE-1 ?
		bcc	\End
		
		move.l	d0, d2			// mulitplicate ud1 by BASE
		swap	d2			//   multiplicate high word
		mulu	BASE+2(PC), d2
		swap	d2
		mulu	BASE+2(PC), d0		//   mulitplicate low word
		add.l	d2, d0
		add.l	d1, d0			// add digit to ud1
		
		addq.w	#1, a1			// go to next character:
		subq.l	#1, a0			//    increase c-addr1 and decrease u1
		bra	\Loop
		
\End:		move.l	d0, -(PSP)		// push return values...
		move.w	a1, -(PSP)
		move.w	a0, -(PSP)
		jmp	(NP)

/********************************************************************************
 * number	   c-addr u -- n|u true | d|ud true | true | false 	CORE
 *              F:          --  | r
 *
 * Try to convert the string in c-addr u to a single or double precision integer
 * or to a floating point number, following the specifications for parsing 
 * numbers in the input source. Depending on STATE return the number or create
 * an according literal (LITERAL, 2LITERAL or FLITERAL). If the number can't be
 * converted only flag=FALSE is returned, else flag is TRUE.
 *
 * Definition:
 *    : number
 *       2DUP  >sign >R                 ( R: sign-flag )  \ save string start, process sign
 *       0 0 2SWAP >NUMBER              ( S: c-addr u ud c-addr2 u2 ) \ convert string to integer
 *       3 PICK OVER = IF               \ if no digit converted (u=u2)
 *          2DROP 2DROP 2DROP R> 0= EXIT   \ return FALSE
 *       THEN
 *       [CHAR] . 1stchar=? IF          \ is there a dot ? 
 *          1 /STRING >NUMBER   TRUE       \ skip dot, convert rest of number, dot-flag=true
 *       ELSE FALSE	  	           \ else dot-flag=false
 *       THEN   R> 2>R                     ( R: dot-flag sign-flag )
 *       [CHAR] E 1stchar=? IF          \ is there an 'E' (floating point number) ?
 *          2R> 2DROP 2DROP 2DROP          \ clear stacks
 *          >FLOAT                         \ redo the conversion from start, using >FLOAT
 *          STATE @ OVER AND IF            \ if compilation state and conversion succeeded
 *             POSTPONE FLITERAL              \ create a float literal
 *          THEN   EXIT  
 *       THEN                           ( S: c-addr u ud c-addr2 u2 ) ( R: dot-flag sign-flag )
 *       2ROT 2DROP                     \ remove c-addr u from stack
 *       NIP IF 	 	        \ if string not fully converted (u2 <> 0):
 *          2R> 2DROP 2DROP FALSE EXIT     \ only return flag=FALSE
 *       THEN
 *       R> 0< IF DNEGATE THEN          \ process the sign flag  ( R: dot-flag )
 *       R> IF		 		\ there was a dot in the number:
 *          STATE @ IF POSTPONE 2LITERAL THEN       \ it's a double cell integer
 *       ELSE  			        \ else if there wasn't a dot:
 *          D>S  STATE @ IF POSTPONE LITERAL THEN   \ it's single cell integer
 *       THEN   TRUE ;                  \ flag=TRUE (success)
 */
INTERNAL_SEC_WORD (number, "number")
	XT(TWO_DUP)  XT(to_sign) XT(TO_R)
	XT(ZERO) XT(ZERO) XT(TWO_SWAP) XT(TO_NUMBER)
	LIT(3) XT(PICK) XT(OVER) XT(EQUALS) IF({
	   XT(TWO_DROP) XT(TWO_DROP) XT(TWO_DROP) XT(R_FROM) XT(ZERO_EQUALS) XT(EXIT)
	/*THEN*/})
	LIT('.') XT(firstchar_equals_question) IF({
	   XT(ONE) XT(SLASH_STRING) XT(TO_NUMBER)   XT(TRUE)    
	/*ELSE*/},{ XT(FALSE)
	/*THEN*/})   XT(R_FROM) XT(TWO_TO_R)
	LIT('E') XT(firstchar_equals_question) IF({
	   XT(TWO_R_FROM) XT(TWO_DROP) XT(TWO_DROP) XT(TWO_DROP)
	   XT(TO_FLOAT)
	   XT(STATE) XT(FETCH) XT(OVER) XT(_AND) IF({
	      XT(FLITERAL)
	   /*THEN*/})   XT(EXIT)
	/*THEN*/})
	XT(TWO_ROT) XT(TWO_DROP)
	XT(NIP) IF({
	   XT(TWO_R_FROM) XT(TWO_DROP) XT(TWO_DROP) XT(FALSE) XT(EXIT)
	/*THEN*/})
	XT(R_FROM) XT(ZERO_LESS) IF({ XT(DNEGATE) /*THEN*/})
        XT(R_FROM) IF({
	   XT(STATE) XT(FETCH) IF({ XT(TWO_LITERAL) /*THEN*/})
	/*ELSE*/},{
	   XT(D_TO_S)   XT(STATE) XT(FETCH) IF({ XT(LITERAL) /*THEN*/})
	/*THEN*/})   XT(TRUE) XT(EXIT)


/*     Number Output Conversion
 *=================================================================================*/

/*-------------------------------------------------------------------------------
 * hld			-- c-addr					CORE
 *
 * Pointer to the current character in the numeric output conversion buffer
 *
 * Definition:   0 VALUE hld
 */
INTERNAL_VALUE (hld, "hld", 0)

/********************************************************************************
 * <#			-- 						CORE
 *
 * Definition:   : <#   HERE 64 + TO hld ;
 */
SEC_WORD (LESS_NUMSIGN, "<#")	
	XT(HERE) XT(C_PER_L) XT(PLUS) TO(hld) XT(EXIT)

/********************************************************************************
 * HOLD			char --						CORE
 *
 * Definition:   : HOLD   hld 1- TO hld   hld HERE < -17 ?throw   hld C!;
 */
SEC_WORD (HOLD, "HOLD")
	XT(hld) XT(ONE_MINUS) TO(hld)   
        XT(hld) XT(HERE) XT(LESS_THAN) LIT(-17) XT(question_throw)   
        XT(hld) XT(C_STORE) XT(EXIT)
	
/********************************************************************************
 * #			ud1 -- ud2					CORE
 *
 * Definition:   
 *    : #
 *        0 BASE @ UM/MOD           ( division in two steps )
 *        >R BASE @ UM/MOD R>       ( S: remainder(digit) ud2 )
 *        ROT [CHAR] 0 +            ( convert digit to character )
 *        DUP [CHAR] 9 >            ( if digit > '9' )
 *           -7 * +                    ( then character from 'A' )
 *        HOLD ;                    ( store ASCII character )
 */
SEC_WORD (NUMSIGN, "#")		
	XT(ZERO) XT(BASE) XT(FETCH) XT(UM_SLASH_MOD)
	XT(TO_R) XT(BASE) XT(FETCH) XT(UM_SLASH_MOD) XT(R_FROM)
	XT(ROT) LIT('0') XT(PLUS)
	XT(DUP) LIT('9') XT(GREATER_THAN)
	   LIT(-7) XT(STAR) XT(PLUS)
	XT(HOLD) XT(EXIT)

/********************************************************************************
 * #S			ud1 -- 0.					CORE
 *
 * Definition:   : #S   BEGIN # 2DUP D0= UNTIL ;
 */
SEC_WORD (NUMSIGN_S, "#S")	
	BEGIN_UNTIL({ XT(NUMSIGN) XT(TWO_DUP) XT(D_ZERO_EQUALS) })/*UNTIL*/ XT(EXIT)				
	
/********************************************************************************
 * SIGN			n -- 						CORE
 *
 * Definition:   : SIGN   0< IF  [CHAR] - HOLD THEN ;
 */
SEC_WORD (SIGN, "SIGN")		
		XT(ZERO_LESS) IF({  LIT('-') XT(HOLD) })/*THEN*/ XT(EXIT)

/********************************************************************************
 * #>			xd -- c-addr u					CORE
 *
 * Definition:   : #>   2DROP  hld  HERE 64 + OVER - ;
 */
SEC_WORD (NUMSIGN_GREATER, "#>")
	XT(TWO_DROP)  XT(hld)   XT(HERE) XT(C_PER_L) XT(PLUS) XT(OVER) XT(MINUS) XT(EXIT)
	
SEC_WORD (D_DOT_R, "D.R")		// ( d n -- )					DOUBLE
		XT(TO_R) XT(_SWAP) XT(OVER) XT(DABS)	// : D.R  >R SWAP OVER DABS  ( d n -- sign dabs )
		XT(LESS_NUMSIGN) XT(NUMSIGN_S) XT(ROT)  //    <# #S ROT SIGN #> ( c-addr u )
                XT(SIGN) XT(NUMSIGN_GREATER)
		XT(R_FROM) XT(OVER) XT(MINUS) XT(SPACES)//    R> OVER - SPACES
		XT(TYPE) XT(EXIT)			//    TYPE ;
SEC_WORD (DOT_R, ".R")			// ( n1 n2 -- )					CORE EXT
		XT(TO_R) XT(S_TO_D) XT(R_FROM)		// : .R  >R S>D R>  D.R ;
                XT(D_DOT_R) XT (EXIT)
SEC_WORD (D_DOT, "D.")			// ( d -- )					DOUBLE	
		XT(ZERO) XT(D_DOT_R) XT(SPACE) XT(EXIT)// : D.  0 D.R SPACE ;
SEC_WORD (DOT, ".")			// ( n -- )					CORE
		XT(S_TO_D) XT(D_DOT) XT(EXIT)		// : .  S>D D. ;
SEC_WORD (U_DOT_R, "U.R")		// ( u n -- )	
		XT(ZERO) XT(_SWAP) XT(D_DOT_R) XT(EXIT)	// : U.R  0 SWAP D.R
SEC_WORD (U_DOT, "U.")			// ( u -- )					CORE
		XT(ZERO) XT(D_DOT) XT(EXIT)		// : U.  0 D. ;
SEC_WORD (QUESTION, "?")		// ( a-addr -- )				TOOLS
		XT(FETCH) XT(DOT) XT(EXIT)		// : ?  @ . ;
SEC_WORD (DECIMAL, "DECIMAL")		// ( -- )					CORE
		LIT(10) XT(BASE) XT(STORE) XT(EXIT)	// : DECIMAL  10 BASE ! ;
SEC_WORD (HEX, "HEX")			// ( -- )					CORE EXT
		LIT(16) XT(BASE) XT(STORE) XT(EXIT)


/*==================================================================================
       String Routines
  ==================================================================================**
   ***********************************************************************************/

/********************************************************************************
 * COMPARE		c-addr1 u1 c-addr2 u2 -- n			STRING
 */
PRIM_WORD (COMPARE, "COMPARE")
		move.w	(PSP)+, d1			// d1 = u2
		move.w	(PSP)+, a1			// a1 = c-addr2
		move.w	(PSP)+, d0			// d0 = u1
		move.w	(PSP)+, a0			// a0 = c-addr1
		adda.l	BP, a0				// calculate the absolute addresses
		adda.l	BP, a1
		moveq	#-1, d2				// d2 = n

\Loop:		tst.w	d1
		beq	\Str2End			// (str1 > str2) or (str1 = str2)
		subq.w	#1, d1
		subq.w	#1, d0
		bmi	\Minus				// (str1 < str2)
		cmpm.b	(a1)+, (a0)+
		blt	\Minus
		bgt	\Plus
		bra	\Loop

\Str2End:	cmp.w	d1, d0
		beq	\Zero

\Plus:		addq.w	#1, d2
\Zero:		addq.w	#1, d2
\Minus:		move.w	d2, -(PSP)
		jmp	(NP)

/*-------------------------------------------------------------------------------
 * slit			-- c-addr u					internal
 *
 * Is compiled by SLITERAL. Return the address of the literal string following
 * slit in threaded code and skip the string.
 *
 * Definition:
 *    : slit   2R> A>R COUNT         ( get address of string )
 *       2DUP + ALIGNED R>A 2>R ;    ( set return address to the address, following the string)
 */
INTERNAL_SEC_WORD (slit, "slit")
			XT(TWO_R_FROM) XT(A_TO_R) XT(COUNT)   XT(TWO_DUP)
__Skip_SLITERAL:	XT(PLUS) XT(ALIGNED) XT(R_TO_A) XT(TWO_TO_R) XT(EXIT)

/********************************************************************************
 * SLITERAL		Compilation: c-addr1 u -- 			STRING
 *                      Run-time:              -- c-addr2 u
 * Definition:
 *    : SLITERAL
 *       compile slit                          ( compile SLITERAL handler )
 *       DUP C,  HERE SWAP                     ( write length-count, S: src-addr dst-addr u )
 *       DUP ALLOT MOVE  ALIGN ;  IMMEDIATE    ( copy the string into dictionary )
 */
SEC_WORD (SLITERAL, "SLITERAL", IMMEDIATE)
		XT(compile) XT(slit)
__Do_SLITERAL:	XT(DUP) XT(C_COMMA)  XT(HERE) XT(_SWAP)
		XT(DUP) XT(ALLOT) XT(_MOVE)  XT(ALIGN) XT(EXIT)

/********************************************************************************
 * S"			Compilation:    "ccc<quote>" -- 		STRING
 *                      Run-time:                    -- c-addr u
 *                      Interpretation: "ccc<quote>" -- c-addr u
 * Definition:
 *    : S"
 *       [CHAR] " PARSE                  ( read the string, S: c-addr u )
 *       STATE @ IF                      ( in compilation state )
 *          POSTPONE SLITERAL EXIT          ( create a literal)
 *       THEN                            ( in interpretation state)
 *       s"-buffer SWAP 80 MIN              ( copy the string to the temporary string buffer )
 *       2DUP 2>R MOVE 2R> ; IMMEDIATE      ( and return c-addr u )
 */
SEC_WORD (S_QUOTE, 'S"', IMMEDIATE)	// ( "ccc<quote>" -- )				CORE
	LIT('"') XT(PARSE)
	XT(STATE) XT(FETCH) IF({
	   XT(SLITERAL) XT(EXIT)
	})/*EXIT*/
	XT(s_quote_buffer) XT(_SWAP) LIT(80) XT(MIN)
	XT(TWO_DUP) XT(TWO_TO_R) XT(_MOVE) XT(TWO_R_FROM) XT(EXIT)

/********************************************************************************
 * /STRING		c-addr1 u1 n -- c-addr2 u2			STRING
 */
PRIM_WORD (SLASH_STRING, "/STRING")
		move.w	(a7)+, d0	// d0 = n
		sub.w	d0, (a7)	// decrease u by n
		add.w	d0, 2(a7)	// increase addr by n
		jmp	(NP)

/********************************************************************************
 * SEARCH		c-addr1 u1 c-addr2 u2 -- c-addr3 u3 flag	STRING
 */
PRIM_WORD (SEARCH, "SEARCH")
		move.w	(PSP)+, d1	// d1 = u2
		movea.w	(PSP)+, a1	// a1 = c-addr2
		adda.l	BP, a1
		move.w	(PSP), d0	// d0 = u1
		movea.w	2(PSP), a0	// a0 = c-addr1
		clr.w	-(PSP)		// push flag=false ( S: c-addr1 u1 false )
		adda.l	BP, a0

		sub.w	d1, d0		// d0 = u1-u2-1 = counter1
		bmi	\NotFound
		subq.w	#1, d1		// d1 = u2-1 = counter2
		bmi	\NotFound

\Loop:				// compare string at current position with c-addr2 u2
		movem.l	a0/a1/d1, -(a7)
\__CmpLoop:	cmpm.b	(a0)+, (a1)+
		dbne	d1, \__CmpLoop
		movem.l	(a7)+, a0/a1/d1
		bne	\__NoMatch

				// occurence found!
		suba.l	BP, a0		// a0 = c-addr3
		move.w	a0, 4(PSP)	// ( S: c-addr3 u1 false )
		addq.w	#1, d0		// d0 = u3
		add.w	d1, d0
		move.w	d0, 2(PSP)	// ( S: c-addr3 u3 false )
		not.w	(PSP)		// ( S: c-addr3 u1 true )
		jmp	(NP)
		
\__NoMatch:	addq.l	#1, a0		// goto next character
		dbra	d0, \Loop

\NotFound:	jmp	(NP)		// no occurence found ( s: c-addr1 u1 false )
		

/********************************************************************************
 * BLANK		c-addr u -- 					STRING
 *
 * Definition:    : BLANK  BL FILL ;
 */
SEC_WORD (BLANK, "BLANK")
		XT(BL) XT(FILL) XT(EXIT)		

/********************************************************************************
 * -TRAILING		c-addr u1 -- c-addr u2				STRING
 */
PRIM_WORD (DASH_TRAILING, "-TRAILING")
		movea.w	(PSP), a0		// a0 = address of end of string plus 1
		adda.w	2(PSP), a0
		adda.l	BP, a0
\Loop:		cmpi.b	#' ', -(a0)		// is there a space at the string's end ?
		bne	\End			// if there isn't: exit
		subq.w	#1, (PSP)		// else: decrease u2 
		bne	\Loop			// and continue if string-length > 0
\End:		jmp	(NP)

/********************************************************************************
 * CMOVE		c-addr1 c-addr2 u -- 				STRING
 * CMOVE>
 */
PrepareMove:	move.l	(a7)+, -(RSP)			// save return-address

		move.w	(PSP)+, d0			// d0 = counter
		bmi	\IllegalAddr			// more than 32767 bytes to move -> error
		move.w	(PSP)+, d2			// d2 = dest-addr
		bmi	\IllegalAddr			// negative dest address -> error
		move.w	d2, d1
		add.w	d0, d1				// d1 = end-of-dest-addr
		chk	#RELP(AccessBoundary), d1	// check that address
		move.w	(PSP)+, d1			// d1 = source-addr
		
		lea	0(BP, d1.w), a0			// a0 = source-addr
		lea	0(BP, d2.w), a1			// a1 = dest-addr
		
		move.l	(RSP)+, -(a7)			// restore return-address
		rts
		
\IllegalAddr:	moveq	#ERR_INV_ADDRESS, d0
		bra	DoThrow
		
PRIM_WORD (CMOVE, "CMOVE")
		bsr	PrepareMove
		subq.w	#1, d0				// d0 = counter
		bmi	\End			
\Loop:		move.b	(a0)+, (a1)+
		dbra	d0, \Loop
		
\End:		jmp	(NP)

PRIM_WORD (CMOVE_UP, "CMOVE>")
		bsr	PrepareMove
		adda.w	d0, a0				// a0 = end-of-source-address
		adda.w	d0, a1				// a1 = end-of-dest-address
		subq.w	#1, d0				// d0 = counter
		bmi	\End			

\Loop:		move.b	-(a0), -(a1)
		dbra	d0, \Loop	
		
\End:		jmp	(NP)


/*==================================================================================
       Programming Tools
  ==================================================================================**
   ***********************************************************************************/

/********************************************************************************
 * .S			--		 				TOOLS
 *
 * Definition:
 *    : .S  ( -- )
 *       DEPTH 0 DO
 *          ." tab"  DEPTH I - 1- PICK .
 *       LOOP ;
 */
SEC_WORD (DOT_S, ".S")		
	XT(DEPTH) XT(ZERO) QDO_LOOP({
	   STR(CHR_TAB)  XT(DEPTH) XT(I) XT(MINUS) XT(ONE_MINUS) XT(PICK) XT(DOT)
	})/*LOOP*/ XT(EXIT)

/********************************************************************************
 * DUMP			c-addr c-addr2 u -- 				TOOLS
 *
 * Definition:
 *   : DUMP  ( c-addr u )
 *      OVER + SWAP ?DO             ( from c-addr to c-addr+u do: )
 *         BASE @  HEX                 ( store old base, new base is 16 )
 *         CR I  5 .R SPACE            ( output current address )
 *         I 8 + I  2DUP DO            ( from c-addr to c-addr+8 do: )
 *            I C@ 0 <# # # #>             ( output one byte as hex, 2-digit number )  
 *            TYPE SPACE  
 *         LOOP
 *         DO                          ( from c-addr to c-addr+8 do (again): )
 *            I C@ DUP BL < IF         ( if byte is not a valid character )
 *               DROP 183                 ( output a dot )
 *            THEN  EMIT               ( else output byte as character )
 *         LOOP
 *         BASE !                      ( restore the base )
 *      8 +LOOP ;                   ( display the next 8 bytes )
 */
SEC_WORD (DUMP, "DUMP")	
	XT(OVER) XT(PLUS) XT(_SWAP) QDO_PLOOP({
 	   XT(BASE) XT(FETCH)  XT(HEX)
	   XT(CR) XT(I)  LIT(5) XT(DOT_R)  XT(SPACE)
           XT(I) LIT(8) XT(PLUS) XT(I)  XT(TWO_DUP) DO_LOOP({
              XT(I) XT(C_FETCH) XT(ZERO) XT(LESS_NUMSIGN) XT(NUMSIGN) XT(NUMSIGN) XT(NUMSIGN_GREATER) 
              XT(TYPE) XT(SPACE)
	   })/*LOOP*/
	   DO_LOOP({ 
	      XT(I) XT(C_FETCH) XT(DUP) XT(BL) XT(LESS_THAN) IF({
	         XT(DROP) LIT(183)
              })/*THEN*/  XT(EMIT)
	   })/*LOOP*/
	   XT(BASE) XT(STORE)
        LIT(8) })/*+LOOP*/
	XT(EXIT)

/*==================================================================================
       Error Checking and Reporting
  ==================================================================================**
   ***********************************************************************************/

ExceptHandler:	dc.l	0	// address of current exception frame (on return stack)

/********************************************************************************
 * CATCH		   i*x xt -- j*x 0 | i*x n			EXCEPTION
 */
PRIM_WORD (CATCH, "CATCH")	
		/** push exception frame */
		move.l	IP, -(RSP)			// save instruction pointer
		move.l	ExceptHandler(PC), -(RSP)	// save previous exception handler
		move.l	PSP, -(RSP)			// save parameter stack pointer
		move.l	FSP, -(RSP)			// save floating stack pointer
		bsr	InputToR			// save input source specification
		move.l	RSP, RELP(ExceptHandler)(BP)	// create new Exception Handler

		/** execute xt */
		lea	\ExecCode, IP			// load code to execute
		move.w	(PSP)+, (IP)			// set code to execute to xt
		jmp	(NP)				// execute it
\ExecCode:	dc.w	0				// here 'xt' is inserted
		XT(\ExecEnd)

		/** restore previous exception handler, return */
\ExecEnd:	XT(\ExecEnd+2)
		lea	SIZEOF_RS_ISS+8(RSP), RSP	// drop exception frame
		move.l	(RSP)+, RELP(ExceptHandler)(BP)	// restore previous exception handler
		move.l	(RSP)+, IP			// restore instruction pointer
		clr.w	-(PSP)				// push 0
		jmp	(NP)

/********************************************************************************
 * THROW		   k*x n -- k*x	| i*x n				EXCEPTION
 *                      R:   l*x -- l*x | j*x
 */
PRIM_WORD (THROW, "THROW")
		move.w	(PSP)+, d0			// d0=n is our exception number
		beq	__THROW_Exit			// if n = 0: do nothing
DoThrow:	move.l	ExceptHandler(PC), d1		// is there an exception handler?
		beq	ExitFORTH			//   if there isn't: exit FORTH
		move.l	d1, RSP				// get exception frame, restore RSP
		bsr	RToInput			// restore input source specification
		move.l	(RSP)+, FSP			// restore FSP
		move.l	(RSP)+, PSP			// restore PSP
		move.l	(RSP)+, RELP(ExceptHandler)(BP)	// restore previous exception handler
		move.w	d0, -(PSP)			// push exception number
		lea	question_reload+2(PC), IP	// ?reload input source
__THROW_Exit:	jmp	(NP)				// ?reload returns to caller of 'CATCH'
							// since return address is still on stack

/*-------------------------------------------------------------------------------
 * ?throw	    	   k*x flag n -- k*x | i*x n			internal
 *			R: 	  l*x -- l*x | j*x
 *
 * If flag is true, trigger a THROW n exception. Else drop flag and n and 
 * continue execution.
 *
 * Definition:   : ?throw   SWAP 0<> AND THROW ;
 */
INTERNAL_SEC_WORD (question_throw, "?throw")	
	XT(_SWAP) XT(ZERO_NOT_EQUALS) XT(_AND) XT(THROW) XT(EXIT)

/*-------------------------------------------------------------------------------
 * ?pairs	    	   k*x x1 x2 -- k*x | i*x n			internal
 *			R: 	 l*x -- l*x | j*x
 *
 * Trigger a "control structure mismatch" exception if x1 <> x2.
 *
 * Definition:   : ?throw   <> -22 ?throw ;
 */
INTERNAL_SEC_WORD (question_pairs, "?pairs")
	XT(NOT_EQUALS) LIT(-22) XT(question_throw) XT(EXIT)

/*-------------------------------------------------------------------------------
 * disp-exc		    	n --   					internal
 *
 * Display an error message for exception n and the place in the input source
 * where it occured. If there is a file "main\fortherr" look for a textual error 
 * message in it.
 *
 * Definition:
 *    : disp-exc
 *       CR                   
 *       interpreted 2@ parsed 2@      ( n c-addr1 u1 c-addr2 u2 )
 *       OVER 4 PICK TUCK - TYPE       \ type string till exceptional word  
 *       2DUP UL TYPE UL	       \ type exceptional word
 *       +   2 PICK -   /STRING        \ type string behind exceptional word
 *       64 MIN TYPE                   
 *       ." \nError: "
 *       S" main\fortherr" R/O OPEN-FILE 0= IF      ( n fileid )
 *          BEGIN   
 *             DUP HERE UNUSED ROT READ-LINE        ( n fileid u flag ior )
 *          DROP WHILE                              ( n fileid u )
 *             HERE 3   4 PICK NEGATE S>D           ( n fileid u c-addr u d )
 *             <# # # # #> COMPARE 0= IF            ( n fileid u )
 *                HERE SWAP 4 /STRING TYPE          \ print text, describing the exception
 *                2DROP EXIT                        ( )
 *             THEN   DROP                          ( n fileid u )
 *          REPEAT                                  ( n fileid u )
 *          DROP                                    ( n fileid )
 *       THEN
 *       DROP                                       ( n )
 *       ." Exception #" . ;   		            \ just pring exception number
 */
INTERNAL_SEC_WORD (disp_exc, "disp-exc")
 	XT(CR) 
	XT(interpreted) XT(TWO_FETCH) XT(parsed) XT(TWO_FETCH)
	XT(OVER) XT(FOUR) XT(PICK) XT(TUCK) XT(MINUS) XT(TYPE)
	XT(TWO_DUP) XT(UL) XT(TYPE) XT(UL)
	XT(PLUS)   XT(TWO) XT(PICK) XT(MINUS)   XT(SLASH_STRING) 
	XT(C_PER_L) XT(MIN) XT(TYPE)
	STR(CHR_LF,"Error: ") 
	SLIT("main\fortherr") XT(R_O) XT(OPEN_FILE) XT(ZERO_EQUALS) IF({
	   BEGIN_WHILE({
	      XT(DUP) XT(HERE) XT(UNUSED) XT(ROT) XT(READ_LINE)
	   XT(DROP) /*WHILE*/},{
	      XT(HERE) LIT(3)   XT(FOUR) XT(PICK) XT(NEGATE) XT(S_TO_D) 
	      XT(LESS_NUMSIGN) XT(NUMSIGN) XT(NUMSIGN) XT(NUMSIGN) XT(NUMSIGN_GREATER) XT(COMPARE) XT(ZERO_EQUALS) IF({
	         XT(HERE) XT(_SWAP) XT(FOUR) XT(SLASH_STRING) XT(TYPE) 
		 XT(TWO_DROP) XT(EXIT)
	      /*THEN*/}) XT(DROP)
	   /*REPEAT*/})
	   XT(DROP)
	/*THEN*/})
	XT(DROP)
	STR("Exception #") XT(DOT) XT(EXIT)
	

/*==================================================================================
       Foating point routines
  ==================================================================================**
   ***********************************************************************************/

/**
 * Some info:
 *    Floating point numbers are stored onto a separate floating point stack, which
 *    is addressed by the address register constant 'FSP'. Floating point numbers have
 *    48 bit. 32 bits are mantisse, 16 bits are used for the exponent. In memory they
 *    are stored in the following way:
 *        0000.l (4 bytes)  mantissse
 *        0004.w (2 bytes)  exponent  
 *    On the floating point stack they are stored "reversed":
 *        FSP-> (2 bytes):   exponent
 *              (4 bytes):   mantisse
 *    Since the floating point stack isn't very big and there is only few overflow
 *    safety, every floating point routine checks for floating point stack over- and
 *    underflows. Since floating point numbers are slow to process anyway, I don't think
 *    that this makes it much slower.
 */

/*---------------------------------------
 * Helper routines DoFloatOp1 and DoFloatOp2: 
 *    Call floting point operation with one/two arguments pointed by a0, do float 
 *    stack check and jump to Next
 */
DoFloatOp2:	move.w	(FSP)+, d3		// load operand2 exponent
		move.l	(FSP)+, d2		// load operand2 mantisse
DoFloatOp1:	move.w	(FSP)+, d1		// load operand1 exponent
		move.l	(FSP)+, d0		// load operand1 mantisse
		jsr	(a0)			// do floating point operation
		move.l	d0, -(FSP)		// push result mantisse
		move.w	d1, -(FSP)		// push result exponent
		// stack check just follows ...

/*---------------------------------------
 * Helper routine: check for floating point stack over- and underflow
 * and jump to Next
 */
CheckFSP:	move.l	FSB, d0	
		sub.l	FSP, d0
		chk	#FS_SIZE*6, d0
		jmp	(NP)

/*    Floating point conversion routintes
 *=================================================================================*/

/********************************************************************************
 * SET-PRECISION	   u -- 					FLOATING EXT
 * 
 * Definition:   : SET-PRECISION   1 MAX 9 MIN TO PRECISION ;
 */
SEC_WORD (SET_PRECISION, "SET-PRECISION")
	XT(ONE) XT(MAX) LIT(9) XT(MIN) TO(PRECISION) XT(EXIT)

/********************************************************************************
 * D>F			   d -- 					FLOATING
 *			F:   --	r	
 */
PRIM_WORD (D_TO_F, "D>F")
		move.l	(PSP)+, d0
		jsr	forthlib::IntToFloat
		move.l	d0, -(FSP)
		move.w	d1, -(FSP)
		bra	CheckFSP
		
/********************************************************************************
 * F>D			      -- d					FLOATING
 *			F: r  --	
 */
PRIM_WORD (F_TO_D, "F>D")
		move.w	(FSP)+, d1
		move.l	(FSP)+, d0
		jsr	forthlib::FloatToInt
		move.l	d0, -(PSP)
		bra	CheckFSP

/********************************************************************************
 * REPRESENT		   c-addr u -- n flag1 flag2			FLOATING
 *			F:        r --	
 * Definition:
 *    : REPRESENT
 *       2>R                               ( R: c-addr u )
 *       FDUP F0<   DUP IF FNEGATE THEN    ( S: flag1 )
 *       FDUP FLOG F>D D>S 1+              ( S: flag1 n ) 
 *       DUP f10** F/                      \ divide r so that .1 <= r < 1 (comma to the left) 
 *       R@ f10** F*  FROUND 0.4E F+ R@ f10** F/  \ round r
 *       FDUP 1.E F< 0= IF                 \ adjust exponent and magnitude to rounded r
 *          10E F/   1+
 *       THEN
 *       2R> OVER + SWAP ?DO                  \ output string representation of number
 *          10E F* FDUP F>D D>S               ( S: flag1 n digit )    
 *          [CHAR] 0 + I C!                   \ store digit to destination string
 *          FDUP FLOOR F-                     \ remove digit from number     
 *       LOOP 
 *       FDROP SWAP TRUE ;                 ( S: n flag1 flag2 )
 */
SEC_WORD (REPRESENT, "REPRESENT")
	XT(TWO_TO_R) 
	XT(FDUP) XT(F_ZERO_LESS)   XT(DUP) IF({ XT(FNEGATE) })/*THEN*/
	XT(FDUP) XT(FLOG) XT(F_TO_D) XT(D_TO_S) XT(ONE_PLUS)
	XT(DUP) XT(f_10_star_star) XT(F_SLASH)
	XT(R_FETCH) XT(f_10_star_star) XT(F_STAR)  XT(FROUND)  FLIT($66666666,-32) XT(F_PLUS) XT(R_FETCH) XT(f_10_star_star) XT(F_SLASH)
	XT(FDUP) FLIT($40000000,-30) XT(F_LESS_THAN) XT(ZERO_EQUALS) IF({
	   XT(TEN_E) XT(F_SLASH)   XT(ONE_PLUS)
	})/*THEN*/
	XT(TWO_R_FROM) XT(OVER) XT(PLUS) XT(_SWAP) QDO_LOOP({
	   XT(TEN_E) XT(F_STAR) XT(FDUP) XT(F_TO_D) XT(D_TO_S)
	   LIT('0') XT(PLUS) XT(I) XT(C_STORE)
	   XT(FDUP) XT(FLOOR) XT(F_MINUS)
	})/*LOOP*/
	XT(FDROP) XT(_SWAP) XT(TRUE) XT(EXIT) 

/*-------------------------------------------------------------------------------
 * fzeroes	   	u --						internal
 *
 * Definition:   : fzeroes   0 ?DO ." 0" LOOP ;
 */
INTERNAL_SEC_WORD (fzeroes, "fzeroes")
	XT(ZERO) QDO_LOOP({ STR("0") })/*LOOP*/ XT(EXIT)

/*-------------------------------------------------------------------------------
 * fcomma	   	--						internal
 *
 * Definition:   : fcomma  ." ." ;
 */
INTERNAL_SEC_WORD (fcomma, "fcomma")
	STR(".") XT(EXIT)

/*-------------------------------------------------------------------------------
 * ftype	   	c-addr u --					internal
 *
 * Print a string of digits, without trailing zeroes.
 *
 * Definition:   
 *    : ftype
 *       BEGIN 
 *          2DUP + 1- C@ [CHAR] 0 =  OVER AND    \ while end of string is zero and u <> 0
 *       WHILE 1- REPEAT                         \ decrease string length
 *       TYPE ;
 */
INTERNAL_SEC_WORD (ftype, "ftype")
	BEGIN_WHILE({
	   XT(TWO_DUP) XT(PLUS) XT(ONE_MINUS) XT(C_FETCH) LIT('0') XT(EQUALS)  XT(OVER) XT(_AND)
	},{/*WHILE*/ XT(ONE_MINUS) })/*REPEAT*/
	XT(TYPE) XT(EXIT)

/*-------------------------------------------------------------------------------
 * fconvert	   	      -- n					internal
 *			F:  r --
 * Definition:   
 *    : fconvert  
 *       HERE PRECISION REPRESENT DROP   \ convert floating point number, dest is HERE
 *       IF ." -" THEN ;                 \ process sign flag of REPRESENT
 */
INTERNAL_SEC_WORD (fconvert, "fconvert")
	XT(HERE) XT(PRECISION) XT(REPRESENT) XT(DROP) 
	IF({ STR("-") })/*THEN*/ XT(EXIT)
	
/********************************************************************************
 * F.		   	F: r --						FLOATING EXT
 *
 * Definition:
 *    : F.   
 *       fconvert                         \ convert and print sign
 *       DUP 0> IF                        \ number greater or equal 1.00E0
 *          HERE OVER PRECISION MIN TYPE     \ print significant digits before comma
 *          HERE PRECISION ROT /STRING DUP 0> IF \ digits behind or before the comma missing?
 *             fcomma ftype                      \ print digits behind comma
 *          ELSE                             \ digits before the comma missing:
 *             NEGATE fzeroes fcomma  DROP       \ fill with zeroes, print comma
 *          THEN
 *       ELSE                             \ number less than 1.00E0:
 *          ." 0." NEGATE fzeroes             \ print "0." and zeroes till significant digits
 *          HERE PRECISION ftype              \ print significant digits
 *       THEN
 *       SPACE ;                          \ trailing space
 *
 */
SEC_WORD (F_DOT, "F.")
	XT(fconvert)
	XT(DUP) XT(ZERO_GREATER) IF({
	   XT(HERE) XT(OVER) XT(PRECISION) XT(MIN) XT(TYPE)
	   XT(HERE) XT(PRECISION) XT(ROT) XT(SLASH_STRING) XT(DUP) XT(ZERO_GREATER) IF({
	      XT(fcomma) XT(ftype)
	   /*ELSE*/},{
	      XT(NEGATE) XT(fzeroes) XT(fcomma) XT(DROP)
	   /*THEN*/})
	/*ELSE*/},{
	   STR("0.") XT(NEGATE) XT(fzeroes)
	   XT(HERE) XT(PRECISION) XT(ftype)
	/*THEN*/})
	XT(SPACE) XT(EXIT)

/********************************************************************************
 * FS.		   	F: r --						FLOATING EXT
 *
 * Definition:
 *    : FS.   
 *       fconvert                         \ convert and print optional sign
 *       HERE 1 TYPE fcomma               \ print digit before comma
 *       HERE PRECISION 1 /STRING ftype   \ print digits behind comma
 *       1- ." E" . ;                     \ print exponent and trailing space
 */
SEC_WORD (FS_DOT, "FS.")
	XT(fconvert) 
	XT(HERE) XT(ONE) XT(TYPE) XT(fcomma)
	XT(HERE) XT(PRECISION) XT(ONE) XT(SLASH_STRING) XT(ftype)
	XT(ONE_MINUS) STR("E") XT(DOT) XT(EXIT)
	
/********************************************************************************
 * FE.		   	F: r --						FLOATING EXT
 *
 * Definition:
 *    : FE.   
 *       HERE 4 [CHAR] 0 FILL fconvert          \ convert and print optional sign
 *       HERE OVER  1- 3 MOD 1+  TUCK TYPE      \ print digit(s) before comma
 *       DUP HERE PRECISION ROT /STRING 0 MAX   \ get string of digits behind comma
 *       fcomma ftype                           \ print comma and digits behind
 *       - ." E" . ;		  	        \ print exponent and trailing space
 */
SEC_WORD (FE_DOT, "FE.")
	XT(HERE) XT(FOUR) LIT('0') XT(FILL) XT(fconvert)
	XT(HERE) XT(OVER)  XT(ONE_MINUS) LIT(3) XT(MOD) XT(ONE_PLUS)  XT(TUCK) XT(TYPE)
	XT(DUP) XT(HERE) XT(PRECISION) XT(ROT) XT(SLASH_STRING) XT(ZERO) XT(MAX)
	XT(fcomma) XT(ftype)
	XT(MINUS) STR("E") XT(DOT) XT(EXIT)

/*-------------------------------------------------------------------------------
 * >float   		   c-addr1 u1 -- c-addr2 u2			FLOATING
 *			F:            -- r 
 * Convert numbers from string c-addr1 u1 to an integral floating point number.
 * >float is almost a floating point equivalent to >NUMBER. Return the remaining
 * string at the first character which is not 0-9.
 *
 * Definition:
 *    : >float
 *       0 S>D D>F                         ( R: r )
 *       BEGIN  
 *          OVER C@ [CHAR] 0 -             ( S: c-addr u digit )
 *          2DUP 10 U< AND                 ( S: c-addr u digit continue?-flag )
 *       WHILE
 *          10E F*   S>D D>F F+            \ add digit to floating point number r
 *          1 /STRING                      \ go to next character
 *       REPEAT
 *       DROP ;
 */
INTERNAL_SEC_WORD  (to_float, ">float")
	XT(ZERO) XT(S_TO_D) XT(D_TO_F)
	BEGIN_WHILE({
	   XT(OVER) XT(C_FETCH) LIT('0') XT(MINUS)
	   XT(TWO_DUP) LIT(10) XT(U_LESS_THAN) XT(_AND)
	/*WHILE*/},{
	   XT(TEN_E) XT(F_STAR)   XT(S_TO_D) XT(D_TO_F) XT(F_PLUS)
	   XT(ONE) XT(SLASH_STRING)
	/*REPEAT*/})
	XT(DROP) XT(EXIT)
  	   
/********************************************************************************
 * >FLOAT   		   c-addr u -- true | false			FLOATING
 *			F:          -- r    |
 * Definition:
 *    : >FLOAT
 *       2DUP -TRAILING NIP IF               \ string doesn't consist of spaces only ?
 *          >sign >R                             \ get sign  ( R: sign-flag ) 
 *          TUCK >float   ROT OVER = >R          \ get <digits>   ( R: sign-flag invalid-flag )
 *          [CHAR] . 1stchar=? IF                \ is there a comma ?
 *             1 /STRING   TUCK >float              \ skip comma, read <digits0>
 *             ROT OVER -   f10** F/   F+           \ add fractional part to float number
 *          THEN
 *	    [CHAR] E 1stchar=? >R  [CHAR] e 1stchar=? >R
 *	    [CHAR] D 1stchar=? >R  [CHAR] d 1stchar=? 
 *	    R> 2R> OR OR OR IF			     \ is there an 'E' - char ?
 *             1 /STRING                             \ skip 'E' chars
 *	    THEN
 *	    >sign >R                             \ read sign of exponent, if exists
 *          0 0 2SWAP >NUMBER 2SWAP              \ read exponent, if exists
 *          D>S R> *   f10** F*                  \ adjust float number to exponent
 *          NIP R> OR IF                         \ is the floating point number invalid ?
 *             R> 0= FDROP EXIT                  \ return false
 *          THEN
 *          R> S>D D>F F* 
 *       ELSE				      \ if only spaces in string: return r=zero 
 *          2DROP 0 S>D D>F               
 *       THEN   TRUE ;                        \ return true
 */
SEC_WORD (TO_FLOAT, ">FLOAT")
	XT(TWO_DUP) XT(DASH_TRAILING) XT(NIP) IF({
	   XT(to_sign) XT(TO_R)
	   XT(TUCK) XT(to_float)   XT(ROT) XT(OVER) XT(EQUALS) XT(TO_R)
	   LIT('.') XT(firstchar_equals_question) IF({
	      XT(ONE) XT(SLASH_STRING)   XT(TUCK) XT(to_float)
	      XT(ROT) XT(OVER) XT(MINUS)   XT(f_10_star_star) XT(F_SLASH)   XT(F_PLUS)
	   /*THEN*/})
	   LIT('E') XT(firstchar_equals_question) XT(TO_R)   LIT('e') XT(firstchar_equals_question) XT(TO_R)
	   LIT('D') XT(firstchar_equals_question) XT(TO_R)   LIT('d') XT(firstchar_equals_question)
 	   XT(R_FROM) XT(TWO_R_FROM) XT(_OR) XT(_OR) XT(_OR) IF({
	      XT(ONE) XT(SLASH_STRING)
	   /*THEN*/})
	   XT(to_sign) XT(TO_R)
	   XT(ZERO) XT(ZERO) XT(TWO_SWAP) XT(TO_NUMBER) XT(TWO_SWAP)
	   XT(D_TO_S) XT(R_FROM) XT(STAR)   XT(f_10_star_star) XT(F_STAR)
	   XT(NIP) XT(R_FROM) XT(_OR) IF({
	      XT(R_FROM) XT(ZERO_EQUALS) XT(FDROP) XT(EXIT)
	   })/*THEN*/
	   XT(R_FROM) XT(S_TO_D) XT(D_TO_F) XT(F_STAR)
	/*ELSE*/},{
	   XT(TWO_DROP) XT(ZERO) XT(S_TO_D) XT(D_TO_F)
	/*THEN*/})   XT(TRUE) XT(EXIT)
	
	
/*    Floating point access
 *=================================================================================*/

/********************************************************************************
 * FLOATS		n1 -- n2					FLOATING
 *
 * Definition:   : FLOATS   6 * ;
 */
SEC_WORD (FLOATS, "FLOATS")
	LIT(6) XT(STAR) XT(EXIT)

/********************************************************************************
 * FLOAT+		f-addr1 -- f-addr2				FLOATING
 *
 * Definition:   : FLOAT+   6 + ;
 */
SEC_WORD (FLOAT_PLUS, "FLOAT+")
	LIT(6) XT(PLUS) XT(EXIT)

/********************************************************************************
 * FALIGNED		addr -- f-addr					FLOATING
 *
 * Definition:   : FALIGNED   ALIGNED ;
 */
SAME_PRIM_WORD (FALIGNED, "FALIGNED", ALIGNED)

/********************************************************************************
 * FALIGN		--						FLOATING
 *
 * Definition:   : FALIGNED   ALIGN ;
 */
SEC_WORD (FALIGN, "FALIGN")
	XT(ALIGN) XT(EXIT)

/********************************************************************************
 * F!			   f-addr -- 					FLOATING
 *			F:      r --		
 */
PRIM_WORD (F_STORE, "F!")
		move.w	(PSP)+, d0
		move.w	(FSP)+, 4(BP, d0.w)
		move.l	(FSP)+, 0(BP, d0.w)
		bra	CheckFSP

/********************************************************************************
 * F@			   f-addr -- 					FLOATING
 *			F:        -- r
 */
PRIM_WORD (F_FETCH, "F@")
		move.w	(PSP)+, d0
		move.l  0(BP, d0.w), -(FSP)
		move.w	4(BP, d0.w), -(FSP)
		bra	CheckFSP

/********************************************************************************
 * F,			F: r --						????
 *
 * Write a floating point value to the current data-space pointer, increase
 * data space pointer to point behind it.
 *
 * Definition:   : F,   HERE 1 FLOATS ALLOT F! ;
 */
SEC_WORD (F_COMMA, "F,")
	XT(HERE) XT(ONE) XT(FLOATS) XT(ALLOT) XT(F_STORE) XT(EXIT)

/*-------------------------------------------------------------------------------
 * flit			F: -- x						internal
 *
 * Is compiled by FLITERAL. Push the float-number pointed to by the current IP 
 * onto the data stack. Continue execution after that number
 */
INTERNAL_PRIM_WORD (flit, "flit")	
		move.l	(IP)+, -(FSP)
		move.w	(IP)+, -(FSP)
		bra	CheckFSP

/********************************************************************************
 * FLITERAL		compilation: (F: r --  )			FLOATING
 *			run-time:    (F:   -- r)
 * Definition:
 *    : FLITERAL   compile flit F, ; IMMEDIATE
 */
SEC_WORD (FLITERAL, "FLITERAL", IMMEDIATE)	
	XT(compile) XT(flit) XT(F_COMMA) XT(EXIT)

/********************************************************************************
 * FVARIABLE		"<spaces>name" --				FLOATING
 *
 * Definition:   : FVARIABLE   CREATE 1 FLOATS ALLOT ;
 */
SEC_WORD (_FVARIABLE, "FVARIABLE")	
	XT(CREATE)  XT(ONE) XT(FLOATS) XT(ALLOT) XT(EXIT)

/********************************************************************************
 * FCONSTANT 	  	"<spaces>name" --				FLOATING
 *			compilation: (F: r --  )
 *			run-time:    (F:   -- r)
 * (possible) Definition:   
 *    : FCONSTANT   CREATE F,   DOES> F@ ;
 */
SEC_WORD (_FCONSTANT, "FCONSTANT")	
	XT(CREATE) XT(F_COMMA)
	XT(semi_code)/* ;CODE */
DoFCon:		move.l  2(BP,WP.w), -(FSP)
		move.w	6(BP,WP.w), -(FSP)
		bra	CheckFSP

/*    Floating point stack routines
 *=================================================================================*/

/********************************************************************************
 * FDEPTH		-- +n						FLOATING
 */
PRIM_WORD (FDEPTH, "FDEPTH")
		move.l	FSB, d0
		sub.l	FSP, d0
		divu	#6, d0
		move.w	d0, -(PSP)
		jmp	(NP)

/********************************************************************************
 * FDROP		F: r --						FLOATING
 */
PRIM_WORD (FDROP, "FDROP")
		addq.l	#6, FSP
		bra	CheckFSP

/********************************************************************************
 * FDUP			F: r -- r r					FLOATING
 */
PRIM_WORD (FDUP, "FDUP")
		move.l	2(FSP), -(FSP)
		move.w	4(FSP), -(FSP)
		bra	CheckFSP

/********************************************************************************
 * FSWAP		F: r1 r2 -- r2 r1				FLOATING
 */
PRIM_WORD (FSWAP, "FSWAP")
		move.w	(FSP)+, d1
		move.l	(FSP)+, d0
		move.w	(FSP)+, d3
		move.l	(FSP)+, d2
		move.l	d0, -(FSP)
		move.w	d1, -(FSP)
		move.l	d2, -(FSP)
		move.w	d3, -(FSP)
		bra	CheckFSP

/********************************************************************************
 * FROT			F: r1 r2 r3 -- r2 r3 r1				FLOATING
 */
PRIM_WORD (FROT, "FROT")
		lea	12(FSP), FSP	
		move.w	(FSP)+, d1		// load r1
		move.l	(FSP)+, d0
		moveq	#2, d2			// shift r2 and r3 upwards 
\Loop:		move.l	-10(FSP), -(FSP)
		dbra	d2, \Loop
		move.l	d0, -(FSP)		// push r1 onto top of stack
		move.w	d1, -(FSP)
		bra	CheckFSP

/********************************************************************************
 * FOVER		F: r1 r2 -- r1 r2 r1				FLOATING
 */
PRIM_WORD (FOVER, "FOVER")
		move.l	8(FSP), -(FSP)
		move.w	10(FSP), -(FSP)
		bra	CheckFSP

/*-------------------------------------------------------------------------------
 * fp!			F: i*r --					internal
 *
 * Clear the floating point stack.
 */
INTERNAL_PRIM_WORD (fp_store, "fp!")	
		move.l	FSB, FSP	// initialize floating point stack pointer (FSP)
		jmp	(NP)

/*    Floating point arithmetic routines
 *=================================================================================*/
	
/********************************************************************************
 * F+			F: r1 r2 -- r3					FLOATING
 */
PRIM_WORD (F_PLUS, "F+")
		lea	forthlib::FloatAdd, a0
		bra	DoFloatOp2

/********************************************************************************
 * F-			F: r1 r2 -- r3					FLOATING
 */
PRIM_WORD (F_MINUS, "F-")
		lea	forthlib::FloatSub, a0
		bra	DoFloatOp2

/********************************************************************************
 * FNEGATE		F: r1 -- r2					FLOATING
 */
PRIM_WORD (FNEGATE, "FNEGATE")
		lea	forthlib::FloatNegate, a0
		bra	DoFloatOp1

/********************************************************************************
 * FLOOR		F: r1 -- r2					FLOATING
 */
PRIM_WORD (FLOOR, "FLOOR")
		lea	forthlib::FloatFloor, a0
		bra	DoFloatOp1

/********************************************************************************
 * FROUND		F: r1 -- r2					FLOATING
 */
PRIM_WORD (FROUND, "FROUND")
		lea	forthlib::FloatRound, a0
		bra	DoFloatOp1

/********************************************************************************
 * F*			F: r1 r2 -- r3					FLOATING
 */
PRIM_WORD (F_STAR, "F*")
		lea	forthlib::FloatMul, a0
		bra	DoFloatOp2

/********************************************************************************
 * F/			F: r1 r2 -- r3					FLOATING
 */
PRIM_WORD (F_SLASH, "F/")
		lea	forthlib::FloatDiv, a0
		bra	DoFloatOp2

/********************************************************************************
 * F<			         -- flag				FLOATING
 *			F: r1 r2 -- 				
 */
PRIM_WORD (F_LESS_THAN, "F<")
		move.w	(FSP)+, d3
		move.l	(FSP)+, d2
		move.w	(FSP)+, d1
		move.l	(FSP)+, d0
		jsr	forthlib::FloatLessThan
		move.w	d0, -(PSP)
		jmp	(NP)

/********************************************************************************
 * F0<			      -- flag					FLOATING
 *			F: r1 -- 				
 */
PRIM_WORD (F_ZERO_LESS, "F0<")
		addq.l	#2, FSP			// remove exponent
		move.l	(FSP)+, d0		// check whether mantisse is less than zero
		bra	__DO_LT__

/********************************************************************************
 * F0=			      -- flag					FLOATING
 *			F: r1 -- 				
 */
PRIM_WORD (F_ZERO_EQUALS, "F0=")
		addq.l	#2, FSP			// remove exponent
		move.l	(FSP)+, d0		// check whether mantisse is zero
		bra	__DO_EQ__

/********************************************************************************
 * FMIN			F: r1 r2 -- r3					FLOATING
 */
PRIM_WORD (FMIN, "FMIN")
		clr.w	-(PSP)			// return the lesser of r1 and r2
		bra	DoFloatMinMax

/********************************************************************************
 * FMAX			F: r1 r2 -- r3					FLOATING
 */
PRIM_WORD (FMAX, "FMAX")
		move.w	#-1, -(PSP)		// return the greater of r1 and r2

DoFloatMinMax:	move.w	(FSP)+, d3		// load r2
		move.l	(FSP)+, d2
		move.w	(FSP), d1		// load r1; r3 = r1
		move.l	2(FSP), d0
		jsr	forthlib::FloatLessThan	// r1 < r2 ?
		cmp.w	(PSP)+, d0		// set r3 = r2 depending on value pushed
		bne	\R1
		move.w	d3, (FSP)		// set r3 = r2
		move.l	d2, 2(FSP)
\R1:		bra	CheckFSP

/********************************************************************************
 * FLN			F: r1 -- r2					FLOATING EXT
 */
PRIM_WORD (FLN, "FLN")
		lea	forthlib::FloatLn, a0
		bra	DoFloatOp1

/********************************************************************************
 * FLOG			F: r1 -- r2					FLOATING EXT
 */
PRIM_WORD (FLOG, "FLOG")
		lea	forthlib::FloatLog, a0
		bra	DoFloatOp1

/********************************************************************************
 * FEXP			F: r1 -- r2					FLOATING EXT
 */
PRIM_WORD (FEXP, "FEXP")
		lea	forthlib::FloatExp, a0
		bra	DoFloatOp1

/********************************************************************************
 * F**			F: r1 r2 -- r3					FLOATING EXT
 */
PRIM_WORD (F_STAR_STAR, "F**")
		lea	forthlib::FloatPow, a0
		bra	DoFloatOp2

/*-------------------------------------------------------------------------------
 * f10**		    n --   					internal
 *                       F:   -- r
 * Return the ten-based power of n. This is an internal word used for floating
 * point conversion.
 *
 * Definition:   : F10**   S>D D>F   [ 10 FLN ] FLITERAL F*   FEXP ;
 */
INTERNAL_SEC_WORD (f_10_star_star, "f10**")
	XT(S_TO_D) XT(D_TO_F)  FLIT(1236190959,-29) XT(F_STAR) XT(FEXP) XT(EXIT)


/*==================================================================================
       Graphic Routines
  ==================================================================================**
   ***********************************************************************************/

/********************************************************************************
 * +GRAPHIC		--		 				????
 *
 * Enable 2 plane (4 shade) graphics mode. After this has been done, the screen
 * will be cleared, the cursor won't be visible any more, and any text send to
 * the terminal (e.g. by .") will be dark grey.
 */
PRIM_WORD (PLUS_GRAPHIC, "+GRAPHIC")
		jsr	forthlib::GraphicOn
		jmp	(NP)

/********************************************************************************
 * -GRAPHIC		--		 				????
 *
 * Disable 2 plane graphics mode. This will reactivate the cursor.
 */
PRIM_WORD (MINUS_GRAPHIC, "-GRAPHIC")
		jsr	forthlib::GraphicOff
		jmp	(NP)
		
/********************************************************************************
 * PIXEL!		n1 n2 n3 --		 			????
 *
 * Store pixel with color n1 to the coordinates (x=n2|y=n3).
 */
PRIM_WORD (PIXEL_STORE, "PIXEL!")
		move.w	(PSP)+, d1		// y
		move.w	(PSP)+, d0		// x
		move.w	(PSP)+, d2		// color
		jsr	forthlib::PutPixel
		jmp	(NP)
		
/********************************************************************************
 * PIXEL@		n1 n2 -- n3		 			????
 *
 * Read pixel at the coordinates (x=n1|y=n2) and return its color n3.
 */
PRIM_WORD (PIXEL_FETCH, "PIXEL@")
		move.w	(PSP)+, d1		// y
		move.w	(PSP)+, d0		// x
		jsr	forthlib::GetPixel
		move.w	d2, -(PSP)		// color
		jmp	(NP)

/********************************************************************************
 * FILL-SCREEN		u --			 			????
 *
 * Fill screen with color u.
 */
PRIM_WORD (FILL_SCREEN, "FILL-SCREEN")
		move.w	(PSP)+, d0
		jsr	forthlib::FillScreen
		jmp	(NP)
		

/********************************************************************************
 * BLANK-SCREEN		--			 			????
 *
 * Clear screen to color white. Faster than FILL-SCREEN.
 */
PRIM_WORD (BLANK_SCREEN, "BLANK-SCREEN")
		jsr	forthlib::BlankScreen
		jmp	(NP)
		

		
/*==================================================================================
       Variables and Constants
  ==================================================================================**
   ***********************************************************************************/


/*     Constants
 *=================================================================================*/

CONSTANT (BL, "BL", ' ')		// ( -- char )			CORE
SAME_PRIM_WORD (TRUE, "TRUE", MINUS_ONE)// ( -- flag )			CORE EXT
SAME_PRIM_WORD (FALSE, "FALSE", ZERO)	// ( -- flag )			CORE EXT
CONSTANT (C_PER_L, "C/L", 64)		// ( -- u )			????
PRIM_WORD (ZERO, "0")			// ( -- n )			????
		clr.w	-(PSP)
		jmp	(NP)
PRIM_WORD (ONE, "1")			// ( -- n )			????
		move.w	#1, -(PSP)
		jmp	(NP)
PRIM_WORD (TWO, "2")			// ( -- n )			????
		move.w	#2, -(PSP)
		jmp	(NP)
CONSTANT (FOUR, "4", 4)			// ( -- n )			????
PRIM_WORD (MINUS_ONE, "-1")		// ( -- n )			????
		move.w	#-1, -(PSP)
		DO_NEXT
CONSTANT (MINUS_TWO, "-2", -2)		// ( -- n )			????
CONSTANT (MINUS_FOUR, "-4", -4)		// ( -- n )			????
					
FCONSTANT (TEN_E, "10.E", $50000000, -27)// ( F: -- r )			????
FCONSTANT (ONE_E,  "1.E", $40000000, -30)// ( F: -- r )			????
FCONSTANT (ZERO_E, "0.E", $00000000, -32768)// ( F: -- r )		????

/**
 * All relative FORTH92 pointers are relative to that label `RelativePointerBase'.
 * The NULL pointer points to zero. The NULL pointer - 2 also points to zero.
 * That's why a NULL-link-field of a word will point to zero.
 */
		dc.w	0
RelativePointerBase:		
		dc.w	0

/*     User Variables
 *=================================================================================*/

/********************************************************************************
 * BASE			-- a-addr					CORE
 *
 * Definition:   VARIABLE SCR
 */
VARIABLE (BASE, "BASE", 10)	

/********************************************************************************
 * >IN			-- a-addr					CORE
 *
 * Definition:   VARIABLE >IN
 */
VARIABLE (TO_IN, ">IN", 0)

/********************************************************************************
 * STATE		-- a-addr					CORE
 *
 * Value:  executing = 0; compiling = 1
 *
 * Definition:   VARIABLE STATE
 */
VARIABLE (STATE, "STATE", 0)

/********************************************************************************
 * BLK			-- a-addr					BLOCK
 */
VARIABLE (BLK, "BLK", 0)

/********************************************************************************
 * SCR			-- a-addr					BLOCK EXT
 *
 * Definition:   VARIABLE SCR
 */
VARIABLE (SCR, "SCR", -1)

/********************************************************************************
 * PAD			-- c-addr					CORE EXT
 *
 * Definition:   VARIABLE SCR
 */
VARIABLE (PAD, "PAD", '  ')	
		ds.b	PAD_SIZE-2
/********************************************************************************
 * PRECISION		-- u						FLOATING EXT
 *
 * Definition:   7 VALUE PRECISION
 */
VALUE (PRECISION, "PRECISION", 7)	

/*     Internal Variables
 *=================================================================================*/

/*-------------------------------------------------------------------------------
 * input-buffer		-- a-addr					internal
 *
 * This is a 2VARIABLE object that contains the address and length of the input
 * buffer in case that input comes from the terminal, from an EVALUATEd string
 * or from a file.
 *
 * Definition:  2VARIABLE input-buffer
 */
INTERNAL_VAR (input_buffer, "input-buffer", 0)
		dc.w	0

/*-------------------------------------------------------------------------------
 * >line 		-- ud   					internal
 *
 * This is a 2VARIABLE object that contains the file position of the current
 * line.
 *
 * Definition:  2VARIABLE >line
 */
INTERNAL_VAR (to_line, ">line", 0)
		dc.w	0
/*-------------------------------------------------------------------------------
 * abort"-str		   -- c-addr					internal
 * 
 * Holds the address of the counted string of the last execution semantics of
 * ABORT".
 */
INTERNAL_VALUE (abort_quote_str, 'abort"-str', 0)

/*-------------------------------------------------------------------------------
 * tib			-- c-addr					internal
 *
 * Terminal input buffer
 *
 * Definition:
 *    CREATE tib {TIB-SIZE} ALLOT
 */
INTERNAL_VAR (tib, "tib", '  ')
		ds.b	TIB_SIZE-2

/*-------------------------------------------------------------------------------
 * fib			-- c-addr					internal
 *
 * File input buffer
 *
 * Definition:
 *    CREATE fib {FIB-SIZE} ALLOT
 */
INTERNAL_VAR (fib, "fib", '  ')
		ds.b	FIB_SIZE-2

/*-------------------------------------------------------------------------------
 * s"-buffer		-- c-addr					internal
 *
 * Buffer, used by S" in interpretation state
 *
 * Definition:
 *    CREATE s"-buffer {80} ALLOT
 */
INTERNAL_VAR (s_quote_buffer, 's"-buffer', '  ')
		ds.b	80-2

/*-------------------------------------------------------------------------------
 * parsed		-- c-addr u					internal
 *
 * Pointer and length of the string which was parsed at the last execution of
 * PARSE.
 *
 * Definition:   2VARIABLE parsed
 */
INTERNAL_VAR (parsed, "parsed", 0)
		dc.w	0

/*-------------------------------------------------------------------------------
 * interpreted		-- c-addr u					internal
 *
 * Pointer and length of the string which was most recently interpreted by 
 * 'interpret'.
 *
 * Definition:   2VARIABLE interpreted
 */
INTERNAL_VAR (interpreted, "interpreted", 0)
		dc.w	0

/*==================================================================================
       Dictionary Data
  ==================================================================================**
   ***********************************************************************************/

/******************************************************************************
 * HERE			-- addr						CORE
 *
 * Definition:   xxx VALUE HERE
 */
WORD_HEADER (HERE, "HERE")
		XT(DoCon)
		LBL(UserDict)
		XT(DoTo)

/*-------------------------------------------------------------------------------
 * #dict-data		-- c-addr					internal
 *
 * Returns the number of bytes from dict-data, that have to be stored and 
 * restored by MARKER.
 */
INTERNAL_CONST (num_dict_data, "#dict-data", dict_data_end-dict_data+2)

/*-------------------------------------------------------------------------------
 * dict-data		-- c-addr					internal
 *
 * Returns the address of the dictionary data, that have to be stored and 
 * restored by MARKER.
 */
INTERNAL_WORD_HEADER (dict_data, "dict-data")
		XT(DoVar)

	// table, that contains the addresses of the latest words of each word list
WordListTbl:	dc.w	1			// number of wordlists
FORTH_Wordlist:	dc.w	RELP(LINK_PTR)		// address of latest word in the FORTH word list
		ds.w	MAX_WORDLISTS		// addresses of latest words in other word lists

SearchOrder:	dc.w	1			// number of word lists in the search order
		dc.w	RELP(FORTH_Wordlist)	// FORTH wid
		ds.w	MAX_ORDER-1		// other wids
Current:	dc.w	RELP(FORTH_Wordlist)	// compilation word list (defaults to FORTH)

dict_data_end:

/****************************************
 * Area For User Defined Words
 */
		EVEN
UserDict:	ds.b	USER_SPACE
DictEnd:	ds.b	DICT_END_SPACE


/******************************************************************************
 * DATA
 ******************************************************************************/

		EVEN
BlockBuffer1:	dcb.b	1024, 0
BlockBuffer2:	dcb.b	1024, 0
AccessBoundary:			// Addresses beyond here mustn't be accessed by FORTH programs
SavedSP:	dc.l	0		// saved a7

_comment:	dc.b	"FORTH92 -- by David Khling", 0
		BSS

TempBuffer2:	ds.b	RS_MAX_OVERFLOW
ReturnStackTop:	ds.l	RS_SIZE		// Returnstack (grows from high to low memory)
ReturnStack:				// Pointer to bottom of Return Stack
TempBuffer1:	ds.b	RS_MAX_UNDERFLOW // savety reagion for return/float stack
FloatStackTop:	ds.b 	FS_SIZE*6	// Floating point stack
FloatStack:				// Pointer to bottom of Float Stack
		ds.b	12		// floating point stack savety mem for underflows
OldInts:	ds.l	REDIRECT_NUM	// array that holds the old vectors of the redirected ints

		END














